#ifndef __CUDA_LIGHTCUTS_H
#define __CUDA_LIGHTCUTS_H

#include "CUDA_Structs.h"
#include "CUDA_Includes.h"
#include "../ResourceManager/TextureManager/Texture.h"
#include "../GameStates/GameStateVBO.h"
#include "../Timer.h"

extern "C" void shadeAllPixels(dim3 grid, dim3 block, float4* colorTextureGPU, float4* cutSizeTextureGPU, float4* positionsInWorldCPU, float4* normalsInWorldCPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* lightCutsHeapGPU, float4* colorsInWorldGPU);
extern "C" void radiosityWithCUDA(dim3 grid, dim3 block, float4* colorTextureGPU, float4* positionsInWorldCPU, float4* normalsInWorldCPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* colorsInWorldGPU);
extern "C" void allocateTextureMemory(float4* positionsInWorldCPU, float4* normalsInWorldCPU, int width, int height);
extern "C" void deallocateTextureMemory(void);
extern "C" void setKernelParams(int offset_cpu, int maxThreadsPerBlock_cpu, int width_cpu, int height_cpu, int anzLights_cpu, int maximumCutSize_cpu, float sceneUnitsToMeters_cpu, int lightSpaceMatrices_size_cpu, float nearZ_cpu, float farZ_cpu, int ism_size_cpu, int ism_atlas_size_cpu, float sumOfWeightedIntensities_cpu);
extern "C" void freeKernelParams(void);


extern "C" void cudaTest_launchKernel(ClusterInCUDA* gpuPointer);


class CUDA_LightCuts
{
public:
//	static ClusterInCUDA* clustersInCUDA;
//	static 	VPL_InCUDA* representativeLights;

	static float4* positionsInWorld;
	static float4* normalsInWorld;
	static float4* colorsInWorld;

	//VPL structure, aligned memory to enable coalesced access:
	static float4* positionOfVPL;
	static float4* normalOfVPL_IntensityRatio; //intensity ratio wird in normalOfVPL_IntensityRatio.w gespeichert!!!
	static float4* colorOfVPL;
	static int4* shadowMapAndLightSpaceOfVPL;

	static float* lightMatricesCUDA;

	//Cluster structure, aligned memory to enable coalesced access:
	static int4* indices; //indices.r = leftChild, indices.g = rightChild, indices.b = index of representative Light (in datastructure above!)
	static float4* bounds_A;
	static float4* bounds_B;
	static float4* color_lch; //geometry term, visibility term, material term, intensity
	static float4* estRad_RadUpperBound_errorBound;

	static float4* colorTexture;
	static float4* cutSizeTexture;

	static GLuint colorTextureGL;
	static GLuint cutSizeTextureGL;

	static void deleteCudaArrays(void);

	static void cudaTest(void);
	static void renderImageWithCUDA();
	static void renderImageWithRadiosityCUDA();
};

#endif