#include "CUDA_PushPull.h"

#include "../GameStates/global.h"



int CUDA_PushPull::maxThreadsPerBlock = 256;
int CUDA_PushPull::maxGlobalMemoryInBytes = 0;

float* CUDA_PushPull::pushPulledData;
float* CUDA_PushPull::notPushPulledData;

void CUDA_PushPull::initCuda(void)
{   
	int device = cutGetMaxGflopsDeviceId();

	cudaGLSetGLDevice(device);

    cudaDeviceProp prop;
    CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceProperties(&prop, device));
    printf("Device name: %s\n", prop.name);
	printf("Major revision number: %i\n", prop.major);
	printf("Minor revision number: %i\n", prop.minor);
    printf("Max threads per block: %i\n", prop.maxThreadsPerBlock);
    printf("Max threads dim: %i %i %i\n", prop.maxThreadsDim[0], prop.maxThreadsDim[1], prop.maxThreadsDim[2]);
    printf("Max grid size: %i %i %i \n", prop.maxGridSize[0], prop.maxGridSize[1], prop.maxGridSize[2]);
    printf("Shared memory per block: %i bytes\n", prop.sharedMemPerBlock);
    printf("Total global memory: %i bytes\n", prop.totalGlobalMem);
    printf("Warp size: %i\n", prop.warpSize);
    printf("Memory pitch: %i\n", prop.memPitch);
    printf("Registers per block: %i\n", prop.regsPerBlock);
    printf("Clock rate: %i\n", prop.clockRate);
    printf("Texture alignment: %i\n", prop.textureAlignment);

	maxThreadsPerBlock = prop.maxThreadsPerBlock;
	maxGlobalMemoryInBytes = prop.totalGlobalMem * 0.8f;
	fflush(stdout);
}

void CUDA_PushPull::pullPushTexture()
{
	
	int resolutionOfISMs = ism_size;
	int anzLights = TextureManager::anzLights;
	int textureW = TextureManager::lightTextureW; 
	int textureH = TextureManager::lightTextureH;
	int maxLevels = pullPushMaxLevel;

	if(shadowMap!=0)
	{
		delete[] shadowMap;
		shadowMap = 0;
	}


	shadowMap = new float[ism_atlas_size * ism_atlas_size * 4];
	for (int i = 0; i < ism_atlas_size * ism_atlas_size * 4; i++)
	{
		shadowMap[i] = 1.0f;
	}
	
	glBindTexture(GL_TEXTURE_2D, ism_atlas);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, shadowMap);
	
	if(!(shadowPointsOnly && pullPush))
	{
		return;
	}

	float* ism_buffer = new float[ism_atlas_size * ism_atlas_size];
/*
	int cnt = 0;
	for(int a=0; a<ism_atlas_size*ism_atlas_size*4; a+=4)
	{
		if(shadowMap[a]<1.0f)
		{
			cout << shadowMap[a] << " ";
			cnt++;
			if(cnt>50)
				break;
		}
	}
*/

	//verdichten der values:
	for (int i = 0; i < ism_atlas_size * ism_atlas_size; i++)
	{
		ism_buffer[i] = shadowMap[i*4];
	}


/*	
	printf("BEFORE PushPull: \n");
	for (int i = 0; i < ism_atlas_size * ism_atlas_size; i++)
	{
		printf("%f  ", ism_buffer[i]);
		if ( i% (ism_size * textureW) == ((ism_size * textureW)-1))
		{
			printf("\n");
		}
	}
	*/
	float* retVal = new float[ism_atlas_size * ism_atlas_size];

//	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, ism_atlas_size, ism_atlas_size, GL_RGBA, GL_FLOAT, shadowMap);

	int numberOfFloats = textureW * textureH * resolutionOfISMs * resolutionOfISMs;

	printf("ISM ATLAS SIZE = %i\n", ism_atlas_size);

	if(printDebugOutput)
	{
		printf("GPU: Allocating %i floats, which equals ", numberOfFloats);
		printf("%i KB for ISM atlas.\n", (numberOfFloats * sizeof(float)) >> 10);
	}

	int maxNumOfThreadsPerBlock = 512; // device - specific, but shouldn't be less than that.

	float* arrayDataOnTheGPU;
	cudaMalloc((void**)&arrayDataOnTheGPU, numberOfFloats * sizeof(float));	
	cudaMemcpy(arrayDataOnTheGPU, ism_buffer, numberOfFloats * sizeof(float), cudaMemcpyHostToDevice);

	//#####################################################
	//			PULL - PHASE
	//#####################################################

	pushPullLevel* currentLevel = new pushPullLevel;
	pushPullLevel* previousLevel = new pushPullLevel;

	previousLevel->next = currentLevel;
	previousLevel->data = arrayDataOnTheGPU;
	previousLevel->thisLevel = 0;
	
	printf("PullPushing Texture...\n");
	Timer timTimer;

	for (int i = 1; i <= maxLevels; i++)
	{
		int numberOfAveragesOneThreadHasToCompute = 1; //depending on the size of the ism and ism - resolution, one thread may need to calculate more than one average
		int requiredNumberOfThreadsPerBlock = int((resolutionOfISMs * resolutionOfISMs) / (pow(4.0f,i))); //we assume square ISMs, and we assume them to be of size 2^i
		while (requiredNumberOfThreadsPerBlock > maxNumOfThreadsPerBlock)
		{
			numberOfAveragesOneThreadHasToCompute *= 4;
			requiredNumberOfThreadsPerBlock /= 4;
		}

		currentLevel->thisLevel = i;
		currentLevel->previous = previousLevel;

		if(printDebugOutput)
			printf("GPU: Allocating %i floats, which equals ", int(((numberOfFloats)/ (pow(4.0f,i)))));
		int temp = int((numberOfFloats / (pow(4.0f,i))) * sizeof(float));
		temp = temp >> 10;
		if(printDebugOutput)
			printf("%i KB for pushPull level %i\n", temp, i);

		cutilSafeCall(cudaMalloc((void**)&currentLevel->data,int ((numberOfFloats/ (pow(4.0f,i))) * sizeof(float))));

//		block.x = requiredNumberOfThreadsPerBlock;
//		block.y = 1;
//		block.z = 1;

//		push_Kernel<<<grid, block>>>(currentLevel->previous, currentLevel->data, numberOfAveragesOneThreadHasToCompute, i);
//		push_Kernel<<<grid, block>>>(currentLevel->data, numberOfAveragesOneThreadHasToCompute, i);
		launch_pull_kernel(resolutionOfISMs, anzLights, textureW, textureH, currentLevel->previous->data, currentLevel->data, numberOfAveragesOneThreadHasToCompute, requiredNumberOfThreadsPerBlock, i);
		cutilSafeCall(cudaThreadSynchronize());
		currentLevel->next = new pushPullLevel;

		previousLevel = currentLevel;
		currentLevel = currentLevel->next;
		currentLevel->previous = previousLevel;
	}

	//now previousLevel holds the data of the uppermost level of the pyramid!

	//#####################################################
	//			PUSH - PHASE
	//#####################################################

	for (int i = maxLevels; i > 0; i--)
	{
		currentLevel = previousLevel; //da bei der pull phase current level am schluss auf das i+1te level zeigt, was ja nicht existiert.
		previousLevel = currentLevel->previous; //in diesem fall bedeutet previouse 'weiter unten in der pyramide'

		int numberOfAveragesOneThreadHasToPushDown = 1;
		int requiredNumberOfThreadsPerBlock = (resolutionOfISMs * resolutionOfISMs) / int((pow(4.0f,i))); //we assume square ISMs, and we assume them to be of size 2^i
		while (requiredNumberOfThreadsPerBlock > maxNumOfThreadsPerBlock)
		{
			numberOfAveragesOneThreadHasToPushDown *= 4;
			requiredNumberOfThreadsPerBlock /= 4;
		}

		launch_push_kernel(resolutionOfISMs, anzLights, textureW, textureH, currentLevel->data, currentLevel->previous->data, numberOfAveragesOneThreadHasToPushDown, requiredNumberOfThreadsPerBlock, i);
		cutilSafeCall(cudaThreadSynchronize());
	}

	cudaMemcpy(retVal, currentLevel->previous->data, numberOfFloats * sizeof(float), cudaMemcpyDeviceToHost);

	const double dTimePerFrame = double(timTimer);
	timTimer.start();
	printf("Done (%f sek).\n", dTimePerFrame);

	currentLevel = previousLevel;
	for (int i = 0; i <= maxLevels; i++)
	{
		cutilSafeCall(cudaFree(currentLevel->data));
		currentLevel = currentLevel->next;
	}
	delete currentLevel;
	delete previousLevel;
	
	/*
	printf("AFTER PushPull: \n");
	for (int i = 0; i < ism_atlas_size * ism_atlas_size; i++)
	{
		printf("%f  ", retVal[i]);
		if ( i% (ism_size * textureW) == ((ism_size * textureW)-1))
		{
			printf("\n");
		}
	}
	*/
	//wieder "aufspreizen"
	for (int i = 0; i < ism_atlas_size * ism_atlas_size; i++)
	{
		shadowMap[(i*4)+0] = retVal[i];
		shadowMap[(i*4)+1] = retVal[i];
		shadowMap[(i*4)+2] = retVal[i];
	}

	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, ism_atlas_size, ism_atlas_size, GL_RGBA, GL_FLOAT, shadowMap);

//	delete[] shadowMap;
//	shadowMap = 0;

	delete[] retVal;
	retVal = 0;

	delete[] ism_buffer;
	ism_buffer = 0;
}

