#include "CUDA_Structs.h"
#include <stdio.h>
#include <cutil_math.h>
#include "../Math/Color.h"
/*
#define CUDA_SAFE_CALL( call) do {                                         \
   cudaError err = call;                                                    \
   if( cudaSuccess != err) {                                                \
       fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
               __FILE__, __LINE__, cudaGetErrorString( err) );              \
   } } while (0)
*/


texture<float4, 2, cudaReadModeElementType> positionsInWorldGPU;
texture<float4, 2, cudaReadModeElementType> normalsInWorldGPU;

cudaArray* cu_array1;
cudaArray* cu_array2;

__device__ __constant__ int offset;
__device__ __constant__ int maxThreadsPerBlock;
__device__ __constant__ int width;
__device__ __constant__ int height;
__device__ __constant__ int anzLights;
__device__ __constant__ int maximumCutSize;
__device__ __constant__ int lightSpaceMatrices_size;
__device__ __constant__ int ism_size;
__device__ __constant__ int ism_atlas_size;

__device__ __constant__ float sumOfWeightedIntensities;
__device__ __constant__ float nearZ;
__device__ __constant__ float farZ;
__device__ __constant__ float sceneUnitsToMeters;


void __device__ debugOutputHeap(float4* lightCutsHeap, int cutSize, int lightCutStartIndex)
{
#ifdef __DEVICE_EMULATION__
	printf("Heap = \n");
	for (int i = 0 + lightCutStartIndex; i < cutSize + lightCutStartIndex; i++)
	{
		printf("%i, %f, %f, %f\n", int(lightCutsHeap[i].x), lightCutsHeap[i].y, lightCutsHeap[i].z, lightCutsHeap[i].w);
	}
	printf("##################\n");
#endif
}

/*
 * Heapsort, add: Add the new node at the end and push it up (= swap with parent)
 * so long as the parents error is still smaller than the one of the new node.
 *
 */
void __device__ addNode(int clusterIndex, float estRad, float radUpperBound, float error, float4* lightCutsHeap, int cutSize, int lightCutStartIndex, int pixelToShade)
{
	//if (pixelToShade == 1)
	//{
//		printf("before adding:\n");
//		debugOutputHeap(lightCutsHeap, cutSize, lightCutStartIndex);
//	}

	//if (pixelToShade == 1 && cutSize == 3)
	//{
	//	printf("test msg...:\n");
	//}


	//We assume that cutSize refers to the cutSize BEFORE adding this node,
	//hence we access lightCutsHeap[cutSize] and not lightCutsHeap[cutSize-1] (index vs. counter...).
	
	//the values are as follows: indexOfTheCluster, estimated radiance, radianceUpperBound, error
	float4 newNode = make_float4(float(clusterIndex), estRad, radUpperBound, error);
	lightCutsHeap[cutSize + lightCutStartIndex] = newNode;

//	int i = 1;

	//Watch out for the (cutSize - 1) >> 1!
	//The parent of the new node can only be obtained by using the cutSize before it was added!
	//The index of the child is trivial (definition of heap), the child is placed at the end of the heap.
	int indexOfParentInHeap = ((cutSize - 1) >> 1) + lightCutStartIndex; 
	int indexOfChildInHeap = cutSize + lightCutStartIndex;

	//if the heap was previously empty, we need to stop here.
	if (cutSize == 0)
	{
		//if (pixelToShade == 1)
		//{
		//	printf("after adding:\n");
		//	debugOutputHeap(lightCutsHeap, cutSize, lightCutStartIndex);
		//}
		return;
	}

	float4 parentNode = lightCutsHeap[indexOfParentInHeap];
	float errorOfParent = parentNode.w;
	
	while (errorOfParent < error)
	{
		//swap elements:
		lightCutsHeap[indexOfChildInHeap] = parentNode;
		lightCutsHeap[indexOfParentInHeap] = newNode;

//		i++;

		indexOfChildInHeap = indexOfParentInHeap;
	//	indexOfParentInHeap = (indexOfParentInHeap - 1) >> 1;
		indexOfParentInHeap = ((indexOfParentInHeap - lightCutStartIndex - 1) >> 1) + lightCutStartIndex;

		if ((indexOfParentInHeap - lightCutStartIndex) < 0)
			break; //we are at the root, pushing up is complete

		parentNode = lightCutsHeap[indexOfParentInHeap];
		errorOfParent = parentNode.w;
	}

	//if (pixelToShade == 1)
	//{
	//	printf("after adding:\n");
	//	debugOutputHeap(lightCutsHeap, cutSize+1, lightCutStartIndex);
	//}

}

/*
 * Heapsort, sift down: Sift down the root (= swap with largest of its childs)
 * so long as the largest error of its childs is still larger than the error of the node
 * being sifted down.
 */
void __device__ siftDown(float4* lightCutsHeap, int cutSize, int lightCutStartIndex, int pixelToShade)
{
//	printf("sifting down, cutSize = %i\n", cutSize);
	if (cutSize <= 1)
		return;

	int i = 0;//(cutSize >> 1) - 1;

	int largestChild;
	float4 ithNode;
	float4 newNode;

	while ((2*i)+lightCutStartIndex + 1 <= (cutSize-1) + lightCutStartIndex) //root has left child
	{
		largestChild = (2*i) + lightCutStartIndex + 1;

		if (largestChild < (cutSize-1) + lightCutStartIndex) // root has right child
		{
			if (lightCutsHeap[largestChild].w < lightCutsHeap[largestChild+1].w)
			{
				largestChild = largestChild + 1; //right child is largest child
			}
		}
		
		ithNode = lightCutsHeap[i + lightCutStartIndex];
		newNode = lightCutsHeap[largestChild];

		if (ithNode.w < newNode.w)
		{
			//swap elements:
			lightCutsHeap[i + lightCutStartIndex] = newNode;
			lightCutsHeap[largestChild] = ithNode;

			i = largestChild - lightCutStartIndex;
		}
		else
		{
			break; //Heap - condition is true
		}
	}

}

/*
 * Heapsort, remove: Remove the node, replace it with the node
 * having the largest index (cutSize) and sift the node down.
 *
 *
 */
void __device__ removeNode(float4* lightCutsHeap, int cutSize, int lightCutStartIndex, int pixelToShade)
{
//	printf("before removing:\n");
//	debugOutputHeap(lightCutsHeap, cutSize, lightCutStartIndex);

//	float4 endOfHeap = lightCutsHeap[cutSize-1 + lightCutStartIndex];

	if (cutSize <= 1)
	{
		lightCutsHeap[0 + lightCutStartIndex] = make_float4(-1.0f);
//		printf("after removing:\n");
//		debugOutputHeap(lightCutsHeap, cutSize-1, lightCutStartIndex);
		return;
	}
	//We assume that cutSize refers to the cutSize BEFORE removing this node.
	
	//Swap root (=nodeToBeRemoved) with the last element
	lightCutsHeap[0 + lightCutStartIndex] = lightCutsHeap[(cutSize-1) + lightCutStartIndex];
	lightCutsHeap[(cutSize-1) + lightCutStartIndex] = make_float4(-1.0f);

	siftDown(lightCutsHeap, cutSize-1, lightCutStartIndex, pixelToShade);

//	printf("after removing:\n");
//	debugOutputHeap(lightCutsHeap, cutSize-1, lightCutStartIndex);
}

//visvis BOOKMARK
float __device__ getVisibility(ClusterInCUDA nodes, VPL_InCUDA repLights, int index, float4 pos)
{
	int tempVal = dot(normalize(make_float3(pos - repLights.positionOfVPL[index])), make_float3(repLights.normalOfVPL_IntensityRatio[index])) < 0.0f;
	
	//if position behindlight -> not visible
	if(tempVal>0)
		return 0.0f;

	int4 int_buffer; // lspmX, lspmY, ismX, ismY
	int_buffer.x = repLights.shadowMapAndLightSpaceOfVPL[index].z;
	int_buffer.y = repLights.shadowMapAndLightSpaceOfVPL[index].w;
	int_buffer.z  = repLights.shadowMapAndLightSpaceOfVPL[index].x;
	int_buffer.w  = repLights.shadowMapAndLightSpaceOfVPL[index].y;

	float lsMatrix[16];

	#pragma unroll 4
	for(tempVal=0; tempVal<4; tempVal++)
	{
		lsMatrix[tempVal] =		nodes.lightSpaceMatrices[int_buffer.y*lightSpaceMatrices_size*4 + int_buffer.x*4  + tempVal];
		lsMatrix[4+tempVal] =		nodes.lightSpaceMatrices[int_buffer.y*lightSpaceMatrices_size*4 + (int_buffer.x+1)*4  + tempVal];
		lsMatrix[8+tempVal] =		nodes.lightSpaceMatrices[(int_buffer.y+1)*lightSpaceMatrices_size*4 + int_buffer.x*4  + tempVal];
		lsMatrix[12+tempVal] =	nodes.lightSpaceMatrices[(int_buffer.y+1)*lightSpaceMatrices_size*4 + (int_buffer.x+1)*4  + tempVal];
	}
	
	//vertex in light space
	float4 vLSpace;

	vLSpace.x = pos.x*lsMatrix[0] + pos.y*lsMatrix[4] + pos.z*lsMatrix[8] + lsMatrix[12];
	vLSpace.y = pos.x*lsMatrix[1] + pos.y*lsMatrix[5] + pos.z*lsMatrix[9] + lsMatrix[13];
	vLSpace.z = pos.x*lsMatrix[2] + pos.y*lsMatrix[6] + pos.z*lsMatrix[10] + lsMatrix[14];
	vLSpace.w = pos.x*lsMatrix[3] + pos.y*lsMatrix[7] + pos.z*lsMatrix[11] + lsMatrix[15];
	
	//perspective division:
	vLSpace /= vLSpace.w;

	vLSpace.w = 0.0f; //so we can reuse the float4 in further calculations -> emulate float3
//paraboloid mapping:
	
	// compute the halfway vector between viewing direction and direction to vertex			  
//	float3 vLSpaceFloat3 = make_float3(vLSpace);
	float dist = sqrt(dot(vLSpace, vLSpace));
	vLSpace.x = (vLSpace.x / dist);
	vLSpace.y = (vLSpace.y / dist);
	vLSpace.z = (vLSpace.z / dist) - 1.0f;

  //adjust length so that z=-1
	vLSpace.x /= -vLSpace.z;
    vLSpace.y /= -vLSpace.z;		
		
//paraboloid code end
			
	//the vertex depth is the distance to the camera
	vLSpace.z = (dist - nearZ) / (farZ-nearZ);
	
	vLSpace.x = (vLSpace.x+1.0f) / 2.0f;
	vLSpace.y = (vLSpace.y+1.0f) / 2.0f;
			
	//shadow test

	//float distToM
	dist = sqrt(pow((vLSpace.x-0.5f), 2.0f) + pow((vLSpace.y-0.5f), 2.0f));
	
	int_buffer.x = int(vLSpace.x * ism_size); //ismOffsetX
	int_buffer.y = int(vLSpace.y * ism_size); //ismOffsetY

	//
	 // Reuse lsMatrix so we don't need to allocate even more memory:
	 // lsMatrix[0,1,2,3,4] = inShadow[5]
	 // lsMatrix[5] = distXLeft
	 // lsMatrix[6] = distYUp
	 // lsMatrix[7] = weightedSum
	 // lsMatrix[8] = visibility
	 // lsMatrix[9] = 
	 
	lsMatrix[5] = vLSpace.x*ism_size - float(int_buffer.x); //distXLeft 
	lsMatrix[6] = vLSpace.y*ism_size - float(int_buffer.y); //distYUp

	lsMatrix[7] = 1.0f;
	lsMatrix[8] = float(vLSpace.z <= nodes.shadowMaps[((int_buffer.w + int_buffer.y)*ism_atlas_size + int_buffer.z + int_buffer.x)]);

	//shadow test
	//return min(1.0f, (lsMatrix[8] / lsMatrix[7]));

	dist = float(((int_buffer.w + int_buffer.y - 1)*ism_atlas_size + int_buffer.z + int_buffer.x) >= 0.0f);
	if (dist > 0.0f)
	{
		lsMatrix[8] += dist * (float(vLSpace.z <= nodes.shadowMaps[((int_buffer.w + int_buffer.y - 1)*ism_atlas_size + int_buffer.z + int_buffer.x)]) * (1.0f- lsMatrix[6]));
		lsMatrix[7] += dist * (1.0f- lsMatrix[6]);
	}
	 
	dist = float(((int_buffer.w + int_buffer.y + 1)*ism_atlas_size + int_buffer.z + int_buffer.x) < ism_atlas_size*ism_atlas_size);
	if (dist > 0.0f)
	{
		lsMatrix[8] += dist * (float(vLSpace.z <= nodes.shadowMaps[((int_buffer.w + int_buffer.y + 1)*ism_atlas_size + int_buffer.z + int_buffer.x)]) * lsMatrix[6]);
		lsMatrix[7] += dist * lsMatrix[6];
	}
	
	dist = float(((int_buffer.w + int_buffer.y)*ism_atlas_size + int_buffer.z + int_buffer.x - 1) >= 0.0f);
	if (dist > 0.0f)
	{	
		lsMatrix[8] += dist * (float(vLSpace.z <= nodes.shadowMaps[((int_buffer.w + int_buffer.y)*ism_atlas_size + int_buffer.z + int_buffer.x - 1)]) * (1.0f- lsMatrix[5]));
		lsMatrix[7] += dist * (1.0f- lsMatrix[5]);
	}

	dist = float(((int_buffer.w + int_buffer.y)*ism_atlas_size + int_buffer.z + int_buffer.x + 1) < ism_atlas_size*ism_atlas_size);
	if (dist > 0.0f)
	{	
		lsMatrix[8] += dist * (float(vLSpace.z <= nodes.shadowMaps[((int_buffer.w + int_buffer.y)*ism_atlas_size + int_buffer.z + int_buffer.x + 1)]) * lsMatrix[5]);
		lsMatrix[7] += dist * lsMatrix[5];
	}

	//shadow test
	return min(1.0f, (lsMatrix[8] / lsMatrix[7]));
	
}

float __device__ getGeometryTermBound(ClusterInCUDA nodes, VPL_InCUDA repLights, int index, float3 pos, float3 worldNormal)
{
	//d = differenzvektor containing (width, height, depth) of bbox
	float4 d = nodes.bounds_B[index] - nodes.bounds_A[index];

	float3 nodesBoundsB = make_float3(nodes.bounds_B[index]);
	float3 nodesBoundsA = make_float3(nodes.bounds_A[index]);


	float3 v1 = make_float3(nodesBoundsB.x - nodesBoundsA.x, 0.0f, 0.0f);
	float3 v2 = make_float3(0.0f, nodesBoundsB.y - nodesBoundsA.y, 0.0f);
	float3 v3 = make_float3(0.0f, 0.0f, nodesBoundsB.z - nodesBoundsA.z);

	float3 v = pos - nodesBoundsA;

	//point is inside bounding box
	if (0.0f <= dot(v, v1) && dot(v,v1) <= dot(v1, v1) && 0.0f <= dot(v, v2) && dot(v,v2) <= dot(v2, v2) && 0.0f <= dot(v, v3) && dot(v,v3) <= dot(v3, v3))
	{
		return 1.0f;
	}
	

	float minDist = 99999999.9f;


	minDist = min( abs(nodesBoundsA.x - pos.x), abs(nodesBoundsB.x - pos.x));
	minDist = min(min( abs(nodesBoundsA.y - pos.y), abs(nodesBoundsB.y - pos.y)), minDist);
	minDist = min(min( abs(nodesBoundsA.z - pos.z), abs(nodesBoundsB.z - pos.z)), minDist);
	
	minDist *= sceneUnitsToMeters;

	return min(1.0f, (30.0f)/((minDist+3.0f)*(minDist+3.0f)));
}

//the geometry term if the representative light is used for the shading:
float __device__ getGeometryTerm(ClusterInCUDA nodes, VPL_InCUDA repLights, int index, float3 pos, float3 worldNormal)
{
	float distance = sceneUnitsToMeters*(length(make_float3(repLights.positionOfVPL[index])-pos));

	return min(1.0f, (30.0f)/((distance+3.0f)*(distance+3.0f)));
}

//TODO pixelToShade aus parameterliste entfernen, nur zum debuggen da
float __device__ getMaterialTermBound(ClusterInCUDA nodes, VPL_InCUDA repLights, int index, float4 pos, float4 worldNormal, int pixelToShade)
{
	// If the node is already a leaf, this method isn't called, so there 
	// really is no reason to do that check, just return 1.0f;
	
	//Cluster BBOX
	//d = differenzvektor containing (width, height, depth) of bbox

	float4 d = nodes.bounds_B[index] - nodes.bounds_A[index];
	
	float4 L = repLights.positionOfVPL[index] - pos;
	L = normalize(L);

//	float4 lightNormal = repLights.normalOfVPL_IntensityRatio[index];

	float3 bbox[8] = {make_float3(nodes.bounds_A[index]), make_float3(nodes.bounds_A[index] + make_float4(d.x, 0.0f, 0.0f, 0.0f)),
					make_float3(nodes.bounds_A[index] + make_float4(0.0f, d.y, 0.0f, 0.0f)), make_float3(nodes.bounds_A[index] + make_float4(0.0f, 0.0f, d.z, 0.0f)),		
					make_float3(nodes.bounds_A[index] + make_float4(d.x, d.y, 0.0f, 0.0f)), make_float3(nodes.bounds_A[index] + make_float4(d.x, 0.0f, d.z, 0.0f)),
					make_float3(nodes.bounds_A[index] + make_float4(0.0f, d.y, d.z, 0.0f)), make_float3(nodes.bounds_B[index])};

	//we want to transform bbox + shaded position in a way, that shaded position = origin and normal vector = zAxis
	
	//TRANSLATION VECTOR for bounding box
	float3 translateLightToOrigin = make_float3(-pos);

	//rotation matrix, explained later
	float mat[9];


	//if the normal vector isn't already (0,0,1):
	bool normalVectorIsFinal = (worldNormal.x == 0.0f && worldNormal.y == 0.0f);
	if (!normalVectorIsFinal)
	{
		//SETUP ROTATION MATRIX so that normal vector of shaded position = zAxis
		float alpha = acos(worldNormal.z / length(worldNormal)); //drehwinkel alpha um folgende drehachse im raum:
			
		//v=Drehachse im Raum:
		float vX = worldNormal.y;
		float vY = (-1.0f)*worldNormal.x;
		float vZ = 0.0f;
		float lenAxis = sqrt(vX*vX+vY*vY);
		vX/=lenAxis;
		vY/=lenAxis;

		float cosA = cos(alpha);
		float sinA = sin(alpha);

		float oneMinusCosAlpha = 1.0f - cosA;
		//ROT MAT:
		mat[0] = cosA + vX*vX*oneMinusCosAlpha;
		mat[1] = vX*vY*oneMinusCosAlpha-vZ*sinA;
		mat[2] = vX*vZ*oneMinusCosAlpha+vY*sinA;
		mat[3] = vY*vX*oneMinusCosAlpha+vZ*sinA;
		mat[4] = cosA+vY*vY*oneMinusCosAlpha;
		mat[5] = vY*vZ*oneMinusCosAlpha-vX*sinA;
		mat[6] = vZ*vX*oneMinusCosAlpha-vY*sinA;
		mat[7] = vZ*vY*oneMinusCosAlpha+vX*sinA;
		mat[8] = cosA+vZ*vZ*oneMinusCosAlpha;
	}

	//transform all 8 vertices of bbox
	#pragma unroll 8
	for(int i=0; i<8; i++)
	{
		//TRANSLATE BBOX (and light) so that shaded position = origin	
		bbox[i] = bbox[i] + translateLightToOrigin;

		if (!normalVectorIsFinal)
		{
			//ROTATE BBOX
			float P[3] = {bbox[i].x, bbox[i].y, bbox[i].z};	
			bbox[i].x = P[0]*mat[0] + P[1]*mat[1] + P[2]*mat[2];
			bbox[i].y = P[0]*mat[3] + P[1]*mat[4] + P[2]*mat[5];
			bbox[i].z = P[0]*mat[6] + P[1]*mat[7] + P[2]*mat[8];	
		}
		
	}

	//GET NEW axis aligned bounds:
	float maxX, maxY, maxZ;
	maxX = maxY = maxZ = -999999.9f;
	float minX, minY, minZ;
	minX = minY = minZ = 999999.9f;
	#pragma unroll 8
	for(int i=0; i<8; i++)
	{
		if(bbox[i].x > maxX) maxX = bbox[i].x;
		if(bbox[i].x < minX) minX = bbox[i].x;
		if(bbox[i].y > maxY) maxY = bbox[i].y;
		if(bbox[i].y < minY) minY = bbox[i].y;
		if(bbox[i].z > maxZ) maxZ = bbox[i].z;
		if(bbox[i].z < minZ) minZ = bbox[i].z;
	}

	//FIND MAXIMUM COSINE:
	float minX2;
	float minY2;
	if(maxX*minX<0.0f) //0 liegt dazwischen |@CUDA: small if's (<= 4 commands) will (hopefully) be optimized by CUDA
		minX2 = 0.0f;
	else
		minX2 = min(maxX*maxX, minX*minX);
	if(maxY*minY<0.0f) //0 liegt dazwischen
		minY2 = 0.0f;
	else
		minY2 = min(maxY*maxY, minY*minY);
//	maxX2 = max(maxX*maxX, minX*minX);
//	maxY2 = max(maxY*maxY, minY*minY);
//	maxZ2 = max(maxZ*maxZ, minZ*minZ);

	//now calculate max cosine

	if(maxZ>=0.0f)
	{
		if(sqrt(minX2+minY2+maxZ*maxZ) == 0.0f)
			return 1.0f;
		else
			return maxZ / sqrt(minX2+minY2+ maxZ*maxZ);
	}	
	else
		return 0.0f; 
}

void __global__ radiosityWithCUDA(float4* colorTextureGPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* colorsInWorldGPU)
{
	//remember: anzLights is already set (const memory) and can be used.

	float3 accColor = make_float3(0.0f);

	//every block deals with blockDim.x pixels
	int anzPixels = width * height; 

	//Looking at the current blockIdx (and the offset) we can determine where the current blocks first pixelToShade is: 
	int startIndex = (blockIdx.x * blockDim.x) + offset; 
	
	
//	startIndex += width * (((blockIdx.x * blockDim.x) + threadIdx.x)/ width); //ganzzahldivision, nicht wegkuerzen!
	
	//the pixel this thread has to shade
	int pixelToShade = startIndex + threadIdx.x;    

	//We use cached texture memory to look up world positions and normals.
	//since we can't store these values in a 1D texture (CUDA size restrictions)
	//we use 2D Textures, with the coordinates x and y.
	int y = pixelToShade / width;
	int x = pixelToShade - (y * width);

	float4 posInWorld4 = tex2D(positionsInWorldGPU, float(x), float(y));
	float4 normalInWorld4 = tex2D(normalsInWorldGPU, float(x), float(y));

	float3 posInWorld3 = make_float3(posInWorld4);
	float3 normalInWorld3 = make_float3(normalInWorld4); 

	// The last block can be too large, we need to throw threads away that would create index-out-of-bounds exceptions:
	if (pixelToShade >= anzPixels) // '>=' because anzPixels refers to the actual number (and not the highest possible index, which is anzPixels - 1)
	{
		return;
	}

	if(length(normalInWorld3)<=0.0f)//don't use the 'w' component to calculate length of normal
	{
		//return Point(1.0f, 0.0f, 0.0f); //red
		colorTextureGPU[pixelToShade] = make_float4(1.0f, 0.0f, 0.0f, 0.0f);
		return;	
	}

	float4 base;
	float4 final;

	float3 color;
	float3 lightNorm;
	float materialTerm;
	float intensity;
	float estimatedRadiance;

	float visibilityTerm;
	float geometryTerm;

	float3 lightPos;
	float3 L;


	//since the 'real' lights (the clusters consisting only of one light e.g. the leafs of the lighttree)
	//were added to the array at the very beginning, we know that we can access all leaf nodes by 
	//using a loop like this:
	for (int i = 0; i < anzLights; i++)
	{
		lightPos = make_float3(vplPointersGPU.positionOfVPL[i]);
		L = lightPos - posInWorld3;
		L = normalize(L);
		lightNorm = make_float3(vplPointersGPU.normalOfVPL_IntensityRatio[i]);


		intensity = vplPointersGPU.colorOfVPL[i].w;
		geometryTerm = getGeometryTerm(clusterPointersGPU, vplPointersGPU, i, posInWorld3, normalInWorld3);
		materialTerm = max(0.0f,(dot(normalInWorld3, L) * dot(lightNorm, L * -1.0f)));
		visibilityTerm = getVisibility(clusterPointersGPU, vplPointersGPU, i, posInWorld4);

		estimatedRadiance = visibilityTerm * geometryTerm * materialTerm * intensity;

		color = make_float3(vplPointersGPU.colorOfVPL[i]);
		accColor += estimatedRadiance * color;
	}

	//IRADIOSITY SHADING	
	base = colorsInWorldGPU[pixelToShade];
	
	accColor = accColor * (1.0f / sumOfWeightedIntensities);

	final.x = accColor.x * base.x;
	final.y = accColor.y * base.y;
	final.z = accColor.z * base.z;

	colorTextureGPU[pixelToShade] = final;
}


void __global__ shadeLightcutsWithCUDA(float4* colorTextureGPU, float4* cutSizeTextureGPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* lightCutsHeap, float4* colorsInWorldGPU)
{
	float3 accColor = make_float3(0.0f);

	//every block deals with blockDim.x pixels
	int anzPixels = width * height; 

	//Looking at the current blockIdx (and the offset) we can determine where the current blocks first pixelToShade is: 
	int startIndex = (blockIdx.x * blockDim.x) + offset; 
	
	//regardless whether this thread deals with a pixel in an odd or even row, this index points to the one in the even row.
	int lightCutStartIndex = (startIndex + threadIdx.x);
	lightCutStartIndex = (lightCutStartIndex - offset) * maximumCutSize;

	//edit: until lightcuts are correct, ignore the following comments
	//Instead of going row by row, we do this:
	//..
	//|thisKernelLaunch|thisKernelLaunch|..
	//|nextKernelLaunch|nextKernelLaunch|..
	//|thisKernelLaunch|thisKernelLaunch|..
	//..
	//the offset parameter specifies whether the current Kernel deals with odd or even rows.
//	startIndex += width * (((blockIdx.x * blockDim.x) + threadIdx.x)/ width); //ganzzahldivision, nicht wegkuerzen!
	
	//the pixel this thread has to shade
	int pixelToShade = startIndex + threadIdx.x;    

	//We use cached texture memory to look up world positions and normals.
	//since we can't store these values in a 1D texture (CUDA size restrictions)
	//we use 2D Textures, with the coordinates x and y.
	int y = pixelToShade / width;
	int x = pixelToShade - (y * width);

	float4 posInWorld4 = tex2D(positionsInWorldGPU, float(x), float(y));
	float4 normalInWorld4 = tex2D(normalsInWorldGPU, float(x), float(y));

	float3 posInWorld3 = make_float3(posInWorld4);
	float3 normalInWorld3 = make_float3(normalInWorld4); 

	int anzClusters = (2*anzLights) - 1;
#ifdef __DEVICE_EMULATION__
	FILE* file = fopen("test.test", "a");
	fprintf(file, "\n######################\nNEUER PIXEL %i\n", pixelToShade);
#endif
	// The last block can be too large, we need to throw threads away that would create index-out-of-bounds exceptions:
	if (pixelToShade >= anzPixels) // '>=' because anzPixels refers to the actual number (and not the highest possible index, which is anzPixels - 1)
	{
		#ifdef __DEVICE_EMULATION__
		fclose(file);
		#endif
		return;
	}

	if(length(normalInWorld3)<=0.0f)//don't use the 'w' component to calculate length of normal
	{
		#ifdef __DEVICE_EMULATION__
		fclose(file);
		#endif
		//return Point(1.0f, 0.0f, 0.0f); //red
		colorTextureGPU[pixelToShade] = make_float4(1.0f, 0.0f, 0.0f, 0.0f);

		cutSizeTextureGPU[pixelToShade] = make_float4(0.0f, 0.0f, 0.0f, 0.0f);
		return;	
	}

	float sumIlluminationEstimates = 0.0f;
	float sumRadianceUpperBounds = 0.0f;
	float criterion = 0.02f;

	float4 base;
	float4 final;
	float4 csize;

	
	float3 color;
	float3 lightNorm;
	float materialTerm;
	float intensity;
	float estimatedRadiance;

	float visibilityTermBound = 1.0f;
	float geometryTermBound;
	float materialTermBound;
	float radianceUpperBound;

	float visibilityTerm;
	float geometryTerm;

	float3 lightPos;
	float3 L;

	/*
	 * Insert root. Calculate estimated radiance / radiance upper bound.
	 *
	 */

	lightPos = make_float3(vplPointersGPU.positionOfVPL[anzClusters-1]);
	lightNorm = make_float3(vplPointersGPU.normalOfVPL_IntensityRatio[anzClusters-1]);
	L = lightPos - posInWorld3;
	L = normalize(L);


	visibilityTerm = getVisibility(clusterPointersGPU, vplPointersGPU, anzClusters-1, posInWorld4);	
	geometryTerm = getGeometryTerm(clusterPointersGPU, vplPointersGPU, anzClusters-1, posInWorld3, normalInWorld3);
	materialTerm = max(0.0f,(dot(normalInWorld3, L) * dot(lightNorm, L * -1.0f))); //shade pixel with this light/cluster
	intensity = vplPointersGPU.colorOfVPL[anzClusters-1].w;

	estimatedRadiance = visibilityTerm * geometryTerm * materialTerm * intensity;

	int leftChildLC = clusterPointersGPU.indices[anzClusters-1].x;
	int rightChildLC = clusterPointersGPU.indices[anzClusters-1].x;

	bool leftChildIsLeaf = leftChildLC < 0;
	bool rightChildIsLeaf = rightChildLC < 0;

	if (!leftChildIsLeaf)
	{
		//errechne radiance upper bound von nextNode 
		geometryTermBound = getGeometryTermBound(clusterPointersGPU, vplPointersGPU, anzClusters-1, posInWorld3, normalInWorld3);
		materialTermBound = getMaterialTermBound(clusterPointersGPU, vplPointersGPU, anzClusters-1, posInWorld4, normalInWorld4, pixelToShade);
		radianceUpperBound = visibilityTermBound*geometryTermBound*materialTermBound*intensity;

		/*
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("Initial radiance upper bound = %f (%f, %f, %f, %f)\n", radianceUpperBound, visibilityTermBound, geometryTermBound, materialTermBound, intensity);
				printf("Initial estimated radiance   = %f (%f, %f, %f, %f)\n", estimatedRadiance, visibilityTerm, geometryTerm, materialTerm, intensity);
			}	
		#endif
			*/
	}
	else
	{
		radianceUpperBound = estimatedRadiance;
	}

	/*
	#ifdef __DEVICE_EMULATION__
	if (pixelToShade == 1275)
	{
		printf("posInWorld = %f, %f, %f, %f\n", posInWorld4.x, posInWorld4.y, posInWorld4.z, posInWorld4.w);
		printf("normalInWorld = %f, %f, %f, %f\n\n", normalInWorld4.x, normalInWorld4.y, normalInWorld4.z, normalInWorld4.w);
	}	
	#endif
	*/
	
	sumIlluminationEstimates = estimatedRadiance;
	sumRadianceUpperBounds = radianceUpperBound;
	color = make_float3(vplPointersGPU.colorOfVPL[anzClusters-1]);
	accColor += color * estimatedRadiance;


	int cutSize = 0;

	addNode(anzClusters-1, estimatedRadiance, radianceUpperBound, abs(radianceUpperBound - estimatedRadiance), lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);
	cutSize += 1;


	bool largestErrorIsSmallEnough = false;

//REFINEMENT:
	#pragma unroll 1
	do
	{
		float4 largestNode = lightCutsHeap[0 + lightCutStartIndex];
		int indexOfClusterToSplit = int(largestNode.x);
		float largestError = largestNode.w;
#ifdef __DEVICE_EMULATION__
		fprintf(file, "\nBeginn einer neuen Iteration\n");
		fprintf(file, "Cluster - position mit dem groessten Fehler: %.2f, %.2f, %.2f\n", vplPointersGPU.positionOfVPL[indexOfClusterToSplit].x, vplPointersGPU.positionOfVPL[indexOfClusterToSplit].y, vplPointersGPU.positionOfVPL[indexOfClusterToSplit].z);
		fprintf(file, "largest.estimatedRadiance,radUpperBound = %.2f, %.2f\n", largestNode.y, largestNode.z);
		fprintf(file, "criterion = %.2f\n", sumIlluminationEstimates * criterion);
#endif
		
		largestErrorIsSmallEnough = (largestError <= sumIlluminationEstimates * criterion);
	/*	
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("Before Splitting:\n");
	//			debugOutputHeap(lightCutsHeap, cutSize, lightCutStartIndex);
				printf("cutSize = %i\n", cutSize);
				printf("largestError = %f\n", largestError);
				printf("isSmallEnough = %i\n\n", largestErrorIsSmallEnough);
				printf("Splitting node with repLight = %f, %f, %f, intensity = %f\n", vplPointersGPU.positionOfVPL[indexOfClusterToSplit].x, vplPointersGPU.positionOfVPL[indexOfClusterToSplit].y, vplPointersGPU.positionOfVPL[indexOfClusterToSplit].z, vplPointersGPU.colorOfVPL[indexOfClusterToSplit].w);
			}
		#endif
*/
		if (largestErrorIsSmallEnough)// && cutSize > 1)
		{
			break;
		}

		int leftChild = clusterPointersGPU.indices[indexOfClusterToSplit].x;
		int rightChild = clusterPointersGPU.indices[indexOfClusterToSplit].y;
			

		if (leftChild == -1 || rightChild == -1)
		{
			break;
		}

		leftChildLC = clusterPointersGPU.indices[leftChild].x;
		rightChildLC = clusterPointersGPU.indices[rightChild].x;

		//I'm just checking the left child, should be equivalent to checking both
		leftChildIsLeaf = leftChildLC < 0;
		rightChildIsLeaf = rightChildLC < 0;

		//remove cluster with the largest error:
		sumIlluminationEstimates -= largestNode.y;
		sumRadianceUpperBounds -= largestNode.z;

		color = make_float3(vplPointersGPU.colorOfVPL[int(largestNode.x)]);
		accColor -= color * largestNode.y;

/*
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("Subtracting: = %f, %f, \n", largestNode.y, largestNode.z);
				printf("New Values: %f, %f\n", sumIlluminationEstimates, sumRadianceUpperBounds);
				printf("Inserting left child.\n");
			}
		#endif
*/
		//Remove root from heap
		removeNode(lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);
		cutSize -= 1;
	
		//Calculate est. radiance / radiance upper bound for left child:
		lightPos = make_float3(vplPointersGPU.positionOfVPL[leftChild]);
		L = lightPos - posInWorld3;
		L = normalize(L);
		lightNorm = make_float3(vplPointersGPU.normalOfVPL_IntensityRatio[leftChild]);


		geometryTerm = getGeometryTerm(clusterPointersGPU, vplPointersGPU, leftChild, posInWorld3, normalInWorld3);
		visibilityTerm = getVisibility(clusterPointersGPU, vplPointersGPU, leftChild, posInWorld4);
		materialTerm = max(0.0f,(dot(normalInWorld3, L) * dot(lightNorm, L * -1.0f)));
		intensity = vplPointersGPU.colorOfVPL[leftChild].w;
		estimatedRadiance = visibilityTerm * geometryTerm * materialTerm * intensity;

/*
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("visibilityTerm = %f\n", visibilityTerm);
				printf("geometryTerm = %f\n", geometryTerm);
				printf("materialTerm = %f\n", materialTerm);
			}
		#endif
	*/
		
		if (!leftChildIsLeaf)
		{
			geometryTermBound = getGeometryTermBound(clusterPointersGPU, vplPointersGPU, leftChild, posInWorld3, normalInWorld3);
			materialTermBound = getMaterialTermBound(clusterPointersGPU, vplPointersGPU, leftChild, posInWorld4, normalInWorld4, pixelToShade);
			radianceUpperBound = visibilityTermBound * geometryTermBound * materialTermBound * intensity;
			

			addNode(leftChild, estimatedRadiance, radianceUpperBound, abs(radianceUpperBound - estimatedRadiance), lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);

			sumIlluminationEstimates += estimatedRadiance;
			sumRadianceUpperBounds += radianceUpperBound;
#ifdef __DEVICE_EMULATION__
			fprintf(file, "Left Child, estimatedRadiance, upperBound = %.2f, %.2f\n", estimatedRadiance, radianceUpperBound);
#endif
			/*
			#ifdef __DEVICE_EMULATION__
				if (pixelToShade == 1275)
				{
					printf("visibilityTermBound = %f\n", visibilityTermBound);
					printf("geometryTermBound = %f\n", geometryTermBound);
					printf("materialTermBound = %f\n", materialTermBound);
				}
			#endif
				*/
		}
		else
		{

			addNode(leftChild, estimatedRadiance, estimatedRadiance, 0.0f, lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);
			sumIlluminationEstimates += estimatedRadiance;
			sumRadianceUpperBounds += estimatedRadiance;
#ifdef __DEVICE_EMULATION__
			fprintf(file, "Left Child, estimatedRadiance, upperBound = %.2f, %.2f\n", estimatedRadiance, estimatedRadiance);
#endif
		//	radianceUpperBound = estimatedRadiance;
		}

		//add left child to heap
		cutSize += 1;


		color = make_float3(vplPointersGPU.colorOfVPL[leftChild]);
		accColor += color * estimatedRadiance;
/*
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("Adding: = %f, %f, \n", estimatedRadiance, radianceUpperBound);
				printf("New Values: %f, %f\n", sumIlluminationEstimates, sumRadianceUpperBounds);
				printf("color = %f, %f, %f\n", vplPointersGPU.colorOfVPL[leftChild].x, vplPointersGPU.colorOfVPL[leftChild].y, vplPointersGPU.colorOfVPL[leftChild].z);
				printf("Inserting right child\n");
			}
		#endif
	*/


		//Calculate est. radiance / radiance upper bound for left child:						
		lightPos = make_float3(vplPointersGPU.positionOfVPL[rightChild]);
		L = lightPos - posInWorld3;
		L = normalize(L);
		lightNorm = make_float3(vplPointersGPU.normalOfVPL_IntensityRatio[rightChild]);

		geometryTerm = getGeometryTerm(clusterPointersGPU, vplPointersGPU, rightChild, posInWorld3, normalInWorld3);
		visibilityTerm = getVisibility(clusterPointersGPU, vplPointersGPU, rightChild, posInWorld4);
		materialTerm = max(0.0f,(dot(normalInWorld3, L) * dot(lightNorm, L * -1.0f)));
		intensity = vplPointersGPU.colorOfVPL[rightChild].w;

		estimatedRadiance = visibilityTerm * geometryTerm * materialTerm * intensity;
	/*	
		#ifdef __DEVICE_EMULATION__
			if (pixelToShade == 1275)
			{
				printf("visibilityTerm = %f\n", visibilityTerm);
				printf("geometryTerm = %f\n", geometryTerm);
				printf("materialTerm = %f\n", materialTerm);
			}
		#endif
*/
		if (!rightChildIsLeaf)
		{
			geometryTermBound = getGeometryTermBound(clusterPointersGPU, vplPointersGPU, rightChild, posInWorld3, normalInWorld3);
			materialTermBound = getMaterialTermBound(clusterPointersGPU, vplPointersGPU, rightChild, posInWorld4, normalInWorld4, pixelToShade);	
			radianceUpperBound = visibilityTermBound * geometryTermBound * materialTermBound * intensity;
				

			addNode(rightChild, estimatedRadiance, radianceUpperBound, abs(radianceUpperBound - estimatedRadiance), lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);
			

			sumIlluminationEstimates += estimatedRadiance;
			sumRadianceUpperBounds += radianceUpperBound;
#ifdef __DEVICE_EMULATION__
			fprintf(file, "Right Child, estimatedRadiance, upperBound = %.2f, %.2f\n", estimatedRadiance, radianceUpperBound);
#endif
		}
		else
		{
	//		radianceUpperBound = estimatedRadiance;


			addNode(rightChild, estimatedRadiance, estimatedRadiance, 0.0f, lightCutsHeap, cutSize, lightCutStartIndex, pixelToShade);
			
			sumIlluminationEstimates += estimatedRadiance;
			sumRadianceUpperBounds += estimatedRadiance;

			#ifdef __DEVICE_EMULATION__
			fprintf(file, "Right Child, estimatedRadiance, upperBound = %.2f, %.2f\n", estimatedRadiance, estimatedRadiance);
#endif
		}


		cutSize += 1;
		
		color = make_float3(vplPointersGPU.colorOfVPL[rightChild]);
		accColor += color * estimatedRadiance;

/*	
		#ifdef __DEVICE_EMULATION__
			printf("cutSize = %i\n", cutSize);
			debugOutputHeap(lightCutsHeap, cutSize);
		#endif
*/

//		printf("AFTER splitting:\n");
//		debugOutputHeap(lightCutsHeap, cutSize, lightCutStartIndex);
	}//while...
	while(cutSize <= maximumCutSize && (!largestErrorIsSmallEnough));

//	printf("pixelToShade = %i\n", pixelToShade);
	/*
	printf("CalculatedCut:\n");
	for (int i = 0; i < cutSize; i++)
	{
		printf("%i\n", int(lightCutsGPU[lightCutStartIndex + i].x));
	}
	*/
//	printf("error = %f\n",abs(sumIlluminationEstimates-sumRadianceUpperBounds)/sumIlluminationEstimates); 
//	printf("\n");


//TODO fuer wiederverwendung einschalten oder umschreiben!!
//	lightCutsGPU[lightCutStartIndex].w = float(cutSize);




	//LIGHTCUTS SHADING	
	base = colorsInWorldGPU[pixelToShade];
	
//	intensity = (sumIlluminationEstimates / sumOfWeightedIntensities);
	intensity = 1.0f / sumOfWeightedIntensities;//sumIlluminationEstimates;
	intensity = min (1.0f, intensity);

	accColor = accColor * intensity;

//	printf("accColor = %f, %f, %f\n", sumIlluminationEstimates / accColor.x, sumIlluminationEstimates / accColor.y, sumIlluminationEstimates / accColor.z);

	final.x = accColor.x * base.x;
	final.y = accColor.y * base.y;
	final.z = accColor.z * base.z;

	colorTextureGPU[pixelToShade] = final;

//	final = (intensity * base);
//	final.x *= accColor.x / sumOfWeightedIntensities;
//	final.y *= accColor.y / sumOfWeightedIntensities;
//	final.z *= accColor.z / sumOfWeightedIntensities;

	#ifdef __DEVICE_EMULATION__
		if (pixelToShade == 1275)
		{
			printf("cutSize = %i\n", cutSize);
			printf("sumIlluminationEstimates = %f\n", sumIlluminationEstimates);
			printf("sumOfWeightedIntensities = %f\n", sumOfWeightedIntensities);
			printf("intensity = %f\n", intensity);
			printf("color = %f, %f, %f\n", final.x, final.y, final.z);
		}
	#endif

	//CUTSIZE VISUALIZATION:
	csize = make_float4(float(cutSize)/anzLights);
	csize.w = 1.0f;

	cutSizeTextureGPU[pixelToShade] = csize;

	#ifdef __DEVICE_EMULATION__
		fprintf(file, "CUTSIZE = %i\n", cutSize);
		fclose(file);
	#endif

	/*
	#ifdef __DEVICE_EMULATION__	
		printf("index = %i\n", pixelToShade);
	#endif
	*/

	/*
	#ifdef __DEVICE_EMULATION__	
	if (x == 0 && blockX == 0 && blockY == 0)
	{
		printf("AUF DER GPU:\n");

		for (int i = 0; i < (anzLights*2)-1; i++)
		{
		//	printf("repLicht: posX = %f; posY = %f; posZ = %f\n", vplData[i].positionX, vplData[i].positionY, vplData[i].positionZ);
			printf("%i'th cluster: repLicht: posX = %f; posY = %f; posZ = %f || leftChild = %i, rightChild = %i\n", i, clusterData[i].representativeLight.positionX, clusterData[i].representativeLight.positionY, clusterData[i].representativeLight.positionZ, clusterData[i].leftChild, clusterData[i].rightChild);
		}
		
	}
	#endif
	*/
}

void __global__ cudaTestKernel()
{
}

extern "C" void cudaTest_launchKernel()
{
	dim3 grid = dim3(1);
	dim3 block = dim3(10, 1, 1);
	cudaTestKernel<<<grid, block>>>();
}

extern "C" void setKernelParams(int offset_cpu, int maxThreadsPerBlock_cpu, int width_cpu, int height_cpu, int anzLights_cpu, int maximumCutSize_cpu, float sceneUnitsToMeters_cpu, int lightSpaceMatrices_size_cpu, float nearZ_cpu, float farZ_cpu, int ism_size_cpu, int ism_atlas_size_cpu, float sumOfWeightedIntensities_cpu)
{
	//Integer params:
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("width", &width_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("height", &height_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("offset", &offset_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("maxThreadsPerBlock", &maxThreadsPerBlock_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("anzLights", &anzLights_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("maximumCutSize", &maximumCutSize_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("lightSpaceMatrices_size", &lightSpaceMatrices_size_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("ism_size", &ism_size_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("ism_atlas_size", &ism_atlas_size_cpu, sizeof(int), 0, cudaMemcpyHostToDevice));

	//Float params:
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("sceneUnitsToMeters", &sceneUnitsToMeters_cpu, sizeof(float), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("nearZ", &nearZ_cpu, sizeof(float), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("farZ", &farZ_cpu, sizeof(float), 0, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol("sumOfWeightedIntensities", &sumOfWeightedIntensities_cpu, sizeof(float), 0, cudaMemcpyHostToDevice));
}

extern "C" void freeKernelParams()
{
	//CUDA_SAFE_CALL(cudaFree(offset));
	//CUDA_SAFE_CALL(cudaFree(maxThreadsPerBlock));
	//CUDA_SAFE_CALL(cudaFree(width));
	//CUDA_SAFE_CALL(cudaFree(height));
	//CUDA_SAFE_CALL(cudaFree(anzLights));
	//CUDA_SAFE_CALL(cudaFree(maximumCutSize));
	//CUDA_SAFE_CALL(cudaFree(lightSpaceMatrices_size));
	//CUDA_SAFE_CALL(cudaFree(ism_size));
	//CUDA_SAFE_CALL(cudaFree(ism_atlas_size));
	//CUDA_SAFE_CALL(cudaFree(sumOfWeightedIntensities));
	//CUDA_SAFE_CALL(cudaFree(nearZ));
	//CUDA_SAFE_CALL(cudaFree(farZ));
	//CUDA_SAFE_CALL(cudaFree(sceneUnitsToMeters));
}

extern "C" void allocateTextureMemory(float4* positionsInWorldCPU, float4* normalsInWorldCPU, int width, int height)
{
	CUDA_SAFE_CALL(cudaMallocArray (&cu_array1, &positionsInWorldGPU.channelDesc, width, height));
	CUDA_SAFE_CALL(cudaMallocArray (&cu_array2, &normalsInWorldGPU.channelDesc, width, height));

	// Bind the array to the texture
	CUDA_SAFE_CALL(cudaBindTextureToArray(positionsInWorldGPU, cu_array1));
	CUDA_SAFE_CALL(cudaBindTextureToArray(normalsInWorldGPU, cu_array2));

    // set texture parameters
//    positionsInWorldGPU.addressMode[0] = cudaAddressModeClamp;
//    positionsInWorldGPU.addressMode[1] = cudaAddressModeClamp;
    positionsInWorldGPU.filterMode = cudaFilterModePoint; //nearest neighbour
    positionsInWorldGPU.normalized = 0;

//	normalsInWorldGPU.addressMode[0] = cudaAddressModeClamp;
//    normalsInWorldGPU.addressMode[1] = cudaAddressModeClamp;
    normalsInWorldGPU.filterMode = cudaFilterModePoint; //nearest neighbour
    normalsInWorldGPU.normalized = 0;

	CUDA_SAFE_CALL(cudaMemcpyToArray(cu_array2, 0, 0, normalsInWorldCPU, width * height * sizeof(float4), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpyToArray(cu_array1, 0, 0, positionsInWorldCPU, width * height * sizeof(float4), cudaMemcpyHostToDevice));
}

extern "C" void deallocateTextureMemory()
{
	CUDA_SAFE_CALL(cudaFreeArray(cu_array1));
	CUDA_SAFE_CALL(cudaFreeArray(cu_array2));
}

extern "C" void shadeAllPixels(dim3 grid, dim3 block, float4* colorTextureGPU, float4* cutSizeTextureGPU, float4* positionsInWorldCPU, float4* normalsInWorldCPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* lightCutsHeapGPU, float4* colorsInWorldGPU)
{
	shadeLightcutsWithCUDA<<<grid,block>>>(colorTextureGPU, cutSizeTextureGPU, clusterPointersGPU, vplPointersGPU, lightCutsHeapGPU, colorsInWorldGPU);		
}

extern "C" void radiosityWithCUDA(dim3 grid, dim3 block, float4* colorTextureGPU, float4* positionsInWorldCPU, float4* normalsInWorldCPU, ClusterInCUDA clusterPointersGPU, VPL_InCUDA vplPointersGPU, float4* colorsInWorldGPU)
{
	radiosityWithCUDA<<<grid, block>>>(colorTextureGPU, clusterPointersGPU, vplPointersGPU, colorsInWorldGPU);

}
