#ifndef _PULL_PUSH_CU
#define _PULL_PUSH_CU


#include <stdio.h>
#include <math.h>
#include "CUDA_Structs.h"


__global__ void pull_Kernel(int _ism_size, int numberOfLights, int textureW, int textureH, float* dataToReadFrom, float* dataToWriteTo, int numberOfAveragesOneThreadHasToCompute, int requiredNumberOfThreadsPerBlock, int level)
{
	int ism_size = _ism_size / int(pow(2.0f, float(level-1))); // size of dataToReadFrom
	int numberOfLight = textureW * blockIdx.y + blockIdx.x;

	/*
	if (numberOfLight == 0 && threadIdx.x == 0)
	{
		#ifdef __DEVICE_EMULATION__
		
		printf("IN CUDA:\n");
			for (int i = 0; i < ism_size * ism_size * textureW * textureH; i++)
			{
				if (dataToReadFrom[i] < 0.0f)
					printf("%f\n", dataToReadFrom[i]);
				
				
				printf("%f  ", dataToReadFrom[i]);
				if ( i% (ism_size * textureW) == ((ism_size * textureW)-1))
				{
					printf("\n");
				}
				
			}
			
		#endif
	}
	*/

	if (numberOfLight >= numberOfLights) 
	{
		return;
	}

	int startingIndex = (numberOfLight / textureW) * ism_size * ism_size * textureW + (numberOfLight % textureW) * ism_size; //ganzzahldivision, nicht wegkuerzen!!

	int offsetInY = int(sqrt(float(blockDim.x)));//number of threads in one row of a block
	offsetInY = (threadIdx.x / offsetInY) * 2 * int(sqrt(float(numberOfAveragesOneThreadHasToCompute))) * textureW * ism_size;
//	int offsetInX = (threadIdx.x % int(sqrt(float(blockDim.x)))) * 2 * numberOfAveragesOneThreadHasToCompute;
	int offsetInX = (threadIdx.x % int(sqrt(float(blockDim.x)))) * 2 * int(sqrt(float(numberOfAveragesOneThreadHasToCompute)));

	#ifdef __DEVICE_EMULATION__
	//	printf("numberOfLight = %i\n", numberOfLight);	
	//	printf("startingIndex = %i\n", startingIndex);
		//printf("numberOfThreadsPerRow = %i\n", numberOfThreadsPerRow);
		//printf("blockIdx.x = %i\n", blockIdx.x);
		//printf("blockIdx.y = %i\n", blockIdx.y);
		//printf("threadIdx.x = %i\n", threadIdx.x);
		//printf("offsetInX = %i\n", offsetInX);
		//printf("offsetInY = %i\n\n", offsetInY);
	//	printf("blockDim.x = %i\n\n", blockDim.x);
		//printf("index4 = %i\n\n", index4);
	#endif

int numberOfThreadsPerRow = int(sqrt(float(blockDim.x)));
	startingIndex = startingIndex + offsetInY;
	startingIndex = startingIndex + offsetInX;





	int maximumIndex = (textureW * textureH * ism_size * ism_size) - 1;//-1 da das sonst die size anzeigt und nicht den maximalen Index

	if (startingIndex > maximumIndex)
	{
		#ifdef __DEVICE_EMULATION__
			printf("ERROR calculating indices, STARTindex out of range!!! (maxInde = %i)\n", maximumIndex);
			printf("numberOfLight = %i\n", numberOfLight);
			printf("offsetInX = %i\n", offsetInX);
			printf("offsetInY = %i\n", offsetInY);
			printf("threadIdx.x = %i\n\n", threadIdx.x);
		#endif
	}

	int sqrtOfNumberAveragesPerThread = int(sqrt(float(numberOfAveragesOneThreadHasToCompute)));
	for (int i = 0; i < sqrtOfNumberAveragesPerThread; i++)
	{
		for (int j = 0; j < sqrtOfNumberAveragesPerThread; j++)
		{
			int index0 = startingIndex + (j*2) + i * 2 * ism_size * textureW + 0;
			int index1 = startingIndex + (j*2) + i * 2 * ism_size * textureW + 1;
			int index2 = startingIndex + ism_size * textureW + (j*2) + i * 2 * ism_size * textureW + 0;
			int index3 = startingIndex + ism_size * textureW + (j*2) + i * 2 * ism_size * textureW + 1;

			int cappedVal = (index0 / (ism_size * textureW));
			int offsetInY_new = cappedVal/2 * (((_ism_size * textureW)/ (int(pow(2.0f, float(level)))))); //ganzzahldivision, nicht kuerzen!
			int offsetInX_new = (index0 - (cappedVal * ism_size * textureW)) / 2;
			int averageIndex = offsetInY_new + offsetInX_new;

	
			if (index0 > maximumIndex || index1 > maximumIndex || index2 > maximumIndex || index3 > maximumIndex)
			{
				#ifdef __DEVICE_EMULATION__
					printf("ERROR calculating indices, index out of range!!! (maxInde = %i)\n", maximumIndex);
					printf("index0 = %i\n", index0);
					printf("index1 = %i\n", index1);
					printf("index2 = %i\n", index2);
					printf("index3 = %i\n\n", index3);
				#endif
			}

			float val0 = dataToReadFrom[index0];
			float val1 = dataToReadFrom[index1];
			float val2 = dataToReadFrom[index2];
			float val3 = dataToReadFrom[index3];

			int isValid0 = int(val0<1.0f);
			int isValid1 = int(val1<1.0f);
			int isValid2 = int(val2<1.0f);
			int isValid3 = int(val3<1.0f);
/*
			#ifdef __DEVICE_EMULATION__
				printf("averageIndex = %i\n", averageIndex);
				printf("index = %i, val0 = %f, isValid = %i\n", index0, val0, isValid0);
				printf("index = %i, val1 = %f, isValid = %i\n", index1, val1, isValid1);
				printf("index = %i, val2 = %f, isValid = %i\n", index2, val2, isValid2);
				printf("index = %i, val3 = %f, isValid = %i\n\n", index3, val3, isValid3);		
			#endif
*/
			int numOfValids = isValid0 + isValid1 + isValid2 + isValid3;
			

			if (numOfValids == 0) // if there are no valid values, we are done: pull the invalid value up:
			{
				#ifdef __DEVICE_EMULATION__
				//	printf("all 4 values are invalid, pulling invalid value...\n\n");
				#endif
				dataToWriteTo[averageIndex] = 1.0f;
				continue;
			}

			float thresh = 0.05f*pow(2.0f, level);

			float minZval = 1.0f;
			//ganz schlechter programmierstil, aber leider kann man scheinbar bei float4's nicht mit [] arbeiten...

			minZval = min(minZval, val0);

			minZval = min(minZval, val1);

			minZval = min(minZval, val2);

			minZval = min(minZval, val3);

			//####################################

			if (isValid0>0)
			{
				if (val0 > minZval + thresh)
				{
					isValid0=0;
				}
			}

			if (isValid1>0)
			{
				if (val1 > minZval + thresh)
				{
					isValid1=0;
				}
			}

			if (isValid2>0)
			{
				if (val2 > minZval + thresh)
				{
					isValid2=0;
				}
			}

			if (isValid3>0)
			{
				if (val3 > minZval + thresh)
				{
					isValid0=0;
				}
			}			
			
			float weight; 

			float average;
			if (numOfValids == 0)
			{
				average = 1.0f;
			}
			else
			{
				weight = 1.0f / float(numOfValids);
				average = (isValid0 * weight * val0 + isValid1 * weight * val1 + isValid2 * weight * val2 + isValid3 * weight * val3);
			}

			dataToWriteTo[averageIndex] = average;
			
			/*
			#ifdef __DEVICE_EMULATION__
				printf("index = %i, val0 = %f, isValid = %i\n", index0, val0, isValid0);
				printf("index = %i, val1 = %f, isValid = %i\n", index1, val1, isValid1);
				printf("index = %i, val2 = %f, isValid = %i\n", index2, val2, isValid2);
				printf("index = %i, val3 = %f, isValid = %i\n", index3, val3, isValid3);
				printf("weight = %f\n", weight);
				printf("averageIndex = %i\n", averageIndex);
				printf("average = %f; minZval = %f;\n\n", average, minZval);
			#endif
			*/
			
			


		}
	}
}

__global__ void push_Kernel(int _ism_size, int numberOfLights, int textureW, int textureH, float* dataToReadFrom, float* dataToWriteTo, int numberOfAveragesOneThreadHasToPushDown, int requiredNumberOfThreadsPerBlock, int level)
{
	int ism_size = _ism_size / int(pow(2.0f, float(level))); //die size des readFrom arrays
	int numberOfLight = textureW * blockIdx.y + blockIdx.x;

	/*
	if (numberOfLight == 0 && threadIdx.x == 0)
	{
		#ifdef __DEVICE_EMULATION__
		
		printf("IN CUDA:\n");
			for (int i = 0; i < ism_size * ism_size * textureW * textureH; i++)
			{
				printf("%f  ", dataToReadFrom[i]);
				if ( i% (ism_size * textureW) == ((ism_size * textureW)-1))
				{
					printf("\n");
				}
			}
			
		#endif
	}
	*/	

	if (numberOfLight >= numberOfLights) 
	{
		return;
	}

	int ismSizeBelow = _ism_size / int(pow(2.0f, float(level-1)));

	int startingIndex = (numberOfLight / textureW) * ismSizeBelow * ismSizeBelow * textureW + (numberOfLight % textureW) * ismSizeBelow; //ganzzahldivision, nicht wegkuerzen!!
	int offsetInY = int(sqrt(float(blockDim.x)));//number of threads in one row of a block
	offsetInY = (threadIdx.x / offsetInY) * 2 * int(sqrt(float(numberOfAveragesOneThreadHasToPushDown))) * textureW * ismSizeBelow;
//	int offsetInX = (threadIdx.x % int(sqrt(float(blockDim.x)))) * 2 * numberOfAveragesOneThreadHasToPushDown;
	int offsetInX = (threadIdx.x % int(sqrt(float(blockDim.x)))) * 2 * int(sqrt(float(numberOfAveragesOneThreadHasToPushDown)));

int numberOfThreadsPerRow = int(sqrt(float(blockDim.x)));
	startingIndex = startingIndex + offsetInY;
	startingIndex = startingIndex + offsetInX;

	#ifdef __DEVICE_EMULATION__
		//printf("numberOfLight = %i\n", numberOfLight);	
		//printf("startingIndex = %i\n", startingIndex);
		//printf("numberOfThreadsPerRow = %i\n", numberOfThreadsPerRow);
		//printf("blockIdx.x = %i\n", blockIdx.x);
		//printf("blockIdx.y = %i\n", blockIdx.y);
		//printf("threadIdx.x = %i\n", threadIdx.x);
		//printf("offsetInX = %i\n", offsetInX);
		//printf("offsetInY = %i\n\n", offsetInY);
	//	printf("blockDim.x = %i\n\n", blockDim.x);
		//printf("index4 = %i\n\n", index4);
	#endif


	
	int maximumIndex = (textureW * textureH * ismSizeBelow * ismSizeBelow) - 1;//-1 da das sonst die size anzeigt und nicht den maximalen Index

	if (startingIndex > maximumIndex)
	{
		#ifdef __DEVICE_EMULATION__
			printf("ERROR calculating indices, STARTindex out of range!!! (maxInde = %i)\n", maximumIndex);
			printf("numberOfLight = %i\n", numberOfLight);
			printf("offsetInX = %i\n", offsetInX);
			printf("offsetInY = %i\n", offsetInY);
			printf("threadIdx.x = %i\n\n", threadIdx.x);
		#endif
	}

	int sqrtOfNumberAveragesPerThread = int(sqrt(float(numberOfAveragesOneThreadHasToPushDown)));
	for (int i = 0; i < sqrtOfNumberAveragesPerThread; i++)
	{
		for (int j = 0; j < sqrtOfNumberAveragesPerThread; j++)
		{
			int index0 = startingIndex + (j*2) + i * 2 * ismSizeBelow * textureW + 0;
			int index1 = startingIndex + (j*2) + i * 2 * ismSizeBelow * textureW + 1;
			int index2 = startingIndex + ismSizeBelow * textureW + (j*2) + i * 2 * ismSizeBelow * textureW + 0;
			int index3 = startingIndex + ismSizeBelow * textureW + (j*2) + i * 2 * ismSizeBelow * textureW + 1;

			int cappedVal = (index0 / (ismSizeBelow * textureW));
			int offsetInY_new = cappedVal/2 * ism_size * textureW; //ganzzahldivision, nicht kuerzen!
			int offsetInX_new = (index0 - (cappedVal * ismSizeBelow * textureW)) / 2;
			int averageIndex = offsetInY_new + offsetInX_new;

	
			if (index0 > maximumIndex || index1 > maximumIndex || index2 > maximumIndex || index3 > maximumIndex)
			{
				#ifdef __DEVICE_EMULATION__
					printf("ERROR calculating indices, index out of range!!! (maxInde = %i)\n", maximumIndex);
					printf("index0 = %i\n", index0);
					printf("index1 = %i\n", index1);
					printf("index2 = %i\n", index2);
					printf("index3 = %i\n\n", index3);
				#endif
			}

			

			

			float val0 = dataToWriteTo[index0];
			float val1 = dataToWriteTo[index1];
			float val2 = dataToWriteTo[index2];
			float val3 = dataToWriteTo[index3];

			float average = dataToReadFrom[averageIndex];

			/*
			#ifdef __DEVICE_EMULATION__
				printf("averageIndex = %i\n", averageIndex);
				printf("cappedVal = %i\n", cappedVal);
				printf("offsetInX_new = %i\n", offsetInX_new);
				printf("offsetInY_new = %i\n", offsetInY_new);
				printf("ism_size = %i\n", ism_size);
				printf("ismSizeBelow = %i\n", ismSizeBelow);
				printf("index = %i, val0 = %f\n", index0, val0);
				printf("index = %i, val1 = %f\n", index1, val1);
				printf("index = %i, val2 = %f\n", index2, val2);
				printf("index = %i, val3 = %f\n\n", index3, val3);	
			#endif
			*/

		/*
			#ifdef __DEVICE_EMULATION__
				printf("averageIndex = %i\n", averageIndex);
				printf("averageValue = %f\n", average);
				printf("index = %i, val0 = %f\n", index0, val0);
				printf("index = %i, val1 = %f\n", index1, val1);
				printf("index = %i, val2 = %f\n", index2, val2);
				printf("index = %i, val3 = %f\n\n", index3, val3);		
			#endif
		*/	


			float thresh = 0.05f*pow(2.0f, level-1);

			//####################################

			
//			if (val0 > average + thresh)
			if (val0 == 1.0f)
			{
				dataToWriteTo[index0] = average;
			}

//			if (val1 > average + thresh)
			if (val1 == 1.0f)
			{
				dataToWriteTo[index1] = average;
			}

//			if (val2 > average + thresh)
			if (val2 == 1.0f)
			{
				dataToWriteTo[index2] = average;
			}
					
//			if (val3 > average + thresh)
			if (val3 == 1.0f)
			{
				dataToWriteTo[index3] = average;
			}

			/*
			#ifdef __DEVICE_EMULATION__
				printf("index = %i, val0 = %f, isValid = %i\n", index0, val0, isValid0);
				printf("index = %i, val1 = %f, isValid = %i\n", index1, val1, isValid1);
				printf("index = %i, val2 = %f, isValid = %i\n", index2, val2, isValid2);
				printf("index = %i, val3 = %f, isValid = %i\n", index3, val3, isValid3);
				printf("weight = %f\n", weight);
				printf("average = %f; minZval = %f;\n\n", average, minZval);
			#endif
			*/

		}
	}
}

/**
 * @textureW, textureH: encode, how many ISMs are per row/column!
 *
 */
extern "C" void launch_pull_kernel(int ism_size, int numberOfLights, int textureW, int textureH, float* dataToReadFrom, float* dataToWriteTo, int numberOfAveragesOneThreadHasToCompute, int requiredNumberOfThreadsPerBlock, int level)
{
	dim3 block;
	dim3 grid;

	grid.x = textureW; //these two values will be const. during the pushpull algorithm!
	grid.y = textureH;
	grid.z = 1;

	block.x = requiredNumberOfThreadsPerBlock;
	block.y = 1;
	block.z = 1;
	//printf("CUDA: pulling level %i...", level);
	pull_Kernel<<<grid, block>>>(ism_size, numberOfLights, textureW, textureH, dataToReadFrom, dataToWriteTo, numberOfAveragesOneThreadHasToCompute, requiredNumberOfThreadsPerBlock, level);
	//printf("Done.\n");
}

/**
 * @textureW, textureH: encode, how many ISMs are per row/column!
 *
 */
extern "C" void launch_push_kernel(int ism_size, int numberOfLights, int textureW, int textureH, float* dataToReadFrom, float* dataToWriteTo, int numberOfAveragesOneThreadHasToPushDown, int requiredNumberOfThreadsPerBlock, int level)
{
	dim3 block;
	dim3 grid;

	grid.x = textureW; //these two values will be const. during the pushpull algorithm!
	grid.y = textureH;
	grid.z = 1;

	block.x = requiredNumberOfThreadsPerBlock;
	block.y = 1;
	block.z = 1;

	//printf("CUDA: pushing level %i...", level);
	
	push_Kernel<<<grid, block>>>(ism_size, numberOfLights, textureW, textureH, dataToReadFrom, dataToWriteTo, numberOfAveragesOneThreadHasToPushDown, requiredNumberOfThreadsPerBlock, level);
		
	//printf("Done.\n");
}

#endif // #ifndef _PUSH_PULL_CU
