#include "Camera.h"



ofstream out;
ifstream fin;
int startTime;

struct point
{
	int index;
	float pos[3];
	float rot[2];
	int event;
}; 

vector<point> allPoints;
int curIndex = 3;	//points to last of the 4 points used for spline-interpolation
float t=0.0f;		//curve-parameter

Camera::Camera()
{
	currentFPS = 0;
	theoretFPS = 0;
	lastX=0; lastY=0;
	posX=35.3345f; posY=138.221f; posZ=658.754f; rotX=4.0f; rotY=-0.2f;

	recording = false;
	playback = false;
}

/**
* Calc view vector and return it
*/
Vector Camera::calcViewVector(void)
{
	Vector view;

	float rotXrad, rotYrad;
	rotXrad = (rotX / 180 * 3.141592654f);
	rotYrad = (rotY / 180 * 3.141592654f);

	view.x = sin(rotYrad);
	view.y = -sin(rotXrad);
	view.z = -cos(rotYrad);

	view.normalize();

	return view;
}


Point Camera::getPosition(void)
{
	Point pos;
	pos.x = posX;
	pos.y = posY;
	pos.z = posZ;
	return pos;
}

void Camera::view()
{
	if (playback)
	{
		t = t+(min ((1/currentFPS),(1-t)));
		if (t>=1.0f)
		{
			t = 0.0f;
			curIndex++;
		}
		if (curIndex>int(allPoints.size()-1))
		{
			printf("Playback ended!\n");
			glRotatef(rotX,1.0,0.0,0.0);
			glRotatef(rotY,0.0,1.0,0.0);
			glTranslated(-posX,-posY,-posZ); 
			playback = false;
			return;
		}

		point p0 = allPoints.at(curIndex-3);
		point p1 = allPoints.at(curIndex-2);
		point p2 = allPoints.at(curIndex-1);
		point p3 = allPoints.at(curIndex);

		switch (p2.event)
		{
		case 0:
			{
				break;
			}
		case 1:
			{
				break;
			}
		case 2:
			{
				break;
			}
		case 3:
			{
				break;
			}
		case 4:
			{
				break;
			}
		}

		point result;
		for (int i = 0; i<3; i++)
		{
			result.pos[i] = 0.5f*((2*p1.pos[i]) + t*(-p0.pos[i]+p2.pos[i]) +t*t*(2*p0.pos[i]-5*p1.pos[i]+4*p2.pos[i]-p3.pos[i]) +t*t*t*(-p0.pos[i]+3*p1.pos[i]-3*p2.pos[i]+p3.pos[i])  );
		}

		for (int i = 0; i<2; i++)
		{
			result.rot[i] = (1-t)*p1.rot[i] + (t)*p2.rot[i];
		}

		posX = result.pos[0];
		posY = result.pos[1];
		posZ = result.pos[2];

		rotX = result.rot[0];
		rotY = result.rot[1];


	}

	glRotatef(rotX,1.0,0.0,0.0);
	glRotatef(rotY,0.0,1.0,0.0);
	glTranslated(-posX,-posY,-posZ); //translate the screen to the position of our camera
}

void Camera::saveSnapShot()
{
	char separator[] = ";";
	char newLine[] = "\n";
	char space[] = " ";
	char pX[32];char pY[32];char pZ[32];char rX[32];char rY[32];char curTime[32];
	sprintf_s(curTime, "%i", startTime++);
	sprintf_s(pX,"%f",posX);
	sprintf_s(pY,"%f",posY);
	sprintf_s(pZ,"%f",posZ);
	sprintf_s(rX,"%f",rotX);
	sprintf_s(rY,"%f",rotY);

	out.write(curTime, strlen(curTime));
	out.write(separator, strlen(separator));

	out.write(pX, strlen(pX));
	out.write(space, strlen(space));
	out.write(pY, strlen(pY));
	out.write(space, strlen(space));
	out.write(pZ, strlen(pZ));
	out.write(separator, strlen(separator));

	out.write(rX, strlen(rX));
	out.write(space, strlen(space));
	out.write(rY, strlen(rY));
	out.write(separator, strlen(separator));
	out.write(newLine, strlen(newLine));
}

void Camera::stopRecording()
{
	if (!recording) return;
	printf("Stopping Recording...");
	recording = false;
	out.close();
	printf("Done.\n");
}

void Camera::startRecording()
{
	if (recording) return;
	printf("Starting recording...");
	
	srand ( time(NULL) );
	int i = rand()*10000 + rand();
	char fileSuffix[32];
	sprintf_s(fileSuffix, "%i.txt", i);

	char fileName[] = "Resources\\kamerafahrt";
	strcat_s(fileName, fileSuffix);
	out.open("Resources\\kamerafahrt.txt");
	if(!out.good()) 
	{
	  printf("Cannot open output file.\n");
	  return;
	}
	
	printf("Now recording!\n");
	recording = true;

	startTime = 0;
	  
 
}

bool Camera::getRecording()
{
	return recording;
}
bool Camera::getPlayback()
{
	return playback;
}

void Camera::stopPlayRecording()
{
	playback = false;
}

void Camera::playRecording(char* myRecordingFile)
{
	allPoints.clear();

	fin.open(myRecordingFile);
	
	if (!fin.good())
	{
		printf("error opening %s\n", myRecordingFile);
		return;
	}

	
	int i;

	while (!fin.eof())
	{
		char times[256];
		fin.getline(times, 256, ';');

		char positions[256];
		fin.getline(positions, 256, ';');

		char rotations[256];
		fin.getline(rotations, 256, ';');

		char events[256];
		fin.getline(events, 256, '\n');

		point a;

		a.index = atoi (times);
		
		char* next_token1;

		char* pos  = strtok_s( positions, " ", &next_token1);
		i = 0;

		point tmp;
		
		
		while (pos != NULL)
		{
			tmp.pos[i++] = float(atoi(pos));
			pos = strtok_s( NULL, " ", &next_token1);
		}
		
		char* rot  = strtok_s( rotations, " ", &next_token1);
		i = 0;
		
		while (rot != NULL)
		{
			tmp.rot[i++] = float(atoi(rot));
			rot = strtok_s( NULL, " ", &next_token1);
		}
		if (strlen(events)>0)
			tmp.event = atoi(events);
		else
			tmp.event = 0;

		allPoints.push_back(tmp);

	}
	fin.seekg(0); //reset!!
	fin.close();
	curIndex = 3;
	if(allPoints.size()<4)
	{
		printf("Not enough points - can't play playback!\n");
		return;
	}

	printf("Playback started!\n");
	playback = true;
}