#include "GameStateVBO.h"


using namespace std;

int GameStateVBO::clusterVisualizationStep = 0;

int GameStateVBO::lastRenderingWidth = 0;
int GameStateVBO::lastRenderingHeight = 0;

float GameStateVBO::farZ = 1000.0f;
float GameStateVBO::nearZ = 0.1f;

bool wPressed, sPressed, aPressed, dPressed;
bool poly = true;

// constants / global variables
unsigned int window_width = 512;
unsigned int window_height = 512;
unsigned int image_width = 512;
unsigned int image_height = 512;

unsigned int size_tex_data;
unsigned int num_texels;
unsigned int num_values;
int blur_radius = 0;

// pbo and fbo variables
GLuint pbo_source;
GLuint pbo_dest;
GLuint fbo_source;
GLuint tex_screen;

// (offscreen) render target
// fbo variables
GLuint GameStateVBO::framebuffer;
GLuint GameStateVBO::colorbufferPositions;
GLuint GameStateVBO::colorbufferNormals;
GLuint GameStateVBO::colorbufferCsize;
GLuint GameStateVBO::depthbuffer;
GLuint GameStateVBO::colorbufferTextureColors;

// imperfect shadow map
// fbo variables
GLuint tex_ism;
GLuint ism_depth_buffer;

float* white=0;

GLuint GameStateVBO::ism_matrices;

void createFBO(GLuint* fbo, GLuint* tex, unsigned int size_x, unsigned int size_y);
void deleteFBO(GLuint* fbo);


void visualizeLights(void);
void processImage(void);
void renderTV(void);
void renderPointRepresentationOfScene(bool shaderAlreadyBound, Vector viewVec, Vector viewPos);
void renderImperfectShadowMaps(void);

void pullPushTexture(float* p, int sizeX, int maxLevel);
float calcAverage(float values[4], float thresh);
void calcReplacement(bool mask[4], float average, float values[4], float thresh);

//toggle render modes
bool renderPoints = false;

/*
extern "C" void launch_cudaProcess(dim3 grid, dim3 block, int sbytes, int* g_data, int* g_odata, 
								   int imgw, int imgh, int tilew, 
								   int radius, float threshold, float highlight);
*/
////////////////////////////////////////////

void GameStateVBO::destroyGameState_static(void)
{
	ModelManager::deleteAllModels();

	glDeleteTextures(1, &CUDA_LightCuts::colorTextureGL);
	glDeleteTextures(1, &CUDA_LightCuts::cutSizeTextureGL);
	glDeleteTextures(1, &colorbufferPositions);
	glDeleteTextures(1, &GameStateVBO::colorbufferNormals);
	glDeleteTextures(1, &GameStateVBO::colorbufferTextureColors);
	glDeleteRenderbuffersEXT(1, &GameStateVBO::depthbuffer);
	glDeleteRenderbuffersEXT(1, &GameStateVBO::framebuffer);

	TextureManager::deleteAllTextures();
	

	if(shadowMap!=0)
	{
		delete[] shadowMap;
		shadowMap = 0;
	}

	if(lightSpaceMatrices!=0)
	{
		delete[] lightSpaceMatrices;
		lightSpaceMatrices = 0;
	}

	if(CUDA_LightCuts::lightMatricesCUDA!=0)
	{
		delete[] CUDA_LightCuts::lightMatricesCUDA;
		CUDA_LightCuts::lightMatricesCUDA = 0;
	}

	if (CUDA_LightCuts::colorTexture!=0)
	{
		delete[] CUDA_LightCuts::colorTexture;
		CUDA_LightCuts::colorTexture = 0;
	}

	if (CUDA_LightCuts::cutSizeTexture!=0)
	{
		delete[] CUDA_LightCuts::cutSizeTexture;
		CUDA_LightCuts::cutSizeTexture = 0;
	}

	CUDA_LightCuts::deleteCudaArrays();
	lightChoice->clear();
}

void GameStateVBO::destroyGameState(void)
{
	delete camera;
	camera = NULL;
	ModelManager::deleteAllModels();

	TextureManager::deleteAllTextures();
	
	ShaderManager::deleteAllShaders();
/*
	deletePBO(&pbo_source);
	deletePBO(&pbo_dest);
*/

	//cudaThreadExit();

	if(shadowMap!=0)
	{
		delete[] shadowMap;
		shadowMap = 0;
	}

	if(lightSpaceMatrices!=0)
	{
		delete[] lightSpaceMatrices;
		lightSpaceMatrices = 0;
	}

	if(CUDA_LightCuts::lightMatricesCUDA!=0)
	{
		delete[] CUDA_LightCuts::lightMatricesCUDA;
		CUDA_LightCuts::lightMatricesCUDA = 0;
	}
	
	delete[] white;
		white=0;

	if (CUDA_LightCuts::colorTexture!=0)
	{
		delete[] CUDA_LightCuts::colorTexture;
		CUDA_LightCuts::colorTexture = 0;
	}

	if (CUDA_LightCuts::cutSizeTexture!=0)
	{
		delete[] CUDA_LightCuts::cutSizeTexture;
		CUDA_LightCuts::cutSizeTexture = 0;
	}

	CUDA_LightCuts::deleteCudaArrays();
	Scene::destroyScene();
}

void GameStateVBO::initGameState(vector<vector<string> >& params)
{
	CUDA_PushPull::initCuda();

	time_t t;
	t = time(NULL);
//	srand((unsigned int)t);

	lastX = 0;
	lastY = 0;
	setDefaultParameterValues();

	setParamsAccordingToFileParams(params);

	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	glClearColor (0.0,0.0,0.0,1.0);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	loadTextures();
	//syntax shadersToLoad: "name", "vertShaderSourceFile", "pixShaderSourceFile"			
	char* shadersToLoad[] = {"skybox", "skyboxShader.ver", "skyboxShader.pix",			
		"manyLights", "manyLights.ver", "manyLights.pix",
		"writeGeometry", "writeGeometry.ver", "writeGeometry.pix",
		"points", "points.ver", "points.pix",
		"pointsParabol", "pointsParabol.ver", "pointsParabol.pix",
		"writeGeometryParabol", "writeGeometryParabol.ver", "writeGeometryParabol.pix",
		"wireframe", "wireframe.ver", "wireframe.pix"
	};
	int numOfShaders = 7;

	ShaderManager::loadAllShaders(shadersToLoad, numOfShaders);

	//SCENE LOADING
	Scene::readSceneFromFile(string("Resources\\Scenes\\cornellbox.txt"));

	ModelManager::loadAllModels(Scene::modelsToLoad);
	TextureManager::generateLightTextures();
	GLfloat diffuse2[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat specular2[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat ambient2[] = {1.0, 1.0, 1.0, 1.0};
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse2);
	glLightfv(GL_LIGHT1, GL_SPECULAR, specular2);
	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient2);

	renderMode = 0;
	printDebugOutput = false;
	renderImageToDisk(2,2);
	render();
	renderImageToDisk(2,2);
	printDebugOutput = true;

	
}


void GameStateVBO::processKey(unsigned char& key, int& x, int& y)
{
	if (sceneIsBeingLoaded)
		return;
	int speed=10;

	switch(key)
	{
	case 13: 
		{
			camera->posX = 35.3345f;
			camera->posY = 138.221f;
			camera->posZ = 658.754f;
			camera->rotX = 4.0f;
			camera->rotY = -0.2f;
		//	cout << "CAM: " << camera->posX << " " << camera->posY << " "  << camera->posZ << " " << camera->rotX << " " << camera->rotY << endl;
			break;
		}
	case 'q': case 'Q':
		{
			camera->rotX += 1;
			if (camera->rotX >360) camera->rotX -= 360;
			break;
		}
	case 'e': case 'E':
		{
			camera->rotX -= 1;
			if (camera->rotX < -360) camera->rotX += 360;
			break;
		}

	case 'w':
		{
			wPressed = true; 
			break;
		}

	case's':
		{
			sPressed = true; 
			break;
		}
	case 'd':
		{
			dPressed = true; 
			break;
		}
	case 'a':
		{
			aPressed = true; 
			break;
		}

	case 'p':
		{
			renderPoints = !renderPoints;
			break;
		}

	case 'u':
	{
		shadowPointsOnly = !shadowPointsOnly;
		glClearColor (1.0, 1.0, 1.0, 1.0);
		renderImperfectShadowMaps();

		glUseProgram(ShaderManager::fetchShader("manyLights"));
		ModelManager::setShaderParams("manyLights");
		
		glUseProgram(ShaderManager::fetchShader(writeGeometryShader));
		ModelManager::setShaderParams(writeGeometryShader);

		glUseProgram(0);
		break;
	}

	case 'r': case 'R':
		{
			if (!(camera->getRecording()))
			{
				camera->startRecording();
			}
			else
			{
				camera->stopRecording();
			}
			break;
		}

	case 'l':
		  {
			  camera->lastPosX = camera->posX;
			  camera->lastPosY = camera->posY;
			  camera->lastPosZ = camera->posZ;
			  camera->lastRotX = camera->rotX;
			  camera->lastRotY = camera->rotY;
			  camera->posX = 5.35;
			  camera->posY = 53.34;
			  camera->posZ = 10.09;
			  camera->rotX = 0.0;
			  camera->rotY = 0.0;
			break;
		  }
	case 'L':
		  {
			  camera->posX = camera->lastPosX;
			  camera->posY = camera->lastPosY;
			  camera->posZ = camera->lastPosZ;
			  camera->rotX = camera->lastRotX;
			  camera->rotY = camera->lastRotY;
			break;
		  }

	case 'c': case 'C':
		{
			debugscreen++;
			break;
		}
	case 'v':
		
		{
			showLights=!showLights;
			break;
		}
	case 173: // -
		{
			clusterVisualizationStep--;
			clusterVisualizationStep = max(0.0f, min((TextureManager::anzLights*2.0f) - 1.0f, clusterVisualizationStep));
			break;
		}
	case 171: // +
		{
			clusterVisualizationStep++;
			clusterVisualizationStep = max(0.0f, min((TextureManager::anzLights*2.0f) - 1.0f, clusterVisualizationStep));
			break;
		}
	}
}

void GameStateVBO::processKeyUp(unsigned char& key, int& x, int& y)
{
	switch(key)
	{
		case 'w':
		{
			wPressed = false;
			break;
		}
		case 's':
		{
			sPressed = false;
			break;
		}
		case 'a':
		{
			aPressed = false;
			break;
		}
		case 'd':
		{
			dPressed = false;
			break;
		}
	}
}

void GameStateVBO::processSpecialKey(int& key, int& x, int& y)
{
	//printf("Special key pressed in GameStateVBO: %i\n", key);
}

void GameStateVBO::processSpecialKeyUp(int& key, int& x, int& y)
{
	//printf("Special key released in GameStateVBO: %i\n", key);
}

void GameStateVBO::processMouseClick(int& button, int& state, int& x, int& y)
{
	switch (button)
	{
	case FL_LEFT_MOUSE:
		{
			//prevent "jumping"
			lastX=x; //set lastx to the current x position
			lastY=y; //set lasty to the current y position
			break;
		}
	}

}

void GameStateVBO::processMousePassiveMovement(int& x, int& y)
{
}

void GameStateVBO::processMouseMovement(int& x, int& y)
{
	int diffx=x-lastX; //check the difference between the current x and the last x position
	int diffy=y-lastY; //check the difference between the current y and the last y position
	lastX=x; //set lastx to the current x position
	lastY=y; //set lasty to the current y position
	camera->rotX += ((float) diffy/5.0); //set the xrot to xrot with the addition of the difference in the y position
	camera->rotY += ((float) diffx/5.0); //set the xrot to yrot with the addition of the difference in the x position
}

/*RRResourceHandlers:*/
void GameStateVBO::render(void)
{	
	if (sceneIsBeingLoaded)
	{
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		return;
	}
		

	movePlayer();
	

//*** NOW RENDER REAL IMAGE ***//

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glViewport (0, 0, width, height);

	camera->view();

	if(renderPoints)
	{	
		glUseProgram(ShaderManager::fetchShader(pointsShader));
		int texLoc = glGetUniformLocation(ShaderManager::fetchShader(pointsShader), "lpos");
		Point camPosPoint = GameStateVBO::camera->getPosition();
		Vector viewVec = camera->calcViewVector();
		
		if (texLoc!=-1)
			glUniform4f(texLoc, camPosPoint.x, camPosPoint.y, camPosPoint.z, 1.0f);

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		renderPointRepresentationOfScene(false, viewVec, Vector(camPosPoint.x, camPosPoint.y, camPosPoint.z));
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
	}
	else
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);

		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glLineWidth(2.0f);
		ModelManager::renderAllWireframe();
		
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);

		if(showLights)
			visualizeLights();

		if (showClustering)
			visualizeClustering(clusterVisualizationStep);
		
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		renderTV();
	}
}


void GameStateVBO::update(void)
{
//	glutPostRedisplay();
}

void GameStateVBO::reshape(int w, int h)
{
	width = w;
	height = h;
	glViewport (0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (60, (GLfloat)w / (GLfloat)h, 1.0, 100000.0);
	glMatrixMode (GL_MODELVIEW);
}

/*
* Just for making the code better to read; attempts to retrieve parameter-values from the 
* given vector of parameter-tuples.
*/
void GameStateVBO::setParamsAccordingToFileParams(vector<vector<string> >& params)
{

	if(params.size()>0)
	{
		for (int i = 0; i<(int)params.size(); i++)
		{
			vector<string> temp = params.at(i);
			string name = temp.at(0);
			string value = temp.at(1);
			if (value.compare("")==0) continue; //empty value -> skip

			if (name.compare("WindowWidth")==0) { width = atoi(value.c_str());continue; }
			if (name.compare("WindowHeight")==0) { height = atoi(value.c_str());continue; }
			if (name.compare("FullScreen")==0) { fullScreen = atoi(value.c_str())>0;continue; }
			if (name.compare("pullPush")==0) { pullPush = atoi(value.c_str())>0;continue; }

			if (name.compare("pullPushMaxLevel")==0) { pullPushMaxLevel = atoi(value.c_str());continue; }
			if (name.compare("ism_size")==0) { ism_size = atoi(value.c_str());continue; }

			if (name.compare("clusterLightTree")==0) { clusterLightTree = atoi(value.c_str())>0;continue; }
	
			//atoi = convertToInteger; atof = convertToFloat
		}
	}


}

/*
* (Re)Sets all parameter-values to default values.
*/
void GameStateVBO::setDefaultParameterValues()
{
	width = 200;
	height = 200;
	fullScreen = false;
}



void GameStateVBO::glEnable2D()
{  
	int vPort[4];

	glGetIntegerv(GL_VIEWPORT, vPort);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(0, vPort[2], 0, vPort[3], -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
}

void GameStateVBO::glDisable2D()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();   
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();	
}


void GameStateVBO::movePlayer(void)
{
	float speed = 10.0f;
	if (wPressed) 
		{
			float rotXrad, rotYrad;
			rotXrad = (camera->rotX / 180 * 3.141592654f);
			rotYrad = (camera->rotY / 180 * 3.141592654f); 
			camera->posX += speed*float(sin(rotYrad)) ;
			camera->posY -= speed*float(sin(rotXrad)) ;
			camera->posZ -= speed*float(cos(rotYrad)) ;
		}

	if (sPressed)
		{
			float rotXrad, rotYrad;
			rotYrad = (camera->rotY / 180 * 3.141592654f);
			rotXrad = (camera->rotX / 180 * 3.141592654f); 
			camera->posX -= speed*float(sin(rotYrad));
			camera->posZ += speed*float(cos(rotYrad)) ;
			camera->posY += speed* float(sin(rotXrad));
		}
	if (dPressed)
		{
			float rotYrad;
			rotYrad = (camera->rotY / 180 * 3.141592654f);
			camera->posX += float(cos(rotYrad)) * speed;
			camera->posZ += float(sin(rotYrad)) * speed;
		}


	if (aPressed)
		{
			float rotYrad;
			rotYrad = (camera->rotY / 180 * 3.141592654f);
			camera->posX -= float(cos(rotYrad)) * speed;
			camera->posZ -= float(sin(rotYrad)) * speed;
		}
}

void GameStateVBO::generateISMTextures()
{
	//CREATE ISM (kleine texture)

	int ismW = ism_size;
	int ismH = ism_size;

	glDeleteTextures(1, &tex_ism);
	glGenTextures(1, &tex_ism);
	glBindTexture(GL_TEXTURE_2D, tex_ism);

	// set basic parameters
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	// buffer data
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, ismW, ismH, 0, GL_RGBA, GL_FLOAT, NULL);

	// create a renderbuffer
	glGenRenderbuffersEXT(1, &ism_depth_buffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, ism_depth_buffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, ismW, ismH);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

	// create and bind a framebuffer
	glGenFramebuffersEXT(1, &ism_framebuffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ism_framebuffer);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex_ism, 0);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, ism_depth_buffer);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	/*
	if (white != 0)
	{
		delete[] white;
		glDeleteTextures(1, &ism_atlas);
	}
	*/
//ISM TEX ATLAS (grosse texture)
	// create ism tex atlas
	glDeleteTextures(1, &ism_atlas);
	glGenTextures(1, &ism_atlas);
	glBindTexture(GL_TEXTURE_2D, ism_atlas);
	// set basic parameters
	GLfloat whiteC[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, whiteC);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	int ism_atlas_size_new = ceil(sqrt(float(TextureManager::anzLights))) * ism_size;
	ism_atlas_size = ism_atlas_size_new; //4096;
	//TODO das muss noch ausgebessert werden

	if (white != 0)
	{
		delete[] white;
		white = 0;
	}
	white = new float[ism_atlas_size*ism_atlas_size*4];

	for(int i=0; i<ism_atlas_size*ism_atlas_size*4; i++)
		white[i]=1.0f;
	
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, ism_atlas_size, ism_atlas_size, 0, GL_RGBA, GL_FLOAT, white);
	delete[] white;
	white = 0;

	lightSpaceMatrices_size = (ism_atlas_size/ism_size) * 2; //2*2 = 4 floatRGBAs pro Licht (=16 werte)

//ISM LIGHT MATRIX TEXTURE
	//Create ism light matrix texture
	glDeleteTextures(1, &ism_matrices);
	glGenTextures(1, &ism_matrices);
	glBindTexture(GL_TEXTURE_2D, ism_matrices);
	// set basic parameters
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, lightSpaceMatrices_size, lightSpaceMatrices_size, 0, GL_RGBA, GL_FLOAT, NULL);

	glClearColor (1.0f, 1.0f, 1.0f, 1.0f);

	renderImperfectShadowMaps();

	glUseProgram(ShaderManager::fetchShader("manyLights"));
	ModelManager::setShaderParams("manyLights");
	glUseProgram(ShaderManager::fetchShader("writeGeometry"));
	ModelManager::setShaderParams("writeGeometry");
	glUseProgram(ShaderManager::fetchShader("writeGeometryParabol"));
	ModelManager::setShaderParams("writeGeometryParabol");

	glUseProgram(0);
	glClearColor (0.0,0.0,0.0,1.0);

}




	/*
	float test[64];

	for(int i=0; i<4; i++)
	{
		for(int j=0; j<4; j++)
		{
			for(int c=0; c<4; c++)
			{
				if(i!=0 && j!=0 && j!=3 && i!=3)
					test[(i*4 + j)*4 + c] = 0.5;
				else if (i == 0 && j == 0)
					test[(i*4 + j)*4 + c] = 0.0;
				else
					test[(i*4 + j)*4 + c] = 1.0;
			}	

			cout << test[(i*4 + j)*4] << "\t"; 
		}
		cout << endl;
	}

	cout << endl;



	pullPushTexture(test, 4, 1);
*/