#ifndef _GAME_STATE_VBO
	#define _GAME_STATE_VBO

#include "GameState.h"
#include "../Math/RaytracingMath.h"
#include "../Math/Triangle.h"
#include "../Math/Point.h"
#include "../Math/Ray.h"
#include "../ResourceManager/ModelManager/particles.h"
#include "../Math/PullPush.h"

#include "GameStateVBO_RenderingStuff.h"
#include "global.h"
#include "../ResourceManager/Scene.h"
#include "../CUDA/CUDA_PushPull.h"
#include "../CUDA/CUDA_Lightcuts.h"
#include <time.h>
#include "../ResourceManager/ModelManager/ModelManager.h"
#include "../ResourceManager/TextureManager/Texture.h"
#include "../ResourceManager/ShaderManager/ShaderManager.h"





class GameStateVBO: public GameState
{
private:

	int width,height;
	bool fullScreen;
	void setParamsAccordingToFileParams(vector<vector<string> >& params);
	void setDefaultParameterValues(void);
	void glEnable2D(void);
	void glDisable2D(void);
	void movePlayer(void);
	void renderSecondPass(void);

public:
	static float scale;
	static GLuint ism_matrices;
	static GLuint framebuffer;
	static GLuint colorbufferPositions;
	static GLuint colorbufferNormals;
	static GLuint colorbufferTextureColors;
	static GLuint colorbufferCsize;
	static GLuint depthbuffer;

	static int clusterVisualizationStep;

	static float farZ;
	static float nearZ;

	static int lastRenderingWidth;
	static int lastRenderingHeight;

	/*Reserve memory, load textures/sounds/models,....
	*/
	virtual void initGameState(vector<vector<string> >& params);

	/*Free memory, prepare for GameState-Change*/
	virtual void destroyGameState(void);

	static void destroyGameState_static(void);

	/*InputHandlers:*/
	virtual void processKey(unsigned char& key, int& x, int& y);
	virtual void processKeyUp(unsigned char& key, int& x, int& y);
	virtual void processSpecialKey(int& key, int& x, int& y);
	virtual void processSpecialKeyUp(int& key, int& x, int& y);
	virtual void processMousePassiveMovement(int& x, int& y);
	virtual void processMouseMovement(int& x, int& y);
	virtual void processMouseClick(int& button, int& state, int& x, int& y);
	
	/*ResourceHandlers:*/


	/*Rendering:*/
	virtual void render(void);
	virtual void update(void);
	void renderOld(void);
	void updateOld(void);
	virtual void reshape(int w, int h);

	void renderFBO(void);

	static void generateISMTextures(void);


};

#endif