#include "GameStateVBO_RenderingStuff.h"


void getTimestamp(char* timestamp);
void rotateByMatrix(Vector* p, float mat[9]);


void storeToDisk(int width, int height)
{
	char vcFilename[1024];
	char vcFilenameCsize[1024];
	char timestamp[128];
	char infoFileName[1024];
	getTimestamp(timestamp);
	

	sprintf_s(vcFilename,1024,"Rendering_%s.jpg",timestamp);
	sprintf_s(vcFilenameCsize,1024,"Rendering_%s_cutsize.jpg",timestamp);



	unsigned int uImage = 0;
	unsigned int uImage2 = 1;
	std::vector<float> vecImage;
	std::vector<float> vecImageSize;
	std::vector<float> vecNormals;
	std::vector<float> vecColors;

	vecImage.resize(width*height*4);
	vecImageSize.resize(width*height*4);
	vecNormals.resize(width*height*4);
	vecColors.resize(width*height*4);

	CUDA_LightCuts::positionsInWorld = new float4[width*height];
	CUDA_LightCuts::normalsInWorld = new float4[width*height];
	CUDA_LightCuts::colorsInWorld = new float4[width*height];

	//glReadPixels(0,0,width,height,GL_RGBA,GL_FLOAT,(void*) &(vecImage.front()) );
	glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferPositions); 
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, (void*) &(vecImage.front()));
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, CUDA_LightCuts::positionsInWorld);

	glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferNormals); 
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, (void*) &(vecNormals.front()));
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, CUDA_LightCuts::normalsInWorld);

	glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferTextureColors); 
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, (void*) &(vecColors.front()));
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, CUDA_LightCuts::colorsInWorld);

	cout << endl;
	//NOW: DO LIGHTCUTS SHADING FOR EACH PIXEL
	if (renderMode == 3) //lightCutsInCUDA
	{
		CUDA_LightCuts::renderImageWithCUDA();
	}
	
	else if (renderMode == 1) //radiosityInCUDA
	{
		CUDA_LightCuts::renderImageWithRadiosityCUDA();
	}
	else if (renderMode == 2) //lightCutsInSoftware
	{

		TextureManager::sumOfCutsizes = 0.0f;

		if(printDebugOutput)
			printf("LightcutsShading(Software):\n");

		Timer timTimer;
		for(unsigned int i=0; i<(vecImage.size()); i+=4)
		{
			float color[6];
			if (printDebugOutput)
			{
				FILE* file = fopen("test2.test", "a");
				fprintf(file, "\n######################\nNEUER PIXEL %i\n", i/4);
				fclose(file);
			}
			TextureManager::shadeLightcuts(vecImage.at(i), vecImage.at(i+1), vecImage.at(i+2), vecNormals.at(i), vecNormals.at(i+1), vecNormals.at(i+2), vecColors.at(i), vecColors.at(i+1), vecColors.at(i+2), color);
			vecImage.at(i) = color[0];
			vecImage.at(i+1) = color[1];
			vecImage.at(i+2) = color[2];
		
			vecImageSize.at(i) = color[3];
			vecImageSize.at(i+1) = color[4];
			vecImageSize.at(i+2) = color[5];

			int aa = (width * height)/10;
			if(aa > 0 && (i/4)%aa == 0 && printDebugOutput)
			{
				printf("%2.0f %% done\r", (float(i/4.0f)*100.0f)/float(width * height));
			}
		}
		const double dTimePerFrame = double(timTimer);
		timTimer.start();
		
		if(printDebugOutput)
			printf("\nDone (%f sek).\n", dTimePerFrame);

		renderingTime = dTimePerFrame;
		cout << "Average Cut Size = " << (TextureManager::sumOfCutsizes/(width*height))*TextureManager::anzLights << " ( " << (TextureManager::sumOfCutsizes*100.0f/(width*height)) << " percent)" << endl;
		avgCutSize = TextureManager::sumOfCutsizes*100.0f/(width*height);

		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::colorTextureGL);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_FLOAT, (void*) &(vecImage.front()));

		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::cutSizeTextureGL);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_FLOAT, (void*) &(vecImageSize.front()));
	}
	else if (renderMode == 0) //radiosityInSoftware
	{
		
		if(printDebugOutput)
			printf("InstantRadiosityShading(Software):\n");
		Timer timTimer;
		for(unsigned int i=0; i<(vecImage.size()); i+=4)
		{
			float color[6];
			TextureManager::shadeIR(vecImage.at(i), vecImage.at(i+1), vecImage.at(i+2), vecNormals.at(i), vecNormals.at(i+1), vecNormals.at(i+2), vecColors.at(i), vecColors.at(i+1), vecColors.at(i+2), color);
			vecImage.at(i) = color[0];
			vecImage.at(i+1) = color[1];
			vecImage.at(i+2) = color[2];
		
			vecImageSize.at(i) = color[3];
			vecImageSize.at(i+1) = color[4];
			vecImageSize.at(i+2) = color[5];

			int aa = (width * height)/10;
			if(aa > 0 && (i/4)%aa == 0 && printDebugOutput)
			{
				printf("%2.0f %% done\r", (float(i/4.0f)*100.0f)/float(width * height));
			}
		}
		const double dTimePerFrame = double(timTimer);
		timTimer.start();
		
		if(printDebugOutput)	
			printf("\nDone (%f sek).\n", dTimePerFrame);
		renderingTime = dTimePerFrame;
	
		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::colorTextureGL);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_FLOAT, (void*) &(vecImage.front()));

		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::cutSizeTextureGL);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_FLOAT, (void*) &(vecImageSize.front()));
	}

	if (printDebugOutput)
	{
		sprintf_s(infoFileName, 1024, "Rendering_%s.txt", timestamp);
		FILE* infoFile = fopen(infoFileName, "wb");
		fprintf(infoFile, "renderingWidth = %i\n",renderingWidth);
		fprintf(infoFile, "renderingHeight = %i\n",renderingHeight);
		fprintf(infoFile, "renderingTime = %f\n",renderingTime);
		fprintf(infoFile, "# lights = %i\n",TextureManager::anzLights);
		fprintf(infoFile, "ism size = %i\n",ism_size);
		fprintf(infoFile, "numberOfThreads = %i\n",optimalNumberOfThreads);
		fprintf(infoFile, "pullpush = %i\n", pullPush);
		fprintf(infoFile, "pullPushMaxLevel = %i\n", pullPushMaxLevel);
		if (renderMode == 2 || renderMode == 3)
		{
			fprintf(infoFile, "avgCutSize = %f\n", avgCutSize);
		}
		fclose(infoFile);
	}


	ilGenImages(1, &uImage);
	
	if (ilGetError() != IL_NO_ERROR || !printDebugOutput)
	{
		if(printDebugOutput)
			std::cerr<<"Unable to create image"<<std::endl;
	}
	else
	{
		ilBindImage(uImage);

		if (ilGetError() != IL_NO_ERROR)
		{
			std::cerr<<"Unable to bind image"<<std::endl;
		}
		else
		{
			if (renderMode == 0 || renderMode == 2)
			{
				ilTexImage(width,height,1,4,IL_RGBA,IL_FLOAT,(void*) &(vecImage.front()) );
			}
			else
			{
				ilTexImage(width,height,1,4,IL_RGBA,IL_FLOAT,(void*) CUDA_LightCuts::colorTexture );
			}
			

			if (ilGetError() != IL_NO_ERROR)
			{
				std::cerr<<"Unable to store image"<<std::endl;
			}
			else
			{
			//	Use DevIL to save a screenshot of the OpenGL window
				ILboolean a = ilSaveImage(vcFilename);

				int e = e = ilGetError();
				if (e != IL_NO_ERROR)
				{
					std::cerr<<"Unable to save image"<<std::endl;
				}
				else
				{
					std::cout << "- Rendered to '" << vcFilename << "' ..." << std::endl;
					
				}
			}
		}
	}

	//we only output the cutSize if we computed lightcuts (renderMode 2 and 3):
	if (renderMode == 0 || renderMode == 1)
	{
		ilBindImage(0);
		ilDeleteImages(1,&uImage);
		vecImage.clear();
		vecNormals.clear();
		vecColors.clear();
		vecImageSize.clear();
		return;
	}

	ilGenImages(1, &uImage2);
	

	if (ilGetError() != IL_NO_ERROR ||!printDebugOutput)
	{
		if(!printDebugOutput)
			std::cerr<<"Unable to create image"<<std::endl;
	}
	else
	{
		ilBindImage(uImage2);

		if (ilGetError() != IL_NO_ERROR)
		{
			std::cerr<<"Unable to bind image"<<std::endl;
		}
		else
		{	
			if (renderMode == 2 || renderMode == 0) //software
			{
				ilTexImage(width,height,1,4,IL_RGBA,IL_FLOAT,(void*) &(vecImageSize.front()) );
			}
			else if (renderMode == 3 || renderMode == 1) //hardware
			{
				ilTexImage(width,height,1,4,IL_RGBA,IL_FLOAT,(void*) CUDA_LightCuts::cutSizeTexture );
			}

			if (ilGetError() != IL_NO_ERROR)
			{
				std::cerr<<"Unable to store image"<<std::endl;
			}
			else
			{
			//	Use DevIL to save a screenshot of the OpenGL window
				ILboolean a = ilSaveImage(vcFilenameCsize);

				int e = e = ilGetError();
				if (e != IL_NO_ERROR)
				{
					std::cerr<<"Unable to save image"<<std::endl;
				}
				else
				{
					std::cout << "- Rendered to '" << vcFilenameCsize << "' ..." << std::endl;
					
				}
			}
		}
	}

	if (uImage && uImage2)
	{
		ilBindImage(0);
		ilDeleteImages(1,&uImage);
		ilDeleteImages(1,&uImage2);
		vecImage.clear();
		vecNormals.clear();
		vecColors.clear();
		vecImageSize.clear();
	}
}

void renderImageToDisk(int width, int height)
{
	
	if(reRenderISMs)
	{
		GameStateVBO::generateISMTextures();
		reRenderISMs = false;
	}

	if(reBuildLightTree)
	{
		TextureManager::createkDTreeAndClusterLightTree();
		reBuildLightTree = false;
	}

	if (width != GameStateVBO::lastRenderingWidth || height != GameStateVBO::lastRenderingHeight)
	{
		//kein check noetig (ob gueltig oder nicht), da openGL ansonsten den Befehl einfach ignoriert
		//note: falls die aktuell gebundene texture geloescht wird, wird textur 0 gebunden (diese kann auf diese weise nicht geloescht werden).	
		glDeleteTextures(1, &CUDA_LightCuts::colorTextureGL); 
		 

		//ONLY FOR DEBUG: TEXTURES FOR CUDA RESULTS:
		glGenTextures(1, &CUDA_LightCuts::colorTextureGL);
		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::colorTextureGL);

		// set basic parameters
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// buffer data
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);

		glDeleteTextures(1, &CUDA_LightCuts::cutSizeTextureGL);
		glGenTextures(1, &CUDA_LightCuts::cutSizeTextureGL);
		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::cutSizeTextureGL);

		// set basic parameters
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// buffer data
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
		//END OF DEBUG TEXTURES


		//COLORBUFFER FOR FRAGMENT POSITIONS
		glDeleteTextures(1, &GameStateVBO::colorbufferPositions);
		glGenTextures(1, &GameStateVBO::colorbufferPositions);
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferPositions);

		// set basic parameters
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// buffer data
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);

		//COLORBUFFER FOR FRAGMENT NORMALS:
		glDeleteTextures(1, &GameStateVBO::colorbufferNormals);
		glGenTextures(1, &GameStateVBO::colorbufferNormals);
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferNormals);

		// set basic parameters
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// buffer data
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
		
		//COLORBUFFER FOR FRAGMENT COLORS:
		glDeleteTextures(1, &GameStateVBO::colorbufferTextureColors);
		glGenTextures(1, &GameStateVBO::colorbufferTextureColors);
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferTextureColors);

		// set basic parameters
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// buffer data
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);	

		glDeleteRenderbuffersEXT(1, &GameStateVBO::depthbuffer);
		glGenRenderbuffersEXT(1, &GameStateVBO::depthbuffer);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, GameStateVBO::depthbuffer);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

		// create and bind a framebuffer
		glDeleteRenderbuffersEXT(1, &GameStateVBO::framebuffer);
		glGenFramebuffersEXT(1, &GameStateVBO::framebuffer);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, GameStateVBO::framebuffer);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, GameStateVBO::colorbufferPositions, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, GameStateVBO::colorbufferNormals, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, GameStateVBO::colorbufferTextureColors, 0);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, GameStateVBO::depthbuffer);
		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, GameStateVBO::framebuffer);
	ModelManager::setShaderParams("manyLights");

	GLenum buffers[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT };
	glDrawBuffers(3, buffers);

	glClearColor(0.0f,0.0f,0.0f,0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glViewport (0, 0, width, height);
	glMatrixMode (GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity ();
	gluPerspective (60, (GLfloat)width / (GLfloat)height, 1.0, 100000.0);
	glMatrixMode (GL_MODELVIEW);
	

	GameStateVBO::camera->view();
	glColor3f (1,1,1);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	//EITHER: DIRECTLY SHADE USING ALL LIGHTS (instant radiosity)
	//ModelManager::renderAll();
	
	//OR: WRITE GEOMETRY FOR LIGHTCUTS RENDERING:
	ModelManager::renderAllGeometryToTexture(0, 0, "writeGeometry");

	storeToDisk(width, height);
	
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	
}


void renderTV(void)
{
	glActiveTexture(GL_TEXTURE0);
	glUseProgram(ShaderManager::fetchShader("skybox"));
	glPushMatrix();

	//DEBUGSCREEN
	if(debugscreen%anzopts==0)
		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::colorTextureGL);
	if(debugscreen%anzopts==1)
		glBindTexture(GL_TEXTURE_2D, CUDA_LightCuts::cutSizeTextureGL);
	if(debugscreen%anzopts==2)
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferPositions);
	if(debugscreen%anzopts==3)
		glBindTexture(GL_TEXTURE_2D, ism_atlas); 
	/*
	if(debugscreen%anzopts==0)
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferPositions);
	if(debugscreen%anzopts==1)
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::colorbufferNormals); 
	if(debugscreen%anzopts==2)
		glBindTexture(GL_TEXTURE_2D, ism_atlas); 
	if(debugscreen%anzopts==3)
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::ism_matrices); 
	if(debugscreen%anzopts==4)
		glBindTexture(GL_TEXTURE_2D, TextureManager::lc); 
	*/
	
	glBegin(GL_QUADS);	
	glColor3f (1,1,1);
	// Das vordere QUAD
	glTexCoord2d(0, 0); glVertex3f(0, 48, 0); 
	glTexCoord2d(1, 0); glVertex3f(100, 48, 0); 
	glTexCoord2d(1, 1); glVertex3f(100, 148, 0); 
	glTexCoord2d(0, 1); glVertex3f(0, 148, 0);
	glEnd();
	glPopMatrix();	
	glUseProgram(0);
}

void renderPointRepresentationOfScene(bool shaderAlreadyBound, Vector viewVec, Vector viewPos)
{
	if(!shaderAlreadyBound)
	{
		glUseProgram(ShaderManager::fetchShader(pointsShader));
		//glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	}

	//int modelCount = 0;

	//culling:
	for(int i=0; i<int(ModelManager::allModels.size()); i++)
	{
		Model* m = ModelManager::allModels.at(i);

		bool partlyVisible = false;

		for(int e=0; e<8; e++)
		{
			if((m->bvertex[e]-viewPos) * viewVec > 0)
			{
				partlyVisible = true;
				break;
			}
		}

		if(partlyVisible)
		{
			m->bindDrawingBuffersForPointSplatting();
			m->drawPointRepresentation();
			//modelCount ++;
		}
	}

	//if(modelCount != ModelManager::allModels.size())
	//	cout << modelCount << endl;
	if(!shaderAlreadyBound)
		glUseProgram(0);
}

void getUpVector(float nx, float ny, float nz, Vector* up)
{
	float n[3] = {nx, ny, nz};
	float minimum = 2.0;
	int minIndex;

	for(int i=0; i<3; i++)
	{
		if(abs(n[i]) < abs(minimum))
		{
			minimum = n[i];
			minIndex = i;
		}
	}
	
	switch(minIndex)
	{
		case 0:
				up->x = 0.0f;
				up->y = -n[2];
				up->z = n[1];
				break;
		case 1:
				up->x = -n[2];
				up->y = 0.0f;
				up->z = n[0];	
				break;
		case 2:
				up->x = -n[1];
				up->y = n[0];
				up->z = 0.0f;
				break;
	}
	up->normalize();
}



void renderImperfectShadowMaps(void)
{
	//fetch light positions and normals
	float* pixelsLP = TextureManager::pixelsLP;
	float* pixelsLN = TextureManager::pixelsLN;


	//alloc buffer memory
	float* pixbuf = new float[4*ism_size*ism_size];

	//bind ISM-framebuffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ism_framebuffer);
	glClearColor(1.0f,1.0f,1.0f,1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	int texLoc;

	int ismsRendered=0;

	//fetch point splatting shader

	glUseProgram(ShaderManager::fetchShader(pointsShader));
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(pointsShader), "farZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::farZ);
	}

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(pointsShader), "nearZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::nearZ);
	}

	//declare matrix variables
	GLfloat modelview[16]; 
	GLfloat projection[16]; 
//	GLfloat modelviewprojection[16];

	//set viewport for rendering a single ism
	glViewport (0, 0, ism_size, ism_size);

	//GET PROJECTION MATRIX (proj. matrix is the same for all lights)
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(90.0f, 1.0f, GameStateVBO::nearZ, GameStateVBO::farZ);
	glGetFloatv(GL_PROJECTION_MATRIX, projection);


	//now set back to modelview mode (which differs with every single light)
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); //save old modelview mat.

	int lightTextureW = TextureManager::lightTextureW;
	int lightTextureH = TextureManager::lightTextureH;

	//RENDER ISMS FOR ALL LIGHTS:
	for(int ax=0; ax<4*TextureManager::anzLights; ax+=4)
	{		
		int lightNr = ax/4; //index number of current light
		int lightH = lightNr/lightTextureW;
		int lightW = lightNr%lightTextureW;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//Calculate MODELVIEW matrix for current light
		glLoadIdentity();

		Vector upVec;
		getUpVector(pixelsLN[ax], pixelsLN[ax+1], pixelsLN[ax+2], &upVec);

		gluLookAt(	pixelsLP[ax], pixelsLP[ax+1], pixelsLP[ax+2], 
			pixelsLP[ax]+ pixelsLN[ax],   pixelsLP[ax+1]+pixelsLN[ax+1],   pixelsLP[ax+2]+pixelsLN[ax+2], 
			upVec.x, upVec.y, upVec.z);
		


		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

		//AND NOW WRITE MODELVIEWPROJECTION MATRIX OF CURRENT LIGHT TO TEXTURE
		glBindTexture(GL_TEXTURE_2D, GameStateVBO::ism_matrices);	
		glTexSubImage2D(GL_TEXTURE_2D, 0, lightW*2, lightH*2, 2, 2, GL_RGBA, GL_FLOAT, modelview);	

		//store location of matrix in huge texture for current vpl
		TextureManager::vplVector.at(lightNr)->lightSpaceMatrixAtPixelX = lightW*2; 
		TextureManager::vplVector.at(lightNr)->lightSpaceMatrixAtPixelY = lightH*2;

		CUDA_LightCuts::shadowMapAndLightSpaceOfVPL[TextureManager::vplVector.at(lightNr)->index].z = lightW*2;
		CUDA_LightCuts::shadowMapAndLightSpaceOfVPL[TextureManager::vplVector.at(lightNr)->index].w = lightH*2;

		int texLoc;

		if(shadowPointsOnly)
		{
			//LIGHT POSITION IN DEN SHADER STECKEN
			texLoc = glGetUniformLocation(ShaderManager::fetchShader(pointsShader), "lpos");
			if (texLoc!=-1)
				glUniform4f(texLoc, pixelsLP[ax], pixelsLP[ax+1], pixelsLP[ax+2], pixelsLP[ax+3]);

			texLoc = glGetUniformLocation(ShaderManager::fetchShader(pointsShader), "biasByPercentageOfLightVector");
			if (texLoc!=-1)
				glUniform1f(texLoc, Scene::biasByPercentageOfLightVector);

			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_NORMAL_ARRAY);
			renderPointRepresentationOfScene(true, Vector(pixelsLN[ax], pixelsLN[ax+1], pixelsLN[ax+2]), Vector(pixelsLP[ax], pixelsLP[ax+1], pixelsLP[ax+2]));
			glDisableClientState(GL_VERTEX_ARRAY);
			glDisableClientState(GL_NORMAL_ARRAY);
		}
		else
		{
			//LIGHT POSITION IN DEN SHADER STECKEN
			texLoc = glGetUniformLocation(ShaderManager::fetchShader(writeGeometryShader), "lpos");
			if (texLoc!=-1)
				glUniform4f(texLoc, pixelsLP[ax], pixelsLP[ax+1], pixelsLP[ax+2], pixelsLP[ax+3]);

			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glEnableClientState(GL_NORMAL_ARRAY);

			ModelManager::renderAllGeometryToTexture(lightW+1, lightH+1, writeGeometryShader);

			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);
			glDisableClientState(GL_NORMAL_ARRAY);
		}


		//AND STORE OBTAINED ISM IN A HUGE TEXTURE
		glReadPixels(0, 0, ism_size, ism_size, GL_RGBA, GL_FLOAT, pixbuf);
		glBindTexture(GL_TEXTURE_2D, ism_atlas);
		glTexSubImage2D(GL_TEXTURE_2D, 0, lightW*ism_size, lightH*ism_size, ism_size, ism_size, GL_RGBA, GL_FLOAT, pixbuf);

		//store location of ism in huge texture for current vpl
		TextureManager::vplVector.at(lightNr)->shadowMapAtPixelX = lightW*ism_size; 
		TextureManager::vplVector.at(lightNr)->shadowMapAtPixelY = lightH*ism_size;

		CUDA_LightCuts::shadowMapAndLightSpaceOfVPL[TextureManager::vplVector.at(lightNr)->index].x = lightW*ism_size; 
		CUDA_LightCuts::shadowMapAndLightSpaceOfVPL[TextureManager::vplVector.at(lightNr)->index].y = lightH*ism_size;

		ismsRendered++;
		if(ismsRendered % 10 == 0 && printDebugOutput)
			printf("Rendering %i ISMS... (%05i remaining)\r", TextureManager::anzLights, TextureManager::anzLights - ismsRendered);

	}
	if(printDebugOutput)
		printf("\n");

	if(lightSpaceMatrices!=0)
	{
		delete[] lightSpaceMatrices;
		lightSpaceMatrices=0;
	}

	if(CUDA_LightCuts::lightMatricesCUDA!=0)
	{
		delete[] CUDA_LightCuts::lightMatricesCUDA;
		CUDA_LightCuts::lightMatricesCUDA=0;
	}

	CUDA_LightCuts::lightMatricesCUDA = new float[lightSpaceMatrices_size * lightSpaceMatrices_size * 4];
	lightSpaceMatrices = new float[lightSpaceMatrices_size * lightSpaceMatrices_size*4];
	glBindTexture(GL_TEXTURE_2D, GameStateVBO::ism_matrices);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, lightSpaceMatrices);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, CUDA_LightCuts::lightMatricesCUDA);

	CUDA_PushPull::pullPushTexture();

	//unbind ism buffer and shader
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glUseProgram(0);

	//restore prior matrices
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	delete[] pixbuf; pixbuf = 0;
}

void visualizeLights(void)
{
	glUseProgram(0);

	float* pixelsLC = TextureManager::pixelsLC;
	float* pixelsLP = TextureManager::pixelsLP;
	float* pixelsLN = TextureManager::pixelsLN;

	for(int ax=0; ax<4*TextureManager::anzLights; ax+=4)
	{
		glPushMatrix();
		glTranslatef(pixelsLP[ax], pixelsLP[ax+1], pixelsLP[ax+2]);

		glColor4f(float(pixelsLC[ax]), float(pixelsLC[ax+1]), float(pixelsLC[ax+2]), float(pixelsLC[ax+3]));
		int temp = int(10*sqrt(pixelsLC[ax+3]));
			
		glutSolidSphere(temp, 5, 5);

		glLineWidth(4.0f);
		glBegin(GL_LINES);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(temp * 4.0f * pixelsLN[ax], temp * 4.0f * pixelsLN[ax+1], temp * 4.0f * pixelsLN[ax+2]);
		glEnd();
		glLineWidth(2.0f);
/*
		Vector upVec;
		getUpVector(pixelsLN[ax], pixelsLN[ax+1], pixelsLN[ax+2], &upVec);

		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glLineWidth(4.0f);
		glBegin(GL_LINES);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(temp * 4.0f * upVec.x, temp * 4.0f * upVec.y, temp * 4.0f * upVec.z);
		glEnd();
		glLineWidth(2.0f);
*/
		glPopMatrix();
	}
}

void visualizeOneCluster(Cluster* cluster)
{
	glPushMatrix();

	glTranslatef(cluster->representativeLight->position.x, cluster->representativeLight->position.y, cluster->representativeLight->position.z);

	glColor3f(1.0f, 1.0f, 1.0f);
	int temp = int(10*sqrt(cluster->intensity));
		
	glutSolidSphere(temp, 5, 5);

	glPopMatrix();

	glLineWidth(10.0f);
	glBegin(GL_QUADS);

	//###unteres quad
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.B.z);

	//###vorderes quad
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.A.z);

	//###linkes quad
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.B.z);

	//###rechtes quad
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.A.z);

	//###hinteres quad
	glVertex3f(cluster->bounds.A.x, cluster->bounds.A.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.A.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.B.z);

	//###oberes quad
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.A.z);
	glVertex3f(cluster->bounds.B.x, cluster->bounds.B.y, cluster->bounds.B.z);
	glVertex3f(cluster->bounds.A.x, cluster->bounds.B.y, cluster->bounds.B.z);
	
	glEnd();
	glLineWidth(2.0f);
}

void visualizeClustering(int step)
{
	if (TextureManager::lightTree == 0)
		return;

	Cluster* root = TextureManager::lightTree->root;

	//If we haven't clustered yet, return:
	if (root == 0)
		return;

	list<Cluster*> allClustersToVisualize;
	allClustersToVisualize.push_back(root);

	for (int i = 0; i < step; i+=1)
	{
		list<Cluster*>::iterator it = allClustersToVisualize.begin();

		Cluster* frontElement = (*it);
		Cluster* leftChild = frontElement->leftChild;
		Cluster* rightChild = frontElement->rightChild;

		if (leftChild == 0)
		{
			allClustersToVisualize.pop_front();
			allClustersToVisualize.push_back(frontElement);
			continue;
		}

		//replace cluster with its children:
		allClustersToVisualize.pop_front();
		allClustersToVisualize.push_front(leftChild);
		allClustersToVisualize.push_front(rightChild);
	}

	list<Cluster*>::iterator it;
	for (it = allClustersToVisualize.begin(); it != allClustersToVisualize.end(); it++)
	{
		visualizeOneCluster((*it));
	}

	
}

void drawSkybox(Camera* camera)
{	
	glActiveTexture(GL_TEXTURE0);
	glUseProgram(ShaderManager::fetchShader("skybox"));
	float sz=100000.0f;
	glPushMatrix();
	glTranslatef(camera->posX, camera->posY, camera->posZ);
	glTranslatef(0, (-sz/2), 0);

	glPushMatrix();
	glTranslatef(-(sz/2), 0, -(sz/2));

	TextureManager::fetchTexture("front.jpg", 0, 0); 
	glBegin(GL_QUADS);	
	glColor3f (1,1,1);
	// Das vordere QUAD
	glTexCoord2d(0, 0); glVertex3f(0, 0, 0); 
	glTexCoord2d(1, 0); glVertex3f( sz, 0, 0); 
	glTexCoord2d(1, 1); glVertex3f( sz, sz, 0); 
	glTexCoord2d(0, 1); glVertex3f(0, sz, 0); 
	glEnd();
	glPopMatrix();

	glPushMatrix();
	glTranslatef(sz/2, 0, sz/2);
	TextureManager::fetchTexture("right.jpg", 0, 0); 
	//	glBindTexture(GL_TEXTURE_2D, tex_screen);
	glBegin(GL_QUADS);	
	// Das rechte QUAD
	glTexCoord2d(1, 0); glVertex3f( 0, 0, 0); 
	glTexCoord2d(1, 1); glVertex3f( 0, sz, 0); 
	glTexCoord2d(0, 1); glVertex3f( 0, sz, -sz); 
	glTexCoord2d(0, 0); glVertex3f( 0, 0, -sz); 
	glEnd();
	glPopMatrix();

	glPushMatrix();
	glTranslatef((-sz/2), 0, sz/2);
	TextureManager::fetchTexture("back.jpg", 0, 0); 
	glBegin(GL_QUADS);	
	// Das hintere QUAD
	glTexCoord2d(1, 0); glVertex3f(0, 0, 0);
	glTexCoord2d(1, 1); glVertex3f(0, sz, 0);
	glTexCoord2d(0, 1); glVertex3f( sz, sz, 0);
	glTexCoord2d(0, 0); glVertex3f( sz, 0, 0);
	glEnd();
	glPopMatrix();

	glPushMatrix();
	glTranslatef((-sz/2), 0, sz/2);
	TextureManager::fetchTexture("left.jpg", 0, 0); 
	glBegin(GL_QUADS);	
	// Das linke QUAD
	glTexCoord2d(0, 0); glVertex3f(0, 0, 0); 
	glTexCoord2d(1, 0); glVertex3f(0, 0, -sz); 
	glTexCoord2d(1, 1); glVertex3f(0, sz, -sz); 
	glTexCoord2d(0, 1); glVertex3f(0, sz, 0); 
	glEnd();
	glPopMatrix();


	glPushMatrix();
	glTranslatef((-sz/2), sz, (sz/2));
	TextureManager::fetchTexture("upper.jpg", 0, 0); 
	glBegin(GL_QUADS);	
	// Das obere QUAD	
	//vertexCoords CW, texCoords CCW
	glTexCoord2f(0, 1); glVertex3f(0,0,0);	
	glTexCoord2f(0, 0); glVertex3f(0,0,-sz);	
	glTexCoord2f(1, 0); glVertex3f(sz,0,-sz);	
	glTexCoord2f(1, 1); glVertex3f(sz,0,0);		


	glEnd();


	glPopMatrix();

	glPopMatrix();		

}

void getTimestamp(char* timestamp)
{
	time_t ltime;
	ltime = time(NULL);
	tm something;
	localtime_s(&something, &ltime);

	sprintf_s(timestamp, 50, "%02i_%02i_%04i-%02i_%02i_%02i", something.tm_mday, something.tm_mon+1, something.tm_year+1900, something.tm_hour, something.tm_min, something.tm_sec);
	
	if (renderMode == 0 || renderMode == 2)
	{
		strcat(timestamp, "_SW_");
	}
	else
	{
		strcat(timestamp, "_HW_");
	}

	if (renderMode == 0 || renderMode == 1)
	{
		strcat(timestamp, "RA");
	}
	else
	{
		strcat(timestamp, "LC");
	}
	
}

