// GraphicsDemo.cpp : Defines the entry point for the console application.
//


#include "GraphicsDemo.h"


int window;
char title[64];
GameState* GameStateManager::activeGameState = 0;

void idleCallback(void* pData)
{
	if (sceneIsBeingLoaded)
		return;
	if(pData!=0)
	{
		MyGlWindow* b = reinterpret_cast<MyGlWindow*>(pData);
		b->redraw();
	}
}


GameStateManager::GameStateManager()
{
	activeGameState = new GameStateVBO();
	activeGameState->initGameState(staticOperations::retrieveParamsFromFile(string("GameStates/GameStateVBO.ini")));
}

GameStateManager::~GameStateManager()
{
	if (activeGameState!=NULL) activeGameState->destroyGameState();
	delete activeGameState;
	activeGameState = NULL;
}

MyGlWindow::~MyGlWindow(void)
{
	if (gameStateManager!=0)
	{
		delete gameStateManager;
		gameStateManager = 0;
	}
	Fl::remove_idle(idleCallback, this);

}
/*
void GameStateManager::killProgram()
{
	if (activeGameState!=NULL) activeGameState->destroyGameState();
	delete activeGameState;
	activeGameState = NULL;
	
	exit(0);
}
*/

void GameStateManager::mouseClick(int button, int state, int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processMouseClick(button, state, x, y);
}

void GameStateManager::mouseMovement(int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processMouseMovement(x,y);
}

void GameStateManager::mousePassiveMovement(int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processMousePassiveMovement(x,y);
}

void GameStateManager::render(void)
{
	if (sceneIsBeingLoaded)
		return;

	Timer timTimer;
	timTimer.start();

	activeGameState->render();
	
	const double dTimePerFrame = double(timTimer);
	const double dFramesPerSecond = dTimePerFrame > 0.0 ? 1.0 / dTimePerFrame : 0.0;
	sprintf_s(title, "- %06.2f frames/second \r",dFramesPerSecond);

	//printf("- %06.2f frames/second \r",dFramesPerSecond);
//	glutSetWindowTitle(title);
}

void GameStateManager::update(void)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->update();
}

void GameStateManager::reshape (int w, int h)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->reshape(w,h);
}



void GameStateManager::keyboard (unsigned char key, int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	/*
	// no matter where you are - ESC quits the program ;-)
	if (key==27)
	{
		deleteStuff();
	//	killProgram();
		return;
	}
*/
	activeGameState->processKey(key, x, y);
}

void GameStateManager::keyboardUp (unsigned char key, int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processKeyUp(key, x, y);
}

void GameStateManager::keyboardSpecial (int key, int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processSpecialKey(key, x, y);
}

void GameStateManager::keyboardSpecialUp (int key, int x, int y)
{
	if (sceneIsBeingLoaded)
		return;
	activeGameState->processSpecialKeyUp(key, x, y);
}


/*
* Opens the given file and attempts to read from it. If file does not exist, 
* returns an empty vector and displays an error-message. If successful, returns
* a vector containing all correctly formatted parameters stored in the file. 
*/
vector<vector<string> > staticOperations::retrieveParamsFromFile(string& fileName)
{
	ifstream inFile;    
	inFile.open(fileName.c_str());
    if (!inFile) 
	{
		printf("Unable to open file %s.\n", fileName.c_str());
		vector<vector<string> > emptyVector;
		inFile.close();
        return emptyVector; 
    }
		
	string ab;
	vector<vector<string> > params;
	vector<string> a;

	while(getline(inFile, ab))
	{
		if (ab.find_first_of("*")!=string::npos)
		{
			continue;
		}
		if(ab.compare("")==0) //skip empty lines
		{
			continue;
		}

		size_t pos = ab.find("=");
		if (ab.find("=") == string::npos) 
		{
			printf("Error reading file %s. Please check for spelling-errors.\n",fileName.c_str());
			continue;
		}
		a.clear();

		string name = ab.substr(0,pos-1);
		string value = ab.substr(pos+1);
		a.push_back(name);
		a.push_back(value);
		params.push_back(a);
	}
	
	inFile.close();
	return params;
}

int MyGlWindow::handle(int e)
{
	if (sceneIsBeingLoaded)
		return 1;
	switch(e)
	{
		case FL_SHORTCUT:
		{
			int key = Fl::event_key();
			switch (key)
			{
				case FL_Escape: case 'v':
					  {
						  return Fl_Gl_Window::handle(e);
					  }
				}
			if (key == FL_Escape)
			{
				return Fl_Gl_Window::handle(e);
			}
			gameStateManager->keyboard (Fl::event_key(), 0,0);
			
			break;
		}
		case FL_KEYUP:
		{
			gameStateManager->keyboardUp (Fl::event_key(), 0,0);
			break;
		}

		case FL_DRAG:
		{
			gameStateManager->mouseMovement(Fl::event_x(), Fl::event_y());
			break;
		}

		case FL_PUSH:
		{
			gameStateManager->mouseClick(Fl::event_button(), FL_BUTTON1 || FL_BUTTON2 || FL_BUTTON3, Fl::event_x(), Fl::event_y());
			break;
		}

		default:
			return Fl_Gl_Window::handle(e);
	}
	//redraw();
	return 1;
}




/*
int main(int argc, char **argv)
{
	#ifdef _DEBUG
		InitMemoryTracker();
	#endif
	activeGameState = new GameStateVBO();

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA | GLUT_STENCIL);

	glutInitWindowSize(300, 300);
	glutInitWindowPosition(100, 100);
	window = glutCreateWindow("Main");
	activeGameState->initGameState(staticOperations::retrieveParamsFromFile(string("GameStates/GameStateVBO.ini")));

//	glutIgnoreKeyRepeat(1);
	glutDisplayFunc(render);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutSpecialFunc(keyboardSpecial);
	glutSpecialUpFunc(keyboardSpecialUp);
	glutMouseFunc(mouseClick);
	glutMotionFunc(mouseMovement);
	glutPassiveMotionFunc(mousePassiveMovement);
	glutIdleFunc(update);

	glutSetWindow(window);

	atexit(killProgram); //~ exit-callback

	glutMainLoop();

	return 0;
}
*/