#ifndef _GRAPHICS_DEMO
#define _GRAPHICS_DEMO

#include <vector>
#include "GameStates/GameState.h"
#include "GameStates/GameStateVBO.h"
#include "Timer.h"
#include <DevIL/il.h>
#include <DevIL/ilu.h>

#include <FL/Fl.h>
#include <FL/Fl_Gl_Window.h>
#include <FL/Fl_Check_Button.h>
#include <FL/Fl_Value_Slider.H>
#include "GameStates/global.h"
#include "ResourceManager/Scene.h"
#include <FL/FL_CHOICE.h>
#include <FL/FL_INPUT.h>

extern Fl_Choice* lightChoice;
extern Fl_Input *primary_lights_input;
extern Fl_Input *vpls_per_pl_input;
extern Fl_Input *intensity_input;
extern Fl_Input* width_input;
extern Fl_Input* height_input;

extern int renderingHeight;
extern int renderingWidth;
extern char width_c[10];
extern char height_c[10];

using namespace std;

void idleCallback(void* pData);

class GameStateManager
{
public:
	static GameState* activeGameState;
	GameStateManager(void);
	~GameStateManager(void);
	void render(void);
	void update(void);
	void reshape (int w, int h);
	void mouseMovement(int x, int y);
	void mousePassiveMovement(int x, int y);
	void mouseClick(int button, int state, int x, int y);
	void keyboard (unsigned char key, int x, int y);
	void keyboardUp (unsigned char key, int x, int y);
	void keyboardSpecial (int key, int x, int y);
	void keyboardSpecialUp (int key, int x, int y);
	//	void killProgram(void);
};


class MyGlWindow : public Fl_Gl_Window 
{

	bool initialized;

	void draw()
	{
		if (!valid()) 
		{ 
			valid(1);

			if(!initialized)
			{
				int err=glewInit();
				if(err!=GLEW_OK)
				{
					printf("FEHLER beim Initialisieren von GLEW. Beende Programm...");
					exit(-1);

				}
				// Initialize DevIL (an image library)
				std::cout << "- Initializing DevIL ..." << std::endl;
				ilInit();
				if (ilGetError() != IL_NO_ERROR)
				{
					std::cerr << "Error initializing DevIL" << std::endl;
				}
				iluInit();
				ilEnable(IL_ORIGIN_SET);
				ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
				ilEnable(IL_FILE_OVERWRITE); 
				std::cout << "- DevIL initialized." << std::endl << std::endl;
				gameStateManager = new GameStateManager();
				gameStateManager->reshape(w(),h()); 


				list<AreaLight*>::iterator it;

				for (it = Scene::areaLightSources.begin(); it != Scene::areaLightSources.end(); it++)
				{
					lightChoice->add((*it)->name.c_str());
				}

				AreaLight* firstLight =  (*Scene::areaLightSources.begin());
				lightChoice->value(0);

				primary_lights_input ->static_value(firstLight->anzPrimaryLights_c);

				vpls_per_pl_input ->static_value(firstLight->virtualToPrimaryRatio_c);

				intensity_input->static_value(firstLight->intensity_c);

				sprintf_s(width_c, "%i", renderingWidth);
				width_input->static_value(width_c);

				sprintf_s(height_c, "%i", renderingHeight);
				height_input->static_value(height_c);


				initialized = true;
			}
		}      

		gameStateManager->render();
	};
	void resize(int x, int y, int w, int h)
	{
		Fl_Gl_Window::resize(x,y,w,h);
		gameStateManager->reshape(w,h);
		redraw();
	}
	int handle(int key);

public:
	GameStateManager* gameStateManager;
	MyGlWindow(int X,int Y,int W,int H,const char*L=0) : Fl_Gl_Window(X,Y,W,H,L) 
	{
		//	gameStateManager = new GameStateManager();
		initialized = false;
		Fl::add_idle(idleCallback, this);

		//		atexit(deleteStuff);
	}
	~MyGlWindow(void);

};

class staticOperations
{
private:
public:
	static vector<vector<string> > retrieveParamsFromFile(string& fileName);
};

#endif
