#include "Color.h"
#include "Vector.h"


float3 Color::lab2lch(float3 lab)
{
	float3 lch;
	lch.x = lab.x;

	lch.y = sqrt(pow(lab.y, 2.0f) + pow(lab.z, 2.0f));

	lch.z = atan2(lab.z , lab.y);

	//rad to grad
	float angle = (180.0f / float(PI)) * lch.z;

	if (angle < 0.0f)
		angle += 360.0f;
	if (angle > 360.0f)
		angle -= 360.0f;

	lch.z = angle;
	return lch;
}

/*
 * Color difference:
 * Delta E (CIE 1994)
 */
float3 Color::rgb2lab(float3 rgb)
{
	//first step: RGB to XYZ
	//@matrix: see http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
	float mat[9] = {
	0.4124564f,  0.3575761f,  0.1804375f,
	0.2126729f,  0.7151522f,  0.0721750f,
	0.0193339f, 0.1191920f,  0.9503041f
	};

	//float mat2[9] = {
	//0.9504700f,  0.0000000f,  0.0000000f,
	//0.0000000f,  1.0000000f,  0.0000000f,
	//0.0000000f,  0.0000000f,  1.0888300f
	//};
	
	float mat2[9] = {
 0.9531874f, -0.0265906f,  0.0238731f,
-0.0382467f,  1.0288406f,  0.0094060f,
 0.0026068f, -0.0030332f,  1.0892565f
	};






//	float gamma = 2.2f;
	float3 rgb_s = make_float3(rgb.x, rgb.y, rgb.z);
/*
	retVal.x = retVal.x * mat2[0] + retVal.y * mat2[1] + retVal.z * mat2[2];
	retVal.y = retVal.x * mat2[3] + retVal.y * mat2[4] + retVal.z * mat2[5];
	retVal.z = retVal.x * mat2[6] + retVal.y * mat2[7] + retVal.z * mat2[8];
*/
	//retVal.x = retVal.x * mat[0] + retVal.y * mat[1] + retVal.z * mat[2];
	//retVal.y = retVal.x * mat[3] + retVal.y * mat[4] + retVal.z * mat[5];
	//retVal.z = retVal.x * mat[6] + retVal.y * mat[7] + retVal.z * mat[8];

	if (rgb_s.x <= 0.04045f)
	{
		rgb_s.x = rgb_s.x / 12.92f;
	}
	else
	{
		rgb_s.x = pow(((rgb_s.x + 0.055f)/1.055f), 2.4f);
	}

	if (rgb_s.y <= 0.04045f)
	{
		rgb_s.y = rgb_s.y / 12.92f;
	}
	else
	{
		rgb_s.y = pow(((rgb_s.y + 0.055f)/1.055f), 2.4f);
	}

	if (rgb_s.z <= 0.04045f)
	{
		rgb_s.z = rgb_s.z / 12.92f;
	}
	else
	{
		rgb_s.z = pow(((rgb_s.z + 0.055f)/1.055f), 2.4f);
	}


	//retVal.x = pow(retVal.x, gamma);
	//retVal.y = pow(retVal.x, gamma);
	//retVal.z = pow(retVal.x, gamma);

	float3 retVal;

	
	/*
	//reference light D50 correction = chromatic adaptation:
	//D65 light: X = 0.95047f, Y = 1.0f, Z = 1.08883f
	float rho =  0.95047f * mat2[0] + 1.0f * mat2[1] + 1.08883f * mat2[2];
	float gamma = 0.95047f * mat2[3] + 1.0f * mat2[4] + 1.08883f * mat2[5];
	float beta = 0.95047f * mat2[6] + 1.0f * mat2[7] + 1.08883f * mat2[8];

	//D50 light: X =  0.96422f, Y = 1.0f, Z = 0.82521f


	retVal.x = rgb_s.x * mat2[0] + rgb_s.y * mat2[1] + rgb_s.z * mat2[2];
	retVal.y = rgb_s.x * mat2[3] + rgb_s.y * mat2[4] + rgb_s.z * mat2[5];
	retVal.z = rgb_s.x * mat2[6] + rgb_s.y * mat2[7] + rgb_s.z * mat2[8];
	*/

	retVal.x = rgb_s.x * mat[0] + rgb_s.y * mat[1] + rgb_s.z * mat[2];
	retVal.y = rgb_s.x * mat[3] + rgb_s.y * mat[4] + rgb_s.z * mat[5];
	retVal.z = rgb_s.x * mat[6] + rgb_s.y * mat[7] + rgb_s.z * mat[8];



	//second step: XYZ to Lab
	float f_x, f_y, f_z;

	if (retVal.x > epsilon)
	{
		f_x = pow(retVal.x, 1.0f/3.0f); 
	}
	else
	{
		f_x = ((kappa * retVal.x) + 16.0f)/116.0f;
	}

	if (retVal.y > epsilon)
	{
		f_y = pow(retVal.y, 1.0f/3.0f);
	}
	else
	{
		f_y = ((kappa * retVal.y) + 16.0f)/116.0f;
	}

	if (retVal.z > epsilon)
	{
		f_z = pow(retVal.z, 1.0f/3.0f);
	}
	else
	{
		f_z = ((kappa * retVal.z) + 16.0f)/116.0f;
	}

	//Lab = retVal.x, y, z
	retVal.x = (116.0f * f_y) - 16.0f;
	retVal.y = 500.0f * (f_x - f_y);
	retVal.z = 200.0f * (f_y - f_z);
	
	return retVal;
}

float3 Color::rgb2lch(float3 rgb)
{
	float3 labVal = rgb2lab(rgb);
	float3 lchVal = lab2lch(labVal);
	return lchVal;
}

float3 Color::rgb2lch(float rgb_v4[4])
{
	float3 rgb = make_float3(rgb_v4[0], rgb_v4[1], rgb_v4[2]);
	return rgb2lch(rgb);
}


float Color::colorDistWithLuminance(float3 lab1, float3 lab2)
{
	//First step, convert Lab to Lch:
	float3 lch1, lch2;
	lch1 = lab2lch(lab1);
	lch2 = lab2lch(lab2);

	float deltaL = lch1.x - lch2.x;
	float c1 = sqrt(pow(lch1.x, 2.0f) + pow(lch1.y, 2.0f));
	float c2 = sqrt(pow(lch2.x, 2.0f) + pow(lch2.y, 2.0f));
	float deltaC = c2 - c1;
	float deltaH = lch2.z - lch1.z;

	float retVal = sqrt(pow(deltaL, 2.0f) + pow ((deltaC / (1+0.045f * c1)) , 2.0f) + pow(deltaH / (1+0.015f * c1), 2.0f));
	return retVal / 100.0f;
}

float Color::colorDistOmittingLuminance_LAB(float3 lab1, float3 lab2)
{
	//First step, convert Lab to Lch:
	float3 lch1, lch2;
	lch1 = lab2lch(lab1);
	lch2 = lab2lch(lab2);

	return colorDistOmittingLuminance_LCH(lch1, lch2);
}

float Color::colorDistOmittingLuminance_LCH(float3 lch1, float3 lch2)
{
	float c1 = sqrt(pow(lch1.x, 2.0f) + pow(lch1.y, 2.0f));
	float c2 = sqrt(pow(lch2.x, 2.0f) + pow(lch2.y, 2.0f));
	float deltaC = c2 - c1;
	float deltaH = lch2.z - lch1.z;//sqrt(pow(deltaA, 2.0f) + pow(deltaB, 2.0f) - pow(deltaC, 2.0f));

	float retVal = sqrt(pow ((deltaC / (1+0.045f * c1)) , 2.0f) + pow(deltaH / (1+0.015f * c1), 2.0f));
	return retVal / 100.0f;
}

float Color::colorDistOmittingLuminance_LCH(float4 lch1_v4, float4 lch2_v4)
{
	float3 lch1 = make_float3(lch1_v4.x, lch1_v4.y, lch1_v4.z);
	float3 lch2 = make_float3(lch2_v4.x, lch2_v4.y, lch2_v4.z);
	return colorDistOmittingLuminance_LCH(lch1, lch2);
}

float Color::colorDistOmittingLuminance_LCH(float lch1_array[4], float lch2_array[4])
{
	float3 lch1 = make_float3(lch1_array[0], lch1_array[1], lch1_array[2]);
	float3 lch2 = make_float3(lch2_array[0], lch2_array[1], lch2_array[2]);
	return colorDistOmittingLuminance_LCH(lch1, lch2);
}



float Color::colorDistOmittingLuminance_RGB(float r1, float g1, float b1, float r2, float g2, float b2)
{
	float3 rgb1 = make_float3(r1, g1, b1);
	float3 rgb2 = make_float3(r2, g2, b2);

	float3 lab1 = rgb2lab(rgb1);
	float3 lab2 = rgb2lab(rgb2);

	float retVal = colorDistOmittingLuminance_LAB(lab1, lab2);
	return retVal;
}