#ifndef __COLOR_H
#define __COLOR_H

#include "../CUDA/CUDA_Includes.h"

#define epsilon 0.008856f
#define kappa 903.3f

class Color
{
private: 
	static float3 lab2lch(float3 lab);
public:
	static float3 rgb2lab(float3 rgb);
	static float3 rgb2lch(float3 rgb);
	static float3 rgb2lch(float rgb[4]);
	
	/**
	 * Expects values to be in the Lab - format. Returned values are in the range [0,1].
	 */
	static float colorDistWithLuminance(float3 lab1, float3 lab2);
	
	/**
	 * Expects values to be in the Lab - format. Returned values are in the range [0,1].
	 */
	static float colorDistOmittingLuminance_LAB(float3 lab1, float3 lab2);

	/**
	 * Expects values to be in the LCH - format. Returned values are in the range [0,1].
	 */
	static float colorDistOmittingLuminance_LCH(float3 lch1, float3 lch2);
	static float colorDistOmittingLuminance_LCH(float4 lch1, float4 lch2);
	static float colorDistOmittingLuminance_LCH(float lch1[4], float lch2[4]);
	
	/**
	 * Expects values to be normalized RGBs. Returned values are in the range [0,100].
	 */
	static float colorDistOmittingLuminance_RGB(float r1, float g1, float b1, float r2, float g2, float b2);
};



#endif