#include <math.h>
#include "CudaStuff.h"
#include "../GameStates/global.h"

LightTree::LightTree(int _numLights)
{
	numLights = _numLights;
	data = new float[numLights*12];//3 for pos, 3 for color, 3 for normal, 1 for intensity, 2 for pointers
	nextFreeSlot = 0;
}

LightTree::~LightTree()
{
	delete[] data;
	data = 0;
}

/*
 * Assertion: data - structure isn't already full!
 */
void LightTree::addVPL(VPL newVLP)
{
	data[nextFreeSlot+0] = newVLP.position.x;
	data[nextFreeSlot+1] = newVLP.position.y;
	data[nextFreeSlot+2] = newVLP.position.z;

	data[nextFreeSlot+3] = newVLP.color[0];
	data[nextFreeSlot+4] = newVLP.color[1];
	data[nextFreeSlot+5] = newVLP.color[2];

	data[nextFreeSlot+6] = newVLP.color[3];

	data[nextFreeSlot+7] = newVLP.normal.x;
	data[nextFreeSlot+8] = newVLP.normal.y;
	data[nextFreeSlot+9] = newVLP.normal.z;

	data[nextFreeSlot+10] = -1.0f;
	data[nextFreeSlot+11] = -1.0f;

	nextFreeSlot+=12;
}

void LightTree::addVPL(Point pos, Vector normal, float* color)
{
	data[nextFreeSlot+0] = pos.x;
	data[nextFreeSlot+1] = pos.y;
	data[nextFreeSlot+2] = pos.z;

	data[nextFreeSlot+3] = color[0];
	data[nextFreeSlot+4] = color[1];
	data[nextFreeSlot+5] = color[2];

	data[nextFreeSlot+6] = color[3];

	data[nextFreeSlot+7] = normal.x;
	data[nextFreeSlot+8] = normal.y;
	data[nextFreeSlot+9] = normal.z;

	data[nextFreeSlot+10] = -1.0f;
	data[nextFreeSlot+11] = -1.0f;

	nextFreeSlot+=12;
}

void LightTree::clusterLightTree(void)
{
	
}

float* LightTree::getLightTree(void)
{
	return data;
}

void LightTree::debugOutput(void)
{
	for (int i = 0; i < numLights*12; i+=12)
	{
		printf("%i: PosX = %f; PosY = %f; PosZ = %f\nRed = %f;Green = %f; Blue = %f; Intensity = %f\n\n", 
			i/12, 
			data[i+0], data[i+1], data[i+2],
			data[i+3], data[i+4], data[i+5],
			data[i+6]);
	}
}

int ID_COUNTER = 0;
int anzFindClosestDown=0;

CpuLightTree::CpuLightTree()
{
	lights = new deque<Cluster*>;
}

CpuLightTree::~CpuLightTree()
{
	delete lights;
	lights = 0;
	kd_tree.recursiveDelete(kd_tree.root);

	if(root!=0)
		recursiveDeleteCluster(root);
}

Cluster::Cluster(VPL* newVPL)
{
	representativeLight = newVPL;//new VPL(newVPL);
	leftChild = 0;
	rightChild = 0;

	bounds.A = representativeLight->position;
	bounds.B = representativeLight->position;

	intensity = representativeLight->color[3];
	ID_COUNTER++;
	ID = ID_COUNTER;
	clusterNumber = anzClusters;

	materialTerm = 0.0;
	geometryTerm = 0.0;
	visibilityTerm = 0.0;
	radianceUpperBound = 0.0;
	estimatedRadiance = 0.0;
	errorBound = 0.0;
}

Cluster::Cluster()
{
	representativeLight = 0;
	leftChild = 0;
	rightChild = 0;
	intensity = 0;
	ID_COUNTER++;
	ID = ID_COUNTER;
	clusterNumber = anzClusters;

	materialTerm = 0.0;
	geometryTerm = 0.0;
	visibilityTerm = 0.0;
	radianceUpperBound = 0.0;
	estimatedRadiance = 0.0;
	errorBound = 0.0;
}

Cluster::Cluster(const Cluster* newCluster)
{
	representativeLight = new VPL(newCluster->representativeLight);
	leftChild = newCluster->leftChild;
	rightChild = newCluster->rightChild;
	intensity = newCluster->intensity;
	bounds = newCluster->bounds;
	ID = newCluster->ID;
	clusterNumber = anzClusters;

	materialTerm = 0.0;
	geometryTerm = 0.0;
	visibilityTerm = 0.0;
	radianceUpperBound = 0.0;
	estimatedRadiance = 0.0;
	errorBound = 0.0;
}

Cluster::~Cluster()
{	
}

void recursiveDeleteCluster(Cluster* a)
{
	if (a->leftChild!=0)
	{
		recursiveDeleteCluster(a->leftChild);
	}
	if (a->rightChild!=0)
	{
		recursiveDeleteCluster(a->rightChild);
	}

	delete a;
	a = 0;
}

bool Cluster::operator==(const Cluster& v)
{
	bool retVal = true;
	
	//retVal =(v.intensity==intensity); 
	retVal&=(v.representativeLight->position.x == representativeLight->position.x);
	retVal&=(v.representativeLight->position.y == representativeLight->position.y);
	retVal&=(v.representativeLight->position.z == representativeLight->position.z);
	float diffX = abs(v.representativeLight->position.x) - abs(representativeLight->position.x);
	float diffY = abs(v.representativeLight->position.y) - abs(representativeLight->position.y);
	float diffZ = abs(v.representativeLight->position.z) - abs(representativeLight->position.z);
	if (abs(diffX)<0.01f && abs(diffY)<0.01f && abs(diffZ)<0.01f && (!retVal))
	{
		int bla = 0;
	}
	return retVal;
}

void CpuLightTree::addCluster(Cluster* cluster)
{
	lights->push_back(cluster);
}

void CpuLightTree::buildClusters2(void)
{
	Timer timTimer;

//	priority_queue<ClusterPair> priority;
	priority_queue<ClusterPair*, vector<ClusterPair*>, ComparePtrToA> priority;
	
	if(printDebugOutput)
		printf("Building kdTree...");

	kd_tree.buildTree(lights);
	
	if(printDebugOutput)
	{
		printf("done!\n");
		printf("Finding nearest neighbours to all individual lights:\n");
	}
//	kd_tree.debugOutputTree(&kd_tree.root);
	deque<Cluster*>::iterator it1;
	float sum = kd_tree.sumOfIntensities(kd_tree.root, 0);
	bool* searchHasEnded = new bool();
	(*searchHasEnded) = false;
	if(lights->size()<2)
	{
		root = lights->at(0);
	}
	else
	{
		for (it1 = lights->begin(); it1 != lights->end(); it1++)
		{
			/*
			KD_Node* searchResult = kd_tree.findNode((*it1), &kd_tree.root, 0);
			KD_Node* a = kd_tree.findClosestNeighbour(searchResult, &kd_tree.root);
			ClusterPair newCluster = ClusterPair((*it1), a->thisNode);
			priority.push(newCluster);
			*/

			KD_Node* searchResult = kd_tree.findNode((*it1), (*it1)->ID, kd_tree.root, 0, searchHasEnded);
			(*searchHasEnded) = false;

			KD_Node* a = kd_tree.findClosestNeighbour(searchResult, kd_tree.root);
		//	ClusterPair newCluster = ClusterPair((*it1), a->thisNode);
			ClusterPair* newCluster = new ClusterPair((*it1), a->thisNode);
			priority.push(newCluster);
			
			if(printDebugOutput)
				printf("Remaining Items = %08i\r", lights->size()-priority.size());
		}
		bool isVal = true;
		debugOutputPriorityQueue(priority);
		
		if(printDebugOutput)
			printf("\ndone.\n\nClustering...\n");

		while(int(priority.size())>1) 
		{
			int a = priority.size();
			
			if(printDebugOutput)
				printf("Remaining Items = %08i\r", a);

			ClusterPair* topPair = priority.top();
			priority.pop();
			KD_Node* firstCluster = kd_tree.findNode(topPair->A, topPair->A->ID, kd_tree.root, 0, searchHasEnded);
			(*searchHasEnded) = false;
			isVal = kd_tree.isValid(topPair->A);
			if (!isVal)
			{
				delete topPair;
				topPair = 0;
				continue;
			}
			KD_Node* secondCluster = kd_tree.findNode(topPair->B, topPair->B->ID, kd_tree.root, 0, searchHasEnded);
			(*searchHasEnded) = false;
			isVal = kd_tree.isValid(topPair->B);
			if (!isVal)
			{
				KD_Node* searchResult = kd_tree.findNode(topPair->A, topPair->A->ID, kd_tree.root, 0, searchHasEnded);
				(*searchHasEnded) = false;
				KD_Node* newClosest = kd_tree.findClosestNeighbour(searchResult, kd_tree.root);
				ClusterPair* newCluster = new ClusterPair(topPair->A, newClosest->thisNode);
				delete topPair;
				topPair = 0;
				priority.push(newCluster);
				continue;
			}

			//both are valid, merge them:
			Cluster* merged = mergeClusters(topPair->A, topPair->B);

			KD_Node* mergedNode = kd_tree.updateTree(firstCluster, secondCluster, merged);
			root = merged;

			KD_Node* newClosest = kd_tree.findClosestNeighbour(mergedNode, kd_tree.root);
			if (newClosest!=0)
			{
				ClusterPair* newCluster = new ClusterPair(merged, newClosest->thisNode);
				priority.push(newCluster);
			}

			delete topPair;
			topPair = 0;
		}

		ClusterPair* topPair = priority.top();
	//	Cluster* merged = mergeClusters(topPair->A, topPair->B);
		delete topPair;
		topPair = 0;
	//	root = merged;
		delete searchHasEnded;
		searchHasEnded = 0;

	}
	debugOutputTreeToFile(root);


	const double dTimePerFrame = double(timTimer);

	timTimer.start();
	if(printDebugOutput)
	{
		printf("Remaining Items = %08i\r", priority.size());
		printf("\nDone (%f sek). Anzahl findClosestDown = %i \n\n", dTimePerFrame, anzFindClosestDown);
	}
		
	
}

void CpuLightTree::debugOutputTreeToFile(Cluster* node)
{
	FILE* file;
	fopen_s(&file, "lightTree.tree", "wb");
	recursiveFileOutput(node, file);
	fclose(file);
}

void CpuLightTree::recursiveFileOutput(Cluster* node, FILE* file)
{
	if (node == 0)
		return;
	fprintf(file, "\n");
	fprintf(file, "posX = %f; posY = %f; posZ = %f; intensity = %f\n", node->representativeLight->position.x,
												node->representativeLight->position.y,
												node->representativeLight->position.z,
												node->intensity);
	recursiveFileOutput(node->leftChild, file);
	recursiveFileOutput(node->rightChild, file);
	fprintf(file, "\n\n");
}

float KD_Tree::sumOfIntensities(KD_Node* node, float value)
{
	if (node==0)
	{
		return value;
	}
	value+=node->thisNode->intensity;

	value+=sumOfIntensities(node->leftChild, 0);
	value+=sumOfIntensities(node->rightChild, 0);
	return value;
}

void KD_Tree::printValids(KD_Node* node)
{
	if (node==0)
	{
		return;
	}
	if (node->isValid)
		printf("posX = %f, posY = %f, posZ = %f, intensity = %f\n", node->thisNode->representativeLight->position.x, 
		node->thisNode->representativeLight->position.y,
		node->thisNode->representativeLight->position.z,
		node->thisNode->intensity);
	printValids(node->leftChild);
	printValids(node->rightChild);
	return;
}

void CpuLightTree::recursiveDeleteCluster(Cluster* cluster)
{
	if (cluster == 0)
		return;

	/*
	if(cluster->representativeLight == 0)
		return;
		*/

	recursiveDeleteCluster(cluster->leftChild);
	recursiveDeleteCluster(cluster->rightChild);

	delete cluster;
	cluster = 0;
}


/*
 * Assertion: root!=0
 */
void CpuLightTree::debugOutput(Cluster* node)
{
	if (node==0)
		return;
	printf("\n");
	printf("Intensity = %f ", node->intensity);
	node->representativeLight->debugOutput();
	
	
	debugOutput(node->rightChild);
	debugOutput(node->leftChild);
	printf("\n\n");
}

void CpuLightTree::debugOutputPriorityQueue(priority_queue<ClusterPair*, vector<ClusterPair*>, ComparePtrToA> someQueue)
{
	priority_queue<ClusterPair*, vector<ClusterPair*>, ComparePtrToA> someOtherQueue = priority_queue<ClusterPair*, vector<ClusterPair*>, ComparePtrToA>(someQueue);
	int i = 0;

	FILE* file;
	fopen_s(&file, "priority.txt", "wb");

	

	while(someOtherQueue.size()>0)
	{
		i++;
		ClusterPair* top = someOtherQueue.top();
		someOtherQueue.pop();
		fprintf(file, "\n%i: %f, %f, %f || %f, %f, %f| %f", i, top->A->representativeLight->position.x,
													top->A->representativeLight->position.y,
													top->A->representativeLight->position.z,

													top->B->representativeLight->position.x,
													top->B->representativeLight->position.y,
													top->B->representativeLight->position.z, top->distance);
	}
	fclose(file);
	
}



void doSomeDebugging()
{
	CpuLightTree* test3 = new CpuLightTree();
	
	deque <Cluster*> lightDeque;
	vector<VPL*> vplVector;
	
		for (int i = 1; i < 17; i++)
	{	
		VPL* newVPL = new VPL();
		float iSquare = pow(float(i), 2.0f);
	//	newVPL->position = Point(iSquare,iSquare,iSquare);
		newVPL->position = Point(1,1,1);
		newVPL->color[0] = 0.0f;
		newVPL->color[1] = 0.0f;
		newVPL->color[2] = 0.0f;
	//	newVPL->color[3] = iSquare;
		newVPL->color[3] = 1;

		Cluster* someCluster = new Cluster(newVPL);

		vplVector.push_back(newVPL);

		lightDeque.push_back(someCluster);	
	}

	deque<Cluster*>::iterator it;

	for (it = lightDeque.begin(); it < lightDeque.end(); it++)
	{
		test3->addCluster(*it);
	}

	test3->buildClusters2();
	
//	test3->debugOutput(test3->root);

	delete test3;
	test3 = 0;

	vector<VPL*>::iterator it2;
	for (it2 = vplVector.begin(); it2 < vplVector.end(); it2++)
	{
		delete (*it2);
		(*it2) = 0;
	}
}

float calculateClusterSize(Cluster* a, Cluster* b)
{
	Bounds newBounds = calcNewBounds(a->bounds, b->bounds);
	float lengthOfDiagonal = (newBounds.B- newBounds.A).getLength();
	//return (aCluster->intensity * pow(lengthOfDiagonal,2.0f));
	
//	#pragma message(__LOC__"Check metric!!!")

//	return lengthOfDiagonal;

	float3 col1 = Color::rgb2lch(a->representativeLight->color);
	float3 col2 = Color::rgb2lch(b->representativeLight->color);

	float colorDist = Color::colorDistOmittingLuminance_LCH(col1, col2);
	
	float clusterSize = colorDist * lengthOfDiagonal + lengthOfDiagonal;

//	return clusterSize;

//	printf("clusterDist = %f\n", (1.0f / 3.0f) * ( (lengthOfDiagonal/Scene::sceneBboxDiagonal) + colorDist + (0.5f - 0.5f*(a->representativeLight->normal*b->representativeLight->normal))));
	return (1.0f / 3.0f) * ( (lengthOfDiagonal/Scene::sceneBboxDiagonal) + colorDist + (0.5f - 0.5f*(a->representativeLight->normal*b->representativeLight->normal)));

	return clusterSize;

	return lengthOfDiagonal + Scene::sceneBboxDiagonal*(0.5f - 0.5f*(a->representativeLight->normal*b->representativeLight->normal));
	
	
}

ClusterPair::ClusterPair(Cluster* _A, Cluster* _B)
{
	A = _A;
	B = _B;
	distance = calculateClusterSize(A, B);
//	distance = calculateClusterSize(A->bounds, B->bounds, A->intensity+B->intensity);
}

ClusterPair::~ClusterPair(void)
{
	//delete A;
	//A = 0;
	//
	//delete B;
	//B = 0;
}
bool KD_Tree::isValid(Cluster* a)
{
	bool* searchHasEnded = new bool();
	(*searchHasEnded) = false;
	KD_Node* searchResult = findNode(a, a->ID, root, 0, searchHasEnded);

	delete searchHasEnded;
	searchHasEnded = 0;

	if (searchResult->isValid == false)
		return false;
	else
	{
		if (searchResult->thisNode->intensity == a->intensity)
			return true;
		//else return false
	}

	return false; 
}

void KD_Tree::buildTree(deque<Cluster*>* data)
{
	root = new KD_Node();
	root->separationAxis = 0;

	recursiveFillTree(root, 0, data, 0);
	
	debugOutputTreeToFile(root);
}

/*
 * @axis: 0 == X, 1 == Y, 2 == Z
 */ 
void KD_Tree::recursiveFillTree(KD_Node* node, KD_Node* parent, deque<Cluster*>* data, int axis)
{
	int nextAxis = (axis+1)%3;
	deque<Cluster*>::iterator searchMedianIt;

	if (int(data->size()) == 0)
	{
		return;
	}


	deque<Cluster*>* allToInsert = new deque<Cluster*>;
	copy(data->begin(), data->end(), back_inserter(*allToInsert));
	//ACHTUNG: copy kopiert so: [a,b) ->....!!!! b muss also eins weiter gehen!!
	switch (axis)
	{
		case 0: {sort(allToInsert->begin(), allToInsert->end(), SortX());break;}
		case 1: {sort(allToInsert->begin(), allToInsert->end(), SortY());break;}
		case 2: {sort(allToInsert->begin(), allToInsert->end(), SortZ());break;}
		default: break;
	}

	searchMedianIt = allToInsert->begin();
	for (int i = 0; i < int(allToInsert->size()/2.0f); i++)
	{
		searchMedianIt++;
	}

	node->thisNode = (*searchMedianIt);
	node->parent = parent;
	

	deque<Cluster*>* leftInsert = new deque<Cluster*>;
	deque<Cluster*>* rightInsert = new deque<Cluster*>;
	
	copy(allToInsert->begin(), searchMedianIt, back_inserter(*leftInsert));
	searchMedianIt++;	
	copy(searchMedianIt, allToInsert->end(), back_inserter(*rightInsert));
	
	if (int(leftInsert->size()) > 0)
	{
		KD_Node* leftChild = new KD_Node();
		leftChild->separationAxis = nextAxis;
		node->leftChild = leftChild;

		recursiveFillTree(leftChild, node, leftInsert, nextAxis);	
		
	}
	if (int(rightInsert->size()) > 0)
	{
		//weitere rechte rekursionen:
		KD_Node* rightChild = new KD_Node();
		rightChild->separationAxis = nextAxis;
		node->rightChild = rightChild;

		recursiveFillTree(rightChild, node, rightInsert, nextAxis);
	}
/*
	searchMedianIt = allToInsert->begin();
	for (int i = 0; i < int(allToInsert->size()); i++)
	{
		delete (*searchMedianIt);
	}
	*/
	allToInsert->clear();
	delete allToInsert;
	allToInsert = 0;

	leftInsert->clear();
	delete leftInsert;
	leftInsert = 0;

	rightInsert->clear();
	delete rightInsert;
	rightInsert = 0;


}

/*
 * Assertion: fixNode is valid!
 */
KD_Node* KD_Tree::findClosestNeighbour(KD_Node* fixNode, KD_Node* currentSearchLoc)
{
	bool* searchHasEnded = new bool();
	(*searchHasEnded) = false;
	//KD_Node* leafPos = findNode(fixNode->thisNode, currentSearchLoc, 0, searchHasEnded);
	KD_Node* leafPos = fixNode;
	//Cluster* hypotheticalCluster = mergeClusters(fixNode->thisNode, leafPos->thisNode);
	//float dist = calculateClusterSize(hypotheticalCluster);

	CurrentBest currentBest;
	currentBest.bestMatch = 0;//currentSearchLoc;
	currentBest.upperBound = 9999999999.0f;
	currentBest.size = 9999999999.0f;

	if (leafPos->leftChild!=0)
		currentBest = findClosestDown(fixNode, leafPos->leftChild, currentBest);
	if (leafPos->rightChild!=0)
		currentBest = findClosestDown(fixNode, leafPos->rightChild, currentBest);

	if (leafPos->parent!=0)//sonst hauts ihn um wenn man nach dem nearest neighbour vom root sucht...
		currentBest = findClosestUp(fixNode, leafPos->parent, leafPos, currentBest);

	delete searchHasEnded;
	searchHasEnded = 0;
	return currentBest.bestMatch;
}

CurrentBest KD_Tree::findClosestDown(KD_Node* fixNode, KD_Node* currentSearchLoc, CurrentBest currentBest)
{
	anzFindClosestDown++;
	//if(anzFindClosestDown % 100000 == 0)
	//	cout << endl << anzFindClosestDown;

	if (currentSearchLoc == 0 || currentSearchLoc->thisNode == 0)
	{
		return currentBest;
	}

	if (currentSearchLoc->isValid)
	{
		float dist = calculateClusterSize(fixNode->thisNode, currentSearchLoc->thisNode);
		if (dist < currentBest.size)
		{
			currentBest.bestMatch = currentSearchLoc;
			currentBest.upperBound = calculateUpperBound(fixNode->thisNode, currentSearchLoc->thisNode);
			currentBest.size = dist;
		}
	}

	float valueOfAxisFixNode;
	float valueOfAxisSearchLoc;

	if(currentSearchLoc->separationAxis==0)
	{
		valueOfAxisFixNode = fixNode->thisNode->representativeLight->position.x;
		valueOfAxisSearchLoc = currentSearchLoc->thisNode->representativeLight->position.x;
	}
	else if(currentSearchLoc->separationAxis==1)
	{
		valueOfAxisFixNode = fixNode->thisNode->representativeLight->position.y;
		valueOfAxisSearchLoc = currentSearchLoc->thisNode->representativeLight->position.y;
	}
	else if(currentSearchLoc->separationAxis==2)
	{
		valueOfAxisFixNode = fixNode->thisNode->representativeLight->position.z;
		valueOfAxisSearchLoc = currentSearchLoc->thisNode->representativeLight->position.z;
	}

	if (currentSearchLoc->leftChild != 0 && currentSearchLoc->leftChild->thisNode!=0)
	{
		//float dist= calculateClusterSize(hypotheticalCluster2);
		
		float distSearchToFixOnAxis = abs(valueOfAxisFixNode - valueOfAxisSearchLoc);

//IFFF

		if (distSearchToFixOnAxis < currentBest.upperBound || valueOfAxisSearchLoc > valueOfAxisFixNode)
		{
			currentBest = findClosestDown(fixNode, currentSearchLoc->leftChild, currentBest);
		}
	}
	if (currentSearchLoc->rightChild != 0 && currentSearchLoc->rightChild->thisNode!=0)
	{
//		float dist = calculateClusterSize(fixNode->thisNode, currentSearchLoc->thisNode);
		float distSearchToFixOnAxis = abs(valueOfAxisFixNode - valueOfAxisSearchLoc);

		if (distSearchToFixOnAxis < currentBest.upperBound || valueOfAxisSearchLoc < valueOfAxisFixNode)
		{
			currentBest = findClosestDown(fixNode, currentSearchLoc->rightChild, currentBest);
		}
	}

	return currentBest;
}


CurrentBest KD_Tree::findClosestUp(KD_Node* fixNode, KD_Node* currentSearchLoc, KD_Node* previousSearchLoc, CurrentBest currentBest)
{
	if (currentSearchLoc->isValid)
	{
		float dist = calculateClusterSize(fixNode->thisNode, currentSearchLoc->thisNode);
		if (dist < currentBest.size)
		{
			currentBest.bestMatch = currentSearchLoc;
			currentBest.upperBound = calculateUpperBound(fixNode->thisNode, currentSearchLoc->thisNode);
			currentBest.size = dist;
		}
	}
	if (currentSearchLoc->leftChild == previousSearchLoc) //TODO check ob man hier dereferenzieren muss
	{
		currentBest = findClosestDown(fixNode, currentSearchLoc->rightChild, currentBest);
	}
	else if (currentSearchLoc->rightChild == previousSearchLoc)
	{
		currentBest = findClosestDown(fixNode, currentSearchLoc->leftChild, currentBest);
	}
	else
	{
		printf("Error merging clusters!!! !\n");
		printf("Hit Enter to exit\n");
		char* a = new char[32];
		cin.getline(a, 32);
		exit(-1);
	}

	if (currentSearchLoc->parent==0) //-->root
	{
	}
	else
	{
		currentBest = findClosestUp(fixNode, currentSearchLoc->parent, currentSearchLoc, currentBest);
	}

	return currentBest;
}

/*
 * Returns the KD_Node containing the Cluster* searchThis, or, if that
 * node doesn't exist, returns the node at the location that searchThis would
 * be if it did exist.
 */
KD_Node* KD_Tree::findNode(Cluster* searchThis, int ID, KD_Node* currentSearchLoc, int axis, bool* searchEnded)
{
	if (currentSearchLoc==0)
	{
		KD_Node* empty = new KD_Node();	
		return empty;//currentSearchLoc;
	}

	if ((currentSearchLoc->thisNode)->ID == ID &&  (*(searchThis) == *(currentSearchLoc->thisNode)))
	{
		(*searchEnded) = true;
		return currentSearchLoc;
	}



	int directionToGo = 0;
	int nextAxis = (axis+1)%3;
	KD_Node* retVal;
	switch (axis)
	{
	case 0:
		{
			if (searchThis->representativeLight->position.x == currentSearchLoc->thisNode->representativeLight->position.x)
			{
				directionToGo = 3; //go both right and left
			}
			else
			{
				if (searchThis->representativeLight->position.x < currentSearchLoc->thisNode->representativeLight->position.x)
					directionToGo = 1; //go left
				else
					directionToGo = 2; //go right
			}
			break;
		}
		case 1:
		{
			if (searchThis->representativeLight->position.y == currentSearchLoc->thisNode->representativeLight->position.y)
			{
				directionToGo = 3; //go both right and left
			}
			else
			{
				if (searchThis->representativeLight->position.y < currentSearchLoc->thisNode->representativeLight->position.y)
					directionToGo = 1; //go left
				else
					directionToGo = 2; //go right
			}
			break;
		}
		case 2:
		{
			if (searchThis->representativeLight->position.z == currentSearchLoc->thisNode->representativeLight->position.z)
			{
				directionToGo = 3; //go both right and left
			}
			else
			{
				if (searchThis->representativeLight->position.z < currentSearchLoc->thisNode->representativeLight->position.z)
					directionToGo = 1; //go left
				else
					directionToGo = 2; //go right
			}
			break;
		}
		default: break;
	}
	switch (directionToGo)
	{
	case 1: 
	{
		if (currentSearchLoc->leftChild!=0)
		{
			retVal = findNode(searchThis, ID, currentSearchLoc->leftChild, nextAxis, searchEnded);
		}
		break;
	}
	case 2: 
	{
		if (currentSearchLoc->rightChild!=0)
		{
			retVal = findNode(searchThis, ID, currentSearchLoc->rightChild, nextAxis, searchEnded);
		}
		break;
	}
	case 3: 
	{
		if (currentSearchLoc->leftChild!=0 && currentSearchLoc->leftChild->thisNode!=0)
		{
			retVal = findNode(searchThis, ID, currentSearchLoc->leftChild, nextAxis, searchEnded);
		}
		if (!(*searchEnded))
		{
			if (currentSearchLoc->rightChild!=0 && currentSearchLoc->rightChild->thisNode!=0)
			{
				retVal = findNode(searchThis, ID, currentSearchLoc->rightChild, nextAxis, searchEnded);
			}
		}
		break;
	}
	default: return 0; //this really shouldn't happen 
	}
	return retVal;
}


/*
 * Updates the kd-tree  (I) by merging the clusters to a singular one 
 * and (II) by invalidating the original clusters.
 * Returns a pointer to the merged Node.
 *
 */
KD_Node* KD_Tree::updateTree(KD_Node* nodeA, KD_Node* nodeB, Cluster* mergedCluster)
{
	KD_Node* retVal;
	if ((nodeA->thisNode)->representativeLight->color[3]>(nodeB->thisNode)->representativeLight->color[3])
	{
		nodeB->isValid = false;
		nodeA->thisNode = mergedCluster; 
		retVal = nodeA;
	}
	else
	{
		nodeA->isValid = false;
		nodeB->thisNode = mergedCluster;
		retVal = nodeB;
	}
	return retVal;

}

void KD_Tree::debugOutputTreeToFile(KD_Node* node)
{
	FILE* file;
	fopen_s(&file, "kdTree.tree", "wb");
	recursiveFileOutput(node, file);
	fclose(file);
}

void KD_Tree::recursiveFileOutput(KD_Node* node, FILE* file)
{
	if (node == 0 || node->thisNode == 0)
		return;
	fprintf(file, "\n");
	fprintf(file, "valid = %i, posX = %f; posY = %f; posZ = %f; intensity = %f\n", int(node->isValid),(node->thisNode)->representativeLight->position.x,
												(node->thisNode)->representativeLight->position.y,
												(node->thisNode)->representativeLight->position.z,
												(node->thisNode)->intensity);
	recursiveFileOutput(node->leftChild, file);
	recursiveFileOutput(node->rightChild, file);
	fprintf(file, "\n\n");
}
void KD_Tree::recursiveDelete(KD_Node* node)
{
	if (node == 0 || node->thisNode == 0)
		return;
	recursiveDelete(node->leftChild);
	recursiveDelete(node->rightChild);

	delete node;
	node = 0;
}
void KD_Tree::debugOutputTree(KD_Node* node)
{
	if (node == 0)
		return;
	printf("\n");
	printf("posX = %f; posY = %f; posZ = %f\n", (node->thisNode)->representativeLight->position.x,
												(node->thisNode)->representativeLight->position.y,
												(node->thisNode)->representativeLight->position.z
												);
	debugOutputTree(node->leftChild);
	debugOutputTree(node->rightChild);
	printf("\n\n");
}

KD_Node::KD_Node()
{
	leftChild = 0;
	rightChild = 0;
	parent = 0;
	thisNode = 0;
	isValid = true;
	maximumBoundingBox.A.x = maximumBoundingBox.A.y = maximumBoundingBox.A.z = -9999999999.9f;
	maximumBoundingBox.B.x = maximumBoundingBox.B.y = maximumBoundingBox.B.z = 9999999999.9f;
	minimumIntensity = 999999999.9f;

}

Cluster* mergeClusters(Cluster* a, Cluster* b)
{
	Cluster* retVal = new Cluster();
	if (a->representativeLight->color[3]>b->representativeLight->color[3])
	{
		retVal->representativeLight = a->representativeLight;
		retVal->ID = a->ID;
	}
	else
	{
		retVal->representativeLight = b->representativeLight;
		retVal->ID = b->ID;
	}
	
	retVal->intensity = a->intensity + b->intensity;
	retVal->leftChild = a;
	retVal->rightChild = b;

//	ClusterInCUDA* cudaCluster = new ClusterInCUDA;
//	VPL_InCUDA* cudaVPL = new VPL_InCUDA;

	

//	CUDA_LightCuts::clustersInCUDA[anzClusters] = *cudaCluster;
//	CUDA_LightCuts::representativeLights[anzClusters] = *cudaVPL;

	retVal->bounds = calcNewBounds(a->bounds, b->bounds);
	retVal->sameOrientation = a->sameOrientation && b->sameOrientation && (a->representativeLight->normal*b->representativeLight->normal > 0.9999999);

	setCUDA_ClusterProperties(retVal);
	anzClusters++;

	return retVal;
}

Bounds calcNewBounds(Bounds a, Bounds b)
{
	Bounds c;

	c.A.x = min(a.A.x, b.A.x);
	c.A.y = min(a.A.y, b.A.y);
	c.A.z = min(a.A.z, b.A.z);

	c.B.x = max(a.B.x, b.B.x);
	c.B.y = max(a.B.y, b.B.y);
	c.B.z = max(a.B.z, b.B.z);
	
	return c;
}

float calculateUpperBound(Cluster* a, Cluster* b)
{	
//	Cluster* hypotheticalCluster = mergeClusters(a, b);
	//float alphaCDsquare = calculateClusterSize(hypotheticalCluster) / hypotheticalCluster->intensity;
	
	Bounds newBounds = calcNewBounds(a->bounds, b->bounds);
	float lengthOfDiagonal = (newBounds.B- newBounds.A).getLength();

	return 3.0f * lengthOfDiagonal;
/*
	Bounds bounds;

	float u, v, w;
	
	u = sqrt (alphaCDsquare - pow((a->bounds.B.y - a->bounds.A.y),2.0f) - pow((a->bounds.B.z - a->bounds.A.z),2.0f));
	v = sqrt (alphaCDsquare - pow((a->bounds.B.z - a->bounds.A.z),2.0f) - pow((a->bounds.B.x - a->bounds.A.x),2.0f));
	w = sqrt (alphaCDsquare - pow((a->bounds.B.x - a->bounds.A.x),2.0f) - pow((a->bounds.B.y - a->bounds.A.y),2.0f));

	bounds.A.x = a->bounds.B.x - u;
	bounds.A.y = a->bounds.B.y - v;
	bounds.A.z = a->bounds.B.z - w;

	bounds.B.x = a->bounds.A.x + u;
	bounds.B.y = a->bounds.A.y + v;
	bounds.B.z = a->bounds.A.z + w;

	Vector diagonal = bounds.B - bounds.A;
	return pow(diagonal.getLength(), 2.0f);
	*/

}

void setCUDA_ClusterProperties(Cluster* original)
{
	original->representativeLight->index = anzClusters;
	CUDA_LightCuts::positionOfVPL[anzClusters] = make_float4(original->representativeLight->position.x,original->representativeLight->position.y,original->representativeLight->position.z, -1.0f);
	CUDA_LightCuts::normalOfVPL_IntensityRatio[anzClusters] = make_float4(original->representativeLight->normal.x,original->representativeLight->normal.y,original->representativeLight->normal.z, original->representativeLight->intensityRatio);
	CUDA_LightCuts::colorOfVPL[anzClusters] = make_float4(original->representativeLight->color[0],original->representativeLight->color[1],original->representativeLight->color[2], original->intensity);
	CUDA_LightCuts::shadowMapAndLightSpaceOfVPL[anzClusters] = make_int4(original->representativeLight->shadowMapAtPixelX, original->representativeLight->shadowMapAtPixelY, original->representativeLight->lightSpaceMatrixAtPixelX, original->representativeLight->lightSpaceMatrixAtPixelY);

	CUDA_LightCuts::indices[anzClusters] = make_int4(-1, -1, -1, -1); //indices.x = leftChild, indices.y = rightChild, indices.z = index of representative Light (in datastructure above!)
	if (original->leftChild!=0)
	{
		CUDA_LightCuts::indices[anzClusters].x = original->leftChild->clusterNumber;;
	}
	if (original->rightChild!=0)
	{
		CUDA_LightCuts::indices[anzClusters].y = original->rightChild->clusterNumber;
	}
	CUDA_LightCuts::indices[anzClusters].z = anzClusters; //not really necessary...

	CUDA_LightCuts::bounds_A[anzClusters] = make_float4(original->bounds.A.x,original-> bounds.A.y, original->bounds.A.z, 0.0f);
	CUDA_LightCuts::bounds_B[anzClusters] = make_float4(original->bounds.B.x,original-> bounds.B.y, original->bounds.B.z, 0.0f);

	float3 lch_color = Color::rgb2lch(make_float3(original->representativeLight->color[0], original->representativeLight->color[1], original->representativeLight->color[2]));
	CUDA_LightCuts::color_lch[anzClusters] = make_float4(lch_color.x, lch_color.y, lch_color.z, -1.0f);
	CUDA_LightCuts::estRad_RadUpperBound_errorBound[anzClusters] = make_float4(original->estimatedRadiance, original->radianceUpperBound, original->errorBound, -1.0f);

	/*
	cudaCluster->bounds_A = make_float4(original->bounds.A.x,original-> bounds.A.y, original->bounds.A.z, 0.0f);
	cudaCluster->bounds_B = make_float4(original->bounds.B.x, original->bounds.B.y, original->bounds.B.z, 0.0f);

	if (original->leftChild!=0 && original->rightChild!=0)
	{
		cudaCluster->leftChild = original->leftChild->clusterNumber;
		cudaCluster->rightChild = original->rightChild->clusterNumber;
	}
	else
	{
		cudaCluster->leftChild = -1;
		cudaCluster->rightChild = -1;
	}


	cudaCluster->intensity = original->intensity;

	cudaCluster->materialTerm = original->materialTerm;
	cudaCluster->geometryTerm = original->geometryTerm;
	cudaCluster->visibilityTerm = original->visibilityTerm;
	cudaCluster->radianceUpperBound = original->radianceUpperBound;
	cudaCluster->estimatedRadiance = original->estimatedRadiance;
	cudaCluster->errorBound = original->errorBound;

	cudaVPL->positionX = original->representativeLight->position.x;
	cudaVPL->positionY = original->representativeLight->position.y;
	cudaVPL->positionZ = original->representativeLight->position.z;

	cudaVPL->normalX = original->representativeLight->normal.x;
	cudaVPL->normalY = original->representativeLight->normal.y;
	cudaVPL->normalZ = original->representativeLight->normal.z;

	cudaVPL->colorR = original->representativeLight->color[0];
	cudaVPL->colorG = original->representativeLight->color[1];
	cudaVPL->colorB = original->representativeLight->color[2];
	cudaVPL->colorW = original->representativeLight->color[3];

	cudaVPL->intensityRatio = original->representativeLight->intensityRatio;

	cudaVPL->shadowMapAtPixelX = original->representativeLight->shadowMapAtPixelX;
	cudaVPL->shadowMapAtPixelY = original->representativeLight->shadowMapAtPixelY;
	cudaVPL->lightSpaceMatrixAtPixelX = original->representativeLight->lightSpaceMatrixAtPixelX;
	cudaVPL->lightSpaceMatrixAtPixelY = original->representativeLight->lightSpaceMatrixAtPixelY;
*/
}
