
//CLASS Vector

//objects of this class provide functionality of three-dim. vectors.
//author: magic carpet group

//for performance reasons attributes (coords) have public visibility

#include <math.h>
#include "Vector.h"


//std-constructor. creates null vector object
Vector::Vector()
{
	x=0.0f; y=0.0f; z=0.0f;
}

//constructor. creates vector from float coords
Vector::Vector(float _x, float _y, float _z)
{
	x=_x; y=_y; z=_z;
}

//constructor. creates vector from float coords
Vector::Vector(float f[])
{
	x=f[0]; y=f[1]; z=f[2];
}


Vector Vector::rotateX(Vector v, float angle)
{
	Vector result;
	result.x = v.x;
	result.y = v.y*cos(angle) - v.z*sin(angle);
	result.z = v.y*sin(angle) + v.z*cos(angle);
	result.normalize();
	return result;
}


Vector Vector::rotateY(Vector v, float angle)
{
	Vector result;
	result.x = v.x*cos(angle)+v.z*sin(angle);
	result.y = v.y;
	result.z = v.x*(-sin(angle))+v.z*cos(angle);
	result.normalize();
	return result;

}

Vector Vector::rotateZ(Vector v, float angle)
{
	Vector result;
	result.x = v.x*cos(angle)-v.y*sin(angle);
	result.y = v.x*sin(angle)+v.y*cos(angle);
	result.z = v.z;
	result.normalize();
	return result;

}


//returns length of this vector (pythagoras)
float Vector::getLength(void)
{
	return sqrt(x*x+y*y+z*z);
}

//normalizes this vector.
//post: this.getLength = 1
void Vector::normalize(void)
{
	float l=this->getLength();
	if(l==0)
		return;
	x/=l;
	y/=l;
	z/=l;
}

bool Vector::operator==(const Vector& v)
{
	return (  (x == v.x) && (y == v.y) && (z == v.z));
}


//Overloads assertion operator
//Enables the use of "=" for Vector-objects
Vector& Vector::operator= (const Vector& v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	return *this;
}





//Global operator overloading of "+" operator for 2 Vectors
//Returns sum of 2 Vectors as a new Vector
Vector operator+ (Vector v1, Vector v2)
{
	Vector result;
	result.x = v1.x + v2.x;
	result.y = v1.y + v2.y;
	result.z = v1.z + v2.z;
	return result;
}
	
//Global operator overloading of "-" operator for 2 Vectors
//Returns difference of 2 Vectors as a new Vector
Vector operator- (Vector v1, Vector v2)
{
	Vector result;
	result.x = v1.x - v2.x;
	result.y = v1.y - v2.y;
	result.z = v1.z - v2.z;
	return result;
}

//Global operator overloading of "*" operator for 2 Vectors
//Returns dot product of 2 Vectors as a float number
float operator* (Vector v1, Vector v2)
{
	float result;
	result = v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
	return result;
}

//Global operator overloading of "*" operator for Vector with a scalar
//Returns given Vector scaled by given float number
Vector operator* (Vector v, float f)
{
	v.x *= f;
	v.y *= f;
	v.z *= f;
	return v;
}

//Provides cross-product operation.
//Returns othogonal vector to 2 given Vectors in threedim. space
Vector crossProduct(Vector v1, Vector v2)
{
	Vector result;
	result.x = v1.y*v2.z - v1.z*v2.y;
	result.y = v1.z*v2.x - v1.x*v2.z;
	result.z = v1.x*v2.y - v1.y*v2.x;
	return result;
}

//to-String analogon: Enables output of Vectors with cout command
ostream& operator<< (ostream& os, const Vector& v)
{
	return os<<"( " << v.x << " / " << v.y << " / " << v.z << " )";
}