
//CLASS Vector

//objects of this class provide functionality of three-dim. vectors.
//author: magic carpet group

//for performance reasons attributes (coords) have public visibility

#include <math.h>
#include <iostream>

using namespace std;

#ifndef _Vector
#define _Vector

#define PI 3.141592654

class Vector
{
public:
	float x,y,z; //coordinates


	//std-constructor. creates null vector object
	Vector(void);

	//constructor. creates vector from float coords
	Vector(float _x, float _y, float _z);	

	//constructor. creates vector from float coords
	Vector(float f[]);

	//rotates *this 
	static Vector rotateX(Vector v, float angle);
	static Vector rotateY(Vector v, float angle);
	static Vector rotateZ(Vector v, float angle);


	//returns length of this vector (pythagoras)
	float getLength(void);

	//normalizes this vector.
	//post: this.getLength = 1
	void normalize(void);

	//Overloads assertion operator
	//Enables the use of "=" for Vector-objects
	Vector& operator= (const Vector& v);

	//Enables the use of "==" for Vector-objects
	bool operator==(const Vector& v);

};

//Global operator overloading of "+" operator for 2 Vectors
//Returns sum of 2 Vectors as a new Vector
Vector operator+ (Vector v1, Vector v2);

//Global operator overloading of "-" operator for 2 Vectors
//Returns difference of 2 Vectors as a new Vector
Vector operator- (Vector v1, Vector v2);

//Global operator overloading of "*" operator for 2 Vectors
//Returns dot product of 2 Vectors as a float number
float operator* (Vector v1, Vector v2);

//Global operator overloading of "*" operator for Vector with a scalar
//Returns given Vector scaled by given float number
Vector operator* (Vector v, float f);

//Provides cross-product operation.
//Returns othogonal vector to 2 given Vectors in threedim. space
Vector crossProduct(Vector v1, Vector v2);

//to-String analogon: Enables output of Vectors with cout command
ostream& operator<< (ostream& os, const Vector& v);

#endif