/******************************************************************************

  This source code accompanies the Journal of Graphics Tools paper:

  "Fast Ray / Axis-Aligned Bounding Box Overlap Tests using Ray Slopes" 
  by Martin Eisemann, Thorsten Grosch, Stefan Müller and Marcus Magnor
  Computer Graphics Lab, TU Braunschweig, Germany and
  University of Koblenz-Landau, Germany
  
  This source code is public domain, but please mention us if you use it.

******************************************************************************/
#include "slope.h"


bool slope(ray_2 *r, aabox *b)
{
	bool retVal = false;
    switch (r->classification)
    {
	case MMM:
		
		if ((r->x < b->x0) || (r->y < b->y0) || (r->z < b->z0)
			|| (r->jbyi * b->x0 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x1 + r->c_yx > 0)
			|| (r->jbyk * b->z0 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z1 + r->c_yz > 0)
			|| (r->kbyi * b->x0 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;
		
		

	case MMP:
		
		if ((r->x < b->x0) || (r->y < b->y0) || (r->z > b->z1)
			|| (r->jbyi * b->x0 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x1 + r->c_yx > 0)
			|| (r->jbyk * b->z1 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z0 + r->c_yz < 0)
			|| (r->kbyi * b->x0 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MPM:
		
		if ((r->x < b->x0) || (r->y > b->y1) || (r->z < b->z0)
			|| (r->jbyi * b->x0 - b->y0 + r->c_xy < 0) 
			|| (r->ibyj * b->y1 - b->x1 + r->c_yx > 0)
			|| (r->jbyk * b->z0 - b->y0 + r->c_zy < 0) 
			|| (r->kbyj * b->y1 - b->z1 + r->c_yz > 0)
			|| (r->kbyi * b->x0 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MPP:
	
		if ((r->x < b->x0) || (r->y > b->y1) || (r->z > b->z1)
			|| (r->jbyi * b->x0 - b->y0 + r->c_xy < 0) 
			|| (r->ibyj * b->y1 - b->x1 + r->c_yx > 0)
			|| (r->jbyk * b->z1 - b->y0 + r->c_zy < 0)
			|| (r->kbyj * b->y1 - b->z0 + r->c_yz < 0)
			|| (r->kbyi * b->x0 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PMM:

		if ((r->x > b->x1) || (r->y < b->y0) || (r->z < b->z0)
			|| (r->jbyi * b->x1 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x0 + r->c_yx < 0)
			|| (r->jbyk * b->z0 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z1 + r->c_yz > 0)
			|| (r->kbyi * b->x1 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PMP:

		if ((r->x > b->x1) || (r->y < b->y0) || (r->z > b->z1)
			|| (r->jbyi * b->x1 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x0 + r->c_yx < 0)
			|| (r->jbyk * b->z1 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z0 + r->c_yz < 0)
			|| (r->kbyi * b->x1 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PPM:

		if ((r->x > b->x1) || (r->y > b->y1) || (r->z < b->z0)
			|| (r->jbyi * b->x1 - b->y0 + r->c_xy < 0)
			|| (r->ibyj * b->y1 - b->x0 + r->c_yx < 0)
			|| (r->jbyk * b->z0 - b->y0 + r->c_zy < 0) 
			|| (r->kbyj * b->y1 - b->z1 + r->c_yz > 0)
			|| (r->kbyi * b->x1 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PPP:

		if ((r->x > b->x1) || (r->y > b->y1) || (r->z > b->z1)
			|| (r->jbyi * b->x1 - b->y0 + r->c_xy < 0)
			|| (r->ibyj * b->y1 - b->x0 + r->c_yx < 0)
			|| (r->jbyk * b->z1 - b->y0 + r->c_zy < 0)
			|| (r->kbyj * b->y1 - b->z0 + r->c_yz < 0)
			|| (r->kbyi * b->x1 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case OMM:

		if((r->x < b->x0) || (r->x > b->x1)
			|| (r->y < b->y0) || (r->z < b->z0)
			|| (r->jbyk * b->z0 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z1 + r->c_yz > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case OMP:

		if((r->x < b->x0) || (r->x > b->x1)
			|| (r->y < b->y0) || (r->z > b->z1)
			|| (r->jbyk * b->z1 - b->y1 + r->c_zy > 0)
			|| (r->kbyj * b->y0 - b->z0 + r->c_yz < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case OPM:

		if((r->x < b->x0) || (r->x > b->x1)
			|| (r->y > b->y1) || (r->z < b->z0)
			|| (r->jbyk * b->z0 - b->y0 + r->c_zy < 0) 
			|| (r->kbyj * b->y1 - b->z1 + r->c_yz > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case OPP:

		if((r->x < b->x0) || (r->x > b->x1)
			|| (r->y > b->y1) || (r->z > b->z1)
			|| (r->jbyk * b->z1 - b->y0 + r->c_zy < 0)
			|| (r->kbyj * b->y1 - b->z0 + r->c_yz < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MOM:

		if((r->y < b->y0) || (r->y > b->y1)
			|| (r->x < b->x0) || (r->z < b->z0) 
			|| (r->kbyi * b->x0 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MOP:

		if((r->y < b->y0) || (r->y > b->y1)
			|| (r->x < b->x0) || (r->z > b->z1) 
			|| (r->kbyi * b->x0 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x1 + r->c_zx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case POM:

		if((r->y < b->y0) || (r->y > b->y1)
			|| (r->x > b->x1) || (r->z < b->z0)
			|| (r->kbyi * b->x1 - b->z1 + r->c_xz > 0)
			|| (r->ibyk * b->z0 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case POP:

		if((r->y < b->y0) || (r->y > b->y1)
			|| (r->x > b->x1) || (r->z > b->z1)
			|| (r->kbyi * b->x1 - b->z0 + r->c_xz < 0)
			|| (r->ibyk * b->z1 - b->x0 + r->c_zx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MMO:

		if((r->z < b->z0) || (r->z > b->z1)
			|| (r->x < b->x0) || (r->y < b->y0) 
			|| (r->jbyi * b->x0 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x1 + r->c_yx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MPO:

		if((r->z < b->z0) || (r->z > b->z1)
			|| (r->x < b->x0) || (r->y > b->y1) 
			|| (r->jbyi * b->x0 - b->y0 + r->c_xy < 0) 
			|| (r->ibyj * b->y1 - b->x1 + r->c_yx > 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PMO:

		if((r->z < b->z0) || (r->z > b->z1)
			|| (r->x > b->x1) || (r->y < b->y0) 
			|| (r->jbyi * b->x1 - b->y1 + r->c_xy > 0)
			|| (r->ibyj * b->y0 - b->x0 + r->c_yx < 0)  
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case PPO:

		if((r->z < b->z0) || (r->z > b->z1)
			|| (r->x > b->x1) || (r->y > b->y1)
			|| (r->jbyi * b->x1 - b->y0 + r->c_xy < 0)
			|| (r->ibyj * b->y1 - b->x0 + r->c_yx < 0)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case MOO:

		if((r->x < b->x0)
			|| (r->y < b->y0) || (r->y > b->y1)
			|| (r->z < b->z0) || (r->z > b->z1)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case POO:

		if((r->x > b->x1)
			|| (r->y < b->y0) || (r->y > b->y1)
			|| (r->z < b->z0) || (r->z > b->z1)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;

	case OMO:

		if((r->y < b->y0)
			|| (r->x < b->x0) || (r->x > b->x1)
			|| (r->z < b->z0) || (r->z > b->z1)
			)
		{
			retVal = false;
			break;
		}

	case OPO:

		if((r->y > b->y1)
			|| (r->x < b->x0) || (r->x > b->x1)
			|| (r->z < b->z0) || (r->z > b->z1)
			)
		{
			retVal = false;
			break;
		}

	case OOM:

		if((r->z < b->z0)
			|| (r->x < b->x0) || (r->x > b->x1)
			|| (r->y < b->y0) || (r->y > b->y1)
			)
		{
			retVal = false;
			break;
		}

	case OOP:

		if((r->z > b->z1)
			|| (r->x < b->x0) || (r->x > b->x1)
			|| (r->y < b->y0) || (r->y > b->y1)
			)
		{
			retVal = false;
		}
		else
		{
			retVal = true;
		}
		break;
	
	}
/*
	float dist1, dist2, dist3, maxDist;
	if (!retVal)
		return retVal;
	switch (r->classification)
	{
		case MMM: 
		{		
			dist1 = r->x + b->x1;
			dist2 = r->y + b->y1;
			dist3 = r->z + b->z1;			
			break;
		}
			
		case MMP: 
		{		
			dist1 = r->x + b->x1;
			dist2 = r->y + b->y1;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case MPM:
		{		
			dist1 = r->x + b->x1;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = r->z + b->z1;			
			break;
		}

		case MPP:
		{		
			dist1 = r->x + b->x1;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case PMM:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = r->y + b->y1;
			dist3 = r->z + b->z1;			
			break;
		}

		case PMP:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = r->y + b->y1;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case PPM:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = r->z + b->z1;			
			break;
		}

		case PPP:
		{	
			dist1 = -1.0f*r->x + b->x0;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = -1.0f*r->z + b->z0;
			break;
		}

		case OMM:
		{		
			dist1 = 0.0f;
			dist2 = r->y + b->y1;
			dist3 = r->z + b->z1;			
			break;
		}

		case OMP:
		{		
			dist1 = 0.0f;
			dist2 = r->y + b->y1;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case OPM:
		{		
			dist1 = 0.0f;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = r->z + b->z1;			
			break;
		}

		case OPP:
		{		
			dist1 = 0.0f;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case MOM:
		{		
			dist1 = r->x + b->x1;
			dist2 = 0.0f;
			dist3 = r->z + b->z1;			
			break;
		}

		case MOP:
		{		
			dist1 = r->x + b->x1;
			dist2 = 0.0f;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case POM:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = 0.0f;
			dist3 = r->z + b->z1;			
			break;
		}

		case POP:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = 0.0f;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

		case MMO:
		{		
			dist1 = r->x + b->x1;
			dist2 = r->y + b->y1;
			dist3 = 0.0f;			
			break;
		}

		case MPO:
		{		
			dist1 = r->x + b->x1;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = 0.0f;			
			break;
		}

		case PMO:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = r->y + b->y1;
			dist3 = 0.0f;			
			break;
		}

		case PPO:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = 0.0f;			
			break;
		}

		case MOO:
		{		
			dist1 = r->x + b->x1;
			dist2 = 0.0f;
			dist3 = 0.0f;			
			break;
		}

		case POO:
		{		
			dist1 = -1.0f*r->x + b->x0;
			dist2 = 0.0f;
			dist3 = 0.0f;			
			break;
		}

		case OMO:
		{		
			dist1 = 0.0f;
			dist2 = r->y + b->y1;
			dist3 = 0.0f;			
			break;
		}

		case OPO:
		{		
			dist1 = 0.0f;
			dist2 = -1.0f*r->y + b->y0;
			dist3 = 0.0f;			
			break;
		}
			
		case OOM:
		{		
			dist1 = 0.0f;
			dist2 = 0.0f;
			dist3 = r->z + b->z1;			
			break;
		}
			
		case OOP:
		{		
			dist1 = 0.0f;
			dist2 = 0.0f;
			dist3 = -1.0f*r->z + b->z0;			
			break;
		}

			

		default: break;
		
	}
	maxDist = max(max(dist1, dist2), dist3);
	*/
	return retVal;
}
