#include "ModelManager.h"
#include "direct.h"



enum
{
	POSITION_OBJECT = 0,
	UV_OBJECT = 1,
	NORMAL_OBJECT = 2,
	POINT_OBJECT = 3,
	POINT_NORMAL_OBJECT = 4
};

static char* newString(char* s);
vector<Model*> ModelManager::allModels;


//**********************************************************************************
// The places to look for Models...
//    this should be modified to look where you might find your Fbx_Models.
//    ultimately, there should be a better way than just sticking this in
//    the code.
//**********************************************************************************
char* ModelManager::modelPaths[] = 
{
	"Resources/Models",
	0
};

Model::Model()
{
	geometry = NULL;
	newVertices = NULL;
	newNormals = NULL;
	newUvs = NULL;

	minX = minY = minZ = 9999999;
	maxX = maxY = maxZ = -9999999;

	textured = false;
}
Model::~Model()
{
	delete[] modelName;
	delete[] modelFile;
	delete[] shader;

	if (textured)
		delete[] texture;

	delete[] newVertices; newVertices = NULL;
	delete[] newNormals; newNormals = NULL;
	delete[] newUvs; newUvs = NULL;

	newPoints.clear();
	newPointNormals.clear();
	
	glmDelete(geometry);
	geometry = NULL;
}

void ModelManager::loadAllModels(vector<string> modelnames)
{
	int i = 0;
	int anzPoints=0; 
	int anzTriangles=0;

	float minX = 999999.9f;
	float minY = 999999.9f;
	float minZ = 999999.9f;
	float maxX = -999999.9f;
	float maxY = -999999.9f;
	float maxZ = -999999.9f;
	

	char temp[50];
	for(i=0; i<int(modelnames.size()); i++)
	{
		printf("Creating point representation of scene... (%02i models remaining)\r",  int(modelnames.size()-i));
		strcpy_s(temp, modelnames.at(i).c_str());
		Model* m = fetchModel(temp);
		anzPoints += m->anzPoints;
		anzTriangles += m->geometry->numtriangles;

		if(m->box.x0 < minX)
			minX = m->box.x0;
		if(m->box.y0 < minY)
			minY = m->box.y0;
		if(m->box.z0 < minZ)
			minZ = m->box.z0;
		if(m->box.x1 > maxX)
			maxX = m->box.x1;
		if(m->box.y1 > maxY)
			maxY = m->box.y1;
		if(m->box.z1 > maxZ)
			maxZ = m->box.z1;
	}

	Scene::sceneBboxDiagonal = sqrt(pow((maxX-minX), 2.0f) + pow((maxY-minY), 2.0f) + pow((maxZ-minZ), 2.0f));

	printf("Creating point representation of scene... (%02i models remaining)\r", int(modelnames.size()-i));
	printf("\nDone. %i triangles to %i points.\n\n", anzTriangles, anzPoints);  
}

void ModelManager::setShaderParams(char* shadername)
{	
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, TextureManager::lp);
	int texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "LightPositions");
	if (texLoc!=-1)
	glUniform1i(texLoc,1); //s. Orange Book, s. 258

	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D,TextureManager::lc);
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "LightColors");
	if (texLoc!=-1)
		glUniform1i(texLoc,2); //s. Orange Book, s. 258

	glActiveTexture(GL_TEXTURE5);
	glBindTexture(GL_TEXTURE_2D,TextureManager::ln);
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "LightNormals");
	if (texLoc!=-1)
		glUniform1i(texLoc,5); //s. Orange Book, s. 258

	
	glActiveTexture(GL_TEXTURE3);
	glBindTexture(GL_TEXTURE_2D, ism_atlas);
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "Ism_Atlas");
	if (texLoc!=-1)
	glUniform1i(texLoc,3); //s. Orange Book, s. 258

	glActiveTexture(GL_TEXTURE4);
	glBindTexture(GL_TEXTURE_2D, GameStateVBO::ism_matrices);
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "Ism_Matrices");
	if (texLoc!=-1)
	glUniform1i(texLoc,4); //s. Orange Book, s. 258

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "texWidth");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,TextureManager::lightTextureW);
	}

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "anzLights");
	if (texLoc!=-1)
	{
		glUniform1i(texLoc,TextureManager::anzLights);
	}

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "ism_atlas_size");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc, ism_atlas_size);
	}
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "ism_size");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc, ism_size);
	}

	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "farZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::farZ);
	}
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "nearZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::nearZ);

	}

}


//*************************************************************************************
//
// * The core guts of the thing...
//=====================================================================================
Model* ModelManager::fetchModel(char* name)
//=====================================================================================
{
	GLMmodel* b=0;     // The buffer data for the image

	//************************************************************
	// Check to see if we've loaded the Model already
	//************************************************************
	for(int i=0; i<int(allModels.size()); i++)
	{
		Model* m = allModels.at(i);
		if (!strcmp(m->modelName, name)) 
		{
			return m;
		}	  
	}

	return loadModel(name);
}

Model* ModelManager::loadModel(char* name)
{
	Model* t  = new Model();

	//	name = new char[strlen(modelname) + 1];
	//	strcpy(name, modelname);

	char* fullFilePath = new char[2048];

	_getcwd(fullFilePath, 2048);
	strcat_s(fullFilePath, 2048, "\\Resources\\Models\\");
	strcat_s(fullFilePath, 2048, name);


	GLfloat diff[4] = {0.5, 0.5, 0.5, 1.0};
	GLfloat spec[4] = {0.2, 0.2, 0.2, 1.0};
	GLfloat shin = 10.0;
	float reflectivity = 0.1f;

	float repeatTexFactor = 1.0f;


	// grab the model.info file from the path
	FILE *modelInfoFile;
	char modelInfoFileName[2048];

	strcpy_s(modelInfoFileName, fullFilePath);
	strcat_s(modelInfoFileName, "\\model.info");

	fopen_s(&modelInfoFile, modelInfoFileName, "rb");

	if (!modelInfoFile)
	{
		delete t;
		delete[] fullFilePath;
		return NULL;
	}

	char buffer[2048];
	char* temp = new char[2048];
	while (fgets(buffer, 2048, modelInfoFile))
	{
		sscanf(buffer, "%s", buffer, sizeof(buffer));

		char token[2048], value[2048];

		int split = strcspn(buffer, "=");

		strncpy_s(token, buffer, split);
		token[split] = '\0';

		strcpy_s(value, buffer + split + 1);

		if (strcmp(token, "#") == 0)
		{
			//comment, skip
		}
		else if (strcmp(token, "objFile") == 0)
		{
			char fullObjFileName[2048];
			strcpy_s(fullObjFileName, fullFilePath);
			strcat_s(fullObjFileName, "\\");
			strcat_s(fullObjFileName, value);

			t->modelFile = newString(fullObjFileName);

		}
		else if (strcmp(token, "shader") == 0)
		{
			t->shader = newString(value);
		}
		else if (strcmp(token, "textured") == 0)
		{
			strcpy_s(temp, 2048, fullFilePath);
			strcat_s(temp, 2048,  "\\texture_COLOR.jpg");
			t->texture = newString(temp);
			t->textured = true;

		}
		else if (strcmp(token, "texture") == 0)
		{
			t->texture = newString(value);
			t->textured = true;
		}
		else if (strcmp(token, "reflectivity") == 0)
		{
			reflectivity = atof(value);
		}
		else if (strcmp(token, "diffuseR") == 0)
		{
			diff[0] = atof(value);
		}
		else if (strcmp(token, "diffuseG") == 0)
		{
			diff[1] = atof(value);
		}
		else if (strcmp(token, "diffuseB") == 0)
		{
			diff[2] = atof(value);
		}
		else if (strcmp(token, "specularR") == 0)
		{
			spec[0] = atof(value);
		}
		else if (strcmp(token, "specularG") == 0)
		{
			spec[1] = atof(value);
		}
		else if (strcmp(token, "specularB") == 0)
		{
			spec[2] = atof(value);
		}
		else if (strcmp(token, "shininess") == 0)
		{
			shin = atof(value);
		}
		else if (strcmp(token, "repeatTexFactor") == 0)
		{
			repeatTexFactor = atof(value);
		}


	}
	
	fclose(modelInfoFile);
	delete[] fullFilePath;
	delete[] temp;

	GLMmodel* b = glmReadOBJ(t->modelFile);	

	if(!b)
	{
		printf("!! Couldn't load Model '%s'~\n",name);
		delete t;
		return NULL;
	}

	for(int a=0; a<4; a++)
	{
		b->materials->diffuse[a] = diff[a];
		b->materials->specular[a] = spec[a];	
	}	
	b->materials->shininess = shin;
	b->materials->reflectivity = reflectivity;

	// If we found a model, load it into Model data structure and put it into the
	// vector

	// Set up the field in the data structure
	t->modelName = new char[strlen(name)+1];
	strcpy_s(t->modelName, strlen(name)+1, name);

	t->geometry = b;
	TextureManager::fetchTexture(t->texture, true, true);
	t->repeatTexFactor = repeatTexFactor;

	//glmFacetNormals(t->geometry);
	//glmVertexNormals(t->geometry,90);

	// Print out the creating message
	//printf("Created Model %s\n",t->modelName);
	allModels.push_back(t);

	t->initVBODrawing();

	t->setupDrawingBuffersForPointSplatting();

	t->setupDrawingBuffersForManyLightRendering();
	

	return t;
}

void Model::drawVBO(void)
{
	bindDrawingBuffersForManyLightRendering();
	drawArrays();
}

void Model::bindDrawingBuffersForManyLightRendering(void)
{
	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POSITION_OBJECT]);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[UV_OBJECT]);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[NORMAL_OBJECT]);
	glNormalPointer(GL_FLOAT, 0, 0);

	int texLoc;

	if(textured)
	{
		glActiveTexture(GL_TEXTURE0);
		TextureManager::fetchTexture(texture, true, true);

	
		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "Texture");
		if (texLoc!=-1)
			glUniform1i(texLoc,0); //s. Orange Book, s. 258
		

		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "sceneUnitsToMeters");
		if (texLoc!=-1)
			glUniform1f(texLoc,Scene::sceneUnitsToMeters); //s. Orange Book, s. 258

		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "repeatTexFactor");
		if (texLoc!=-1)
		{
			glUniform1f(texLoc,repeatTexFactor);
		}
	}

	glMaterialfv(GL_FRONT, GL_DIFFUSE, geometry->materials->diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, geometry->materials->specular);
	glMaterialf(GL_FRONT, GL_SHININESS, geometry->materials->shininess);

}

void Model::setupDrawingBuffersForManyLightRendering(void)
{
	//build buffer objects
	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POSITION_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (vc)*sizeof(GLfloat), newVertices, GL_STREAM_DRAW);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[UV_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (uvc)*sizeof(GLfloat), newUvs, GL_STREAM_DRAW);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[NORMAL_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (ns)*sizeof(GLfloat), newNormals, GL_STREAM_DRAW);
	glNormalPointer(GL_FLOAT, 0, 0);

	int texLoc;

	if(textured)
	{
		glActiveTexture(GL_TEXTURE0);
		TextureManager::fetchTexture(texture, true, true);

		
		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "Texture");
		if (texLoc!=-1)
			glUniform1i(texLoc,0); //s. Orange Book, s. 258


		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "sceneUnitsToMeters");
		if (texLoc!=-1)
			glUniform1f(texLoc,Scene::sceneUnitsToMeters); //s. Orange Book, s. 258

		texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "repeatTexFactor");
		if (texLoc!=-1)
		{
			glUniform1f(texLoc,repeatTexFactor);
		}
	}



	glMaterialfv(GL_FRONT, GL_DIFFUSE, geometry->materials->diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, geometry->materials->specular);
	glMaterialf(GL_FRONT, GL_SHININESS, geometry->materials->shininess);

}

void Model::setupDrawingBuffersForPointSplatting(void)
{

	newPointsArray = new GLfloat[newPoints.size()];
	newPointNormalsArray = new GLfloat[newPointNormals.size()];

	vector<GLfloat>::iterator itP;
	vector<GLfloat>::iterator itN=newPointNormals.begin();

	//TODO backInserter...
	int i=0;
	for (itP = newPoints.begin(); itP != newPoints.end() && itN != newPointNormals.end() ;itP++, itN++)
	{
		newPointsArray[i] = *itP;
		newPointNormalsArray[i] = *itN;
		i++;
	}

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POINT_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, newPoints.size()*sizeof(GLfloat), newPointsArray, GL_STREAM_DRAW);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	
	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POINT_NORMAL_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, newPointNormals.size()*sizeof(GLfloat), newPointNormalsArray, GL_STREAM_DRAW);
	glNormalPointer(GL_FLOAT, 0, 0);

	delete[] newPointsArray;
	delete[] newPointNormalsArray;
}

void Model::bindDrawingBuffersForPointSplatting(void)
{
	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POINT_OBJECT]);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POINT_NORMAL_OBJECT]);
	glNormalPointer(GL_FLOAT, 0, 0);
}

void Model::drawPointRepresentation(void)
{
	int anzPointCoords = newPoints.size();
	glDrawArrays(GL_POINTS, 0, anzPointCoords);
}

void Model::setupDrawingBuffersForWritingGeometry(void)
{
	//build buffer objects
	glBindBuffer(GL_ARRAY_BUFFER, BufferName[POSITION_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (vc)*sizeof(GLfloat), newVertices, GL_STREAM_DRAW);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[UV_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (uvc)*sizeof(GLfloat), newUvs, GL_STREAM_DRAW);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, BufferName[NORMAL_OBJECT]);
	glBufferData(GL_ARRAY_BUFFER, (ns)*sizeof(GLfloat), newNormals, GL_STREAM_DRAW);
	glNormalPointer(GL_FLOAT, 0, 0);

	if (textured)
	{
		glActiveTexture(GL_TEXTURE0);
		TextureManager::fetchTexture(texture, true, true);
	
		int texLoc = glGetUniformLocation(ShaderManager::fetchShader(shader), "Texture");
		if (texLoc!=-1)
			glUniform1i(texLoc,0); //s. Orange Book, s. 258
	}
}


void Model::drawArrays(void)
{
	glDrawArrays(GL_TRIANGLES, 0, (vc));
}

void Model::initVBODrawing(void)
{		
	BufferName[BUFFERSIZE];

	glGenBuffers(BUFFERSIZE, BufferName);

	newVertices = new GLfloat[(geometry->numtriangles)*9];
	newNormals = new GLfloat[(geometry->numtriangles)*9];
	newUvs = new GLfloat[(geometry->numtriangles)*6];

	char *drawonly = 0;
	GLuint i;
	GLMgroup* group;
	GLMtriangle* triangle;

	assert(geometry);
	assert(geometry->vertices);
	assert(geometry->normals);

	group = geometry->groups;
	vc=uvc=ns=0;

	float cX,cY,cZ, ux, uy, uz, vx, vy, vz, area;
	float minAr = 9999999.0f;
	float maxAr = -9999999.0f;
	float sumAr = 0.0f;
	float cntTr = 0.0f;
	float avgAr;
	anzPoints = 0;
	bool createPoints = false;

	//ORDER VERTICES AND UVS
	while (group) 
	{
		if (drawonly)
			if (strcmp(group->name,drawonly))
			{
				group=group->next;
				continue;
			}	

			for (i = 0; i < group->numtriangles; i++) 
			{
				triangle = &R(group->triangles[i]);

				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[0]];
				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[0]+1];

				newNormals[ns++] = geometry->normals[3 * triangle->nindices[0]];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[0]+1];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[0]+2];

				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[0]];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[0]+1];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[0]+2];

				minX = min(minX, newVertices[vc-3]);
				maxX = max(maxX, newVertices[vc-3]);
				minY = min(minY, newVertices[vc-2]);
				maxY = max(maxY, newVertices[vc-2]);
				minZ = min(minZ, newVertices[vc-1]);
				maxZ = max(maxZ, newVertices[vc-1]);

				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[1]];
				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[1]+1];

				newNormals[ns++] = geometry->normals[3 * triangle->nindices[1]];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[1]+1];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[1]+2];

				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[1]];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[1]+1];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[1]+2];

				minX = min(minX, newVertices[vc-3]);
				maxX = max(maxX, newVertices[vc-3]);
				minY = min(minY, newVertices[vc-2]);
				maxY = max(maxY, newVertices[vc-2]);
				minZ = min(minZ, newVertices[vc-1]);
				maxZ = max(maxZ, newVertices[vc-1]);
			
		

				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[2]];
				newUvs[uvc++] = geometry->texcoords[2 * triangle->tindices[2]+1];

				newNormals[ns++] = geometry->normals[3 * triangle->nindices[2]];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[2]+1];
				newNormals[ns++] = geometry->normals[3 * triangle->nindices[2]+2];

				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[2]];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[2]+1];
				newVertices[vc++] = geometry->vertices[3 * triangle->vindices[2]+2];
				
				minX = min(minX, newVertices[vc-3]);
				maxX = max(maxX, newVertices[vc-3]);
				minY = min(minY, newVertices[vc-2]);
				maxY = max(maxY, newVertices[vc-2]);
				minZ = min(minZ, newVertices[vc-1]);
				maxZ = max(maxZ, newVertices[vc-1]);
				
				
				//calculate parallelogram area (cross prod)
				vx = newVertices[vc-6] - newVertices[vc-9];
				vy = newVertices[vc-5] - newVertices[vc-8];
				vz = newVertices[vc-4] - newVertices[vc-7];
				ux = newVertices[vc-3] - newVertices[vc-9];
				uy = newVertices[vc-2] - newVertices[vc-8];
				uz = newVertices[vc-1] - newVertices[vc-7];
				
				cX = uy*vz - uz*vy;
				cY = (-1)*(ux*vz - uz*vx);
				cZ =  ux*vy-uy*vx;
				
				area = sqrt(cX*cX + cY*cY + cZ*cZ);
				
				maxAr = max(area, maxAr);
				minAr = min(area, minAr);
				sumAr += area;
				cntTr += 1.0;
				avgAr = sumAr / cntTr;

				//determine number of points for this triangle
				//int pointsInThisTriangle = int(area)/10 + 1;
				int pointsInThisTriangle = int(area)/20 + 1;

				for(int p=0; p<pointsInThisTriangle; p++)
				{	
					//random point using barycentric coordinates P = aA + bB + cC where a+b+c=1
					int randomFactor = 1000;
					float a = float(rand() % randomFactor)/float(randomFactor);
					float b = float(rand() % randomFactor)/float(randomFactor);
					float c;
					if(a+b > 1.0)
					{
						a = 1.0 - a;
						b = 1.0 - b;
					}
					c = 1 - (a+b);

					//generate random points
					for(int j=0; j<3; j++)
					{
						newPoints.push_back(newVertices[vc-9+j] * a + newVertices[vc-6+j] * b + newVertices[vc-3+j] * c);
					}

					newPointNormals.push_back(cX/area);
					newPointNormals.push_back(cY/area);
					newPointNormals.push_back(cZ/area);

					anzPoints++;
				}
			}

			//cout << "min: " << minAr << " max: " << maxAr << " avg: " << sumAr/cntTr << endl;
			//cout <<  geometry->numtriangles  << " triangles reduced to " << anzPoints << " points!" << endl;
			group = group->next;
	}

	bvertex[0] = Vector(minX, minY, minZ);
	bvertex[1] = Vector(maxX, minY, minZ);
	bvertex[2] = Vector(maxX, maxY, minZ);
	bvertex[3] = Vector(minX, maxY, minZ);
	bvertex[4] = Vector(minX, minY, maxZ);
	bvertex[5] = Vector(maxX, minY, maxZ);
	bvertex[6] = Vector(maxX, maxY, maxZ);
	bvertex[7] = Vector(minX, maxY, maxZ);

	make_aabox(minX, minY, minZ, maxX, maxY, maxZ, &box);
}

void Model::deleteModelBuffers(void)
{
	glDeleteBuffers(BUFFERSIZE, BufferName);
}

bool Model::rayIntersectsBoundingBox(ray_2 ray)
{
	bool retVal = slope(&ray, &box);
	return retVal;
}

//For freeing all reserved space (including shaders).
void ModelManager::deleteAllModels(void)
{
	printf("\n");
	for(int i=0; i<int(allModels.size()); i++)
	{
		printf("\rDeleting Models (%02i of %02i)", i+1, int(allModels.size()));
		allModels.at(i)->deleteModelBuffers();
		delete allModels.at(i);
	}
	printf(" Done.\n");

	allModels.clear();


}


void ModelManager::renderAllWireframe(void)
{
	glActiveTexture(GL_TEXTURE0);

	for(int i=0; i<int(allModels.size()); i++)
	{		
		glBindBuffer(GL_ARRAY_BUFFER,allModels.at(i)->BufferName[POSITION_OBJECT]);
		glVertexPointer(3, GL_FLOAT, 0, 0);

		glBindBuffer(GL_ARRAY_BUFFER, allModels.at(i)->BufferName[UV_OBJECT]);
		glTexCoordPointer(2, GL_FLOAT, 0, 0);

		glUseProgram(ShaderManager::fetchShader("wireframe"));
		TextureManager::fetchTexture(allModels.at(i)->texture, true, true);
		int texLoc = glGetUniformLocation(ShaderManager::fetchShader("wireframe"), "Texture");
		if (texLoc!=-1)
			glUniform1i(texLoc,0); //s. Orange Book, s. 258

		allModels.at(i)->drawArrays();
	}
}

void ModelManager::renderAll(void)
{
	for(int i=0; i<int(allModels.size()); i++)
	{
		glUseProgram(ShaderManager::fetchShader(allModels.at(i)->shader));
		allModels.at(i)->drawVBO();
	}
}

void ModelManager::renderAllGeometryToTexture(int w, int h, char* shadername)
{
	glUseProgram(ShaderManager::fetchShader(shadername));
	
	int texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "farZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::farZ);
	}
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "nearZ");
	if (texLoc!=-1)
	{
		glUniform1f(texLoc,GameStateVBO::nearZ);
	}
	
/*
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "h");
	if (texLoc!=-1)
	{
		glUniform1i(texLoc,h);
	}
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "w");
	if (texLoc!=-1)
	{
		glUniform1i(texLoc,w);
	}

	glActiveTexture(GL_TEXTURE4);
	glBindTexture(GL_TEXTURE_2D, GameStateVBO::ism_matrices);
	
	texLoc = glGetUniformLocation(ShaderManager::fetchShader(shadername), "Ism_Matrices");
	if (texLoc!=-1)
		glUniform1i(texLoc,4); //s. Orange Book, s. 258
*/
	
	for(int i=0; i<int(allModels.size()); i++)
	{
		allModels.at(i)->setupDrawingBuffersForWritingGeometry();
		allModels.at(i)->drawArrays();
	}
}

void ModelManager::generateVPL(VPL* vpl, Point lightPos, Vector lightNormal, float lightColor[4])
{
	float minDistance = 99999999;
	Point nearestIntersection;

	int indexOfObject = -1;
	bool intersectionFound = false;

	while(!intersectionFound)
	{

		////Vector rayNorm = lightNormal+Vector(float(rand()%300-150)/150.0, 0.0, float(rand()%300-150)/150.0);
		//int s1 = ((rand() % 2)-0.5)*2;
		//int s2 = ((rand() % 2)-0.5)*2;
		//Vector rayNorm = Vector(s1*asin(float(rand()%300/300.0))/1.57, -1.57+asin(float(rand()%300/300.0)), s2*asin(float(rand()%300)/300.0)/1.57);
		//rayNorm.normalize();
		//Ray R = Ray(lightPos, rayNorm);
		

		float u1 = rand() / float(RAND_MAX);
		float u2 = rand() / float(RAND_MAX);

		Vector randVector = Vector(u1, u2, 0.0f);
		Vector diskCoords = toUnitDisk(randVector);

		float r = sqrt(pow(diskCoords.x, 2.0f) + pow(diskCoords.y, 2.0f));
		
		//imagine the plane created by lightNorm, and a transformation
		//so that lightNorm becomes (0,1,0), where the y - axis faces upward.
		Vector sampledVector = Vector(diskCoords.x, sqrt(1.0f - pow(r, 2.0f)), diskCoords.y);
		
		Vector up = Vector(0.0f, 1.0f, 0.0f);
		Vector rotationAxis = crossProduct(up, lightNormal);
		float rotationAngle = acos(lightNormal * up);

		float rotMat[9];
		float cosA = cos(rotationAngle);
		float sinA = sin(rotationAngle);
		rotMat[0] = cosA + pow(rotationAxis.x, 2.0f) * (1.0f - cosA);
		rotMat[1] = rotationAxis.x * rotationAxis.y * (1.0f - cosA) - (rotationAxis.z * sinA);
		rotMat[2] = rotationAxis.x * rotationAxis.z * (1.0f - cosA) + (rotationAxis.y * sinA);
		rotMat[3] = rotationAxis.y * rotationAxis.x * (1.0f - cosA) + (rotationAxis.z * sinA);
		rotMat[4] = cosA + pow(rotationAxis.y, 2.0f) * (1.0f - cosA);
		rotMat[5] = rotationAxis.y * rotationAxis.z * (1.0f - cosA) - (rotationAxis.x * sinA);
		rotMat[6] = rotationAxis.z * rotationAxis.x * (1.0f - cosA) - (rotationAxis.y * sinA);
		rotMat[7] = rotationAxis.z * rotationAxis.y * (1.0f - cosA) + (rotationAxis.x * sinA);
		rotMat[8] = cosA + pow(rotationAxis.z, 2.0f) * (1.0f - cosA);


		//now we must 'reverse' the previous transformation to get the vector in world space:
		Vector rayNorm;
		rayNorm.x = sampledVector.x * rotMat[0] + sampledVector.y * rotMat[1] + sampledVector.z * rotMat[2];
		rayNorm.y = sampledVector.x * rotMat[3] + sampledVector.y * rotMat[4] + sampledVector.z * rotMat[5];
		rayNorm.z = sampledVector.x * rotMat[6] + sampledVector.y * rotMat[7] + sampledVector.z * rotMat[8];
		
		
		Ray R = Ray(lightPos, rayNorm);

		for(int i = 0; i < (int)allModels.size(); i++)
		{
			Model* m = allModels.at(i);

			ray_2 r;
			make_ray(lightPos.x, lightPos.y, lightPos.z, rayNorm.x, rayNorm.y, rayNorm.z, &r);
			if (!m->rayIntersectsBoundingBox(r))
				continue;

			for (int j = 0; j < (int)m->geometry->numtriangles; j++)
			{
				Point p0 = Point(  m->newVertices[(j*9)+0],  m->newVertices[(j*9)+1],  m->newVertices[(j*9)+2]  );
				Point p1 = Point(  m->newVertices[(j*9)+3],  m->newVertices[(j*9)+4],  m->newVertices[(j*9)+5]  );
				Point p2 = Point(  m->newVertices[(j*9)+6],  m->newVertices[(j*9)+7],  m->newVertices[(j*9)+8]  );

				Triangle T = Triangle(p0,p1,p2);
				
			
				Point result; 
				int ab = intersect_RayTriangle(R, T, &result);
				bool continueLoop = true;
				
				switch (ab)
				{
					case 0: continue;  //0 = disjoint (no intersect)
					case 1: {		   //1 = intersect in unique point I1
								Vector lightVec = lightPos-result;
								if (lightVec.getLength()<minDistance)
								{
									minDistance = lightVec.getLength();
									nearestIntersection=result;

									Vector v1(p1.x-p0.x, p1.y-p0.y, p1.z-p0.z);
									Vector v2(p2.x-p0.x, p2.y-p0.y, p2.z-p0.z);
									
									Vector normalOfModel = (crossProduct(v1,v2));
									normalOfModel.normalize();
									Vector refl = normalOfModel * ((normalOfModel * 2.0f) * (R.direction * -1.0f));
									refl = refl - (R.direction * -1.0f);
									vpl->normal = refl;
								//	vpl->normal = normalOfModel;
									vpl->normal.normalize();

									float uv1[2] = {m->newUvs[(j*6)+0], m->newUvs[(j*6)+1]};
									float uv2[2] = {m->newUvs[(j*6)+2], m->newUvs[(j*6)+3]};
									float uv3[2] = {m->newUvs[(j*6)+4], m->newUvs[(j*6)+5]};

									Texture* tex = TextureManager::getTextureObject(m->texture);
									float h = tex->height - 1.0f;
									float w = tex->width - 1.0f;
									
									float red1 = tex->bits[4* int(uv1[1]*h*w + uv1[0]*w)];
									float green1 = tex->bits[4* int(uv1[1]*h*w + uv1[0]*w) +1];
									float blue1 = tex->bits[4* int(uv1[1]*h*w + uv1[0]*w) +2];

									float red2 = tex->bits[4* int(uv2[1]*h*w + uv2[0]*w)];
									float green2 = tex->bits[4* int(uv2[1]*h*w + uv2[0]*w) +1];
									float blue2 = tex->bits[4* int(uv2[1]*h*w + uv2[0]*w) +2];

									float red3 = tex->bits[4* int(uv3[1]*h*w + uv3[0]*w)];
									float green3 = tex->bits[4* int(uv3[1]*h*w + uv3[0]*w) +1];
									float blue3 = tex->bits[4* int(uv3[1]*h*w + uv3[0]*w) +2];
									
									//annahme weisses Licht!!
									//float3 col_rgb = make_float3(   ,  ,  ); 
									//float3 col_lch = Color::rgb2lch(col_rgb);
									
									//vpl->color[3] = -1.0f;
									//vpl->color[1] = col_lch.y;
									//vpl->color[2] = col_lch.z;

									vpl->color[0] =  (red1+red2+red3)/(255.0f*3.0f);
									vpl->color[1] = (green1+green2+green3)/(255.0f*3.0f);
									vpl->color[2] = (blue1+blue2+blue3)/(255.0f*3.0f);

									float len = minDistance*Scene::sceneUnitsToMeters;
							
									//intensity
									lightVec.normalize();
									vpl->intensityRatio = min(1.0f, (30.0f)/((len+3.0f)*(len+3.0f)))*m->geometry->materials->reflectivity * max(0.0f, normalOfModel*lightVec) * max(0.0f, (lightNormal*-1.0f)*lightVec);
									//TODO check ob man hier die luminosity so verwenden kann (vlt. muss man die luminosity von diesem vpl, = col_lch.x, auch beruecksichtigen!
									vpl->color[3] = lightColor[3] * vpl->intensityRatio; 

									//Vector diffuseVector = Vector(m->geometry->materials->diffuse[0]/3.0f, m->geometry->materials->diffuse[1]/3.0f, m->geometry->materials->diffuse[2]/3.0f);
									//vpl->color[3]*=diffuseVector.getLength();

									vpl->position = result; //+ vpl->normal;
									
									indexOfObject=i;
									intersectionFound = true;
								}
								continue; 
							}
					case 2: continue;  //2 = are in the same plane
					case -1:  continue;//-1 = triangle is degenerate (a segment or point)
	            
					default: continue;
				}
			}
		}
		
	}
}

//*************************************************************************************
//
// * Copy and create a new string from the s
//=====================================================================================
static char* newString(char* s)
//=====================================================================================
{
	char* n = new char [strlen(s)+1];
	strcpy_s(n, strlen(s)+1, s);
	return n;
}


//2D Vectors, z == 0!!
Vector ModelManager::toUnitDisk(Vector onSquare)
{
	float phi, r, u, v;
	float a = 2.0f*onSquare.x-1.0f; // (a,b) is now on [-1,1]2
	float b = 2.0f*onSquare.y-1.0f;

	if (a > -b) // region 1 or 2
	{
		if (a > b) // region 1, also |a| > |b|
		{
			r = a;
			phi = (PI/4.0f) * (b/a);
		}
		else // region 2, also |b| > |a|
		{
			r = b;
			phi = (PI/4.0f) * (2.0f - (a/b));
		}
	}
	else // region 3 or 4
	{
		if (a < b) // region 3, also |a| >= |b|, a != 0
		{
			r = -a;
			phi = (PI/4.0f) * (4.0f + (b/a));
		}
		else // region 4, |b| >= |a|, but a==0 and b==0 could occur.
		{
			r = -b;
			if (b != 0)
			{
				phi = (PI/4.0f) * (6.0f - (a/b));
			}
			else
			{
				phi = 0;
			}
		}
	}

	u = r * cos(phi);
	v = r * sin(phi);

	return Vector(u,v, 0.0f);
}