

#ifndef _ModelManager
#define _ModelManager

#include <vector>
#include <stdio.h>
#include <string.h>
#include "OBJ/glm.h"
#include "../ShaderManager/ShaderManager.h"
#include "../../Math/RaytracingMath.h"
#include "../../Math/VPL.h"
#include "../../GameStates/global.h"
#include "../../Math/ray_2.h"
#include "../../Math/slope.h"
#include "../../Math/aabox.h"
#include <GL/glew.h>
#include <FL/gl.H>
#include <list>
#include "../TextureManager/Texture.h"
#include "../../Math/Color.h"


using namespace std;

#define BUFFERSIZE 5
#define R(x) (geometry->triangles[(x)])

class Model 
{
	public:
		Model(void);
		~Model(void);	

		char* shader;

		
		//geometry 
		GLMmodel* geometry;

		//for vbo rendering
		GLuint BufferName[BUFFERSIZE];	

		//name of model
		char* modelName;

		//model-file
		char* modelFile;

		//Texture of this model
		bool textured;
		char* texture;

		//BoundingBox
		aabox box;
		float minX, minY, minZ;
		float maxX, maxY, maxZ;
		Vector bvertex[8];

		//geordnete vertices und uvs zum VBO zeichnen
		GLfloat* newVertices;
		GLfloat* newNormals;
		GLfloat* newUvs;

		vector<GLfloat> newPoints;
		vector<GLfloat> newPointNormals;
		GLfloat* newPointsArray;
		GLfloat* newPointNormalsArray;
		int anzPoints;

		float repeatTexFactor;

		//anzahl der vertices und uvs im VBO (auch mehrmals, wenn in mehreren triangles verwendet)
		GLuint vc, uvc, ns;
		
		//initialize (vertex etc.) buffers
		void initVBODrawing();

		//for freeing sapce
		void deleteModelBuffers();

		//method for drawing this model in vbo mode
		void drawVBO();

		//methods for building VBO and drawing arrays
		void setupDrawingBuffersForManyLightRendering(void);
		void setupDrawingBuffersForWritingGeometry();
		void setupDrawingBuffersForPointSplatting(void);
		void bindDrawingBuffersForPointSplatting(void);
		void bindDrawingBuffersForManyLightRendering(void);
		void bindDrawingBuffersForWritingGeometry(void);
		void drawPointRepresentation(void);
		void drawArrays();

		bool rayIntersectsBoundingBox(ray_2 ray);

		


};

class ModelManager
{
	public:
		static vector<Model*> allModels; 

		static void generateVPL(VPL* vpl, Point lightPos, Vector lightNormal, float lightColor[4]);

		static char* modelPaths[];

		static void loadAllModels(vector<string> modelnames);

		// Get a model- will hunt it down and load it if its not already loaded
		static Model* fetchModel(char* modelName);

		static Model* loadModel(char* modelName);

		static void deleteAllModels(void);

		static void renderAll(void);
		static void renderAllWireframe(void);
		static void renderAllGeometryToTexture(int w, int h, char* shadername);

		static void setShaderParams(char* shadername);
		
		static Vector toUnitDisk(Vector onSquare);
};


#endif