#include "ShaderManager.h"




vector<Shader*> ShaderManager::allShaders;




//**********************************************************************************
// The places to look for Shaders...
// WIRD ZURZEIT IGNORIERT! TODO...
//**********************************************************************************
char* ShaderManager::shaderPaths[] = 
{
	"Resources/Shaders",
	0
};

Shader::Shader(char* shName, char* vertShaderFile, char* pixShaderFile)
{
	shaderName = new char[strlen(shName)+1];
	strcpy_s(shaderName, strlen(shName)+1, shName);
	// Shader loading.

	char buf[200];
	char buf2[200];
	_snprintf_s(buf, 200,"%s/%s", ShaderManager::shaderPaths[0], vertShaderFile);
	_snprintf_s(buf2, 200,"%s/%s", ShaderManager::shaderPaths[0], pixShaderFile);

	std::string VertexShader = ShaderManager::LoadShader(buf);
	std::string PixelShader = ShaderManager::LoadShader(buf2);
    const char* VertexShaderBuffer = VertexShader.c_str();
	const char* PixelShaderBuffer = PixelShader.c_str();

	glShaderName = glCreateProgram();

	GLuint VertShader = glCreateShader(GL_VERTEX_SHADER);
	GLuint PixShader = glCreateShader(GL_FRAGMENT_SHADER);

	glShaderSource(VertShader, 1, &VertexShaderBuffer, NULL);
	glShaderSource(PixShader, 1, &PixelShaderBuffer, NULL);

	glCompileShader(VertShader);
	glCompileShader(PixShader);

if  (!ShaderManager::CheckShader(VertShader, GL_COMPILE_STATUS, "Compile\n", buf)) 
	{
		printf("Error compiling shader %s!\n", vertShaderFile);
		printf("Hit Enter to exit\n");
		char* a = new char[32];
		cin.getline(a, 32);
		//error-message provided by CheckShader()
		exit(-1);
	}
	if (!ShaderManager::CheckShader(PixShader, GL_COMPILE_STATUS, "Compile\n", buf2))
	{
		printf("Error compiling shader %s!\n", pixShaderFile);
		printf("Hit Enter to exit\n");
		char* a = new char[32];
		cin.getline(a, 32);
		//error-message provided by CheckShader()
		exit(-1);
	}

    glAttachShader(glShaderName, VertShader);
	glAttachShader(glShaderName, PixShader);

    glLinkProgram(glShaderName);
    glDeleteShader(VertShader);
	glDeleteShader(PixShader);
    if (!ShaderManager::CheckProgram(glShaderName, GL_LINK_STATUS, "Link\n"))
	{
		printf("Hit Enter to exit\n");
		char* a = new char[32];
		cin.getline(a, 32);
		//error-message provided by CheckProgram()
		exit(-1);
	}

//	printf("Successfully loaded&compiled shader %s!\n", shName);

}
Shader::~Shader()
{
	delete[] shaderName; shaderName = NULL;
	glDeleteProgram(glShaderName);
}

void ShaderManager::loadAllShaders(char* shaderNames[], int numOfShaders)
{
	for(int i=0; i<numOfShaders*3; i+=3)
	{
		loadShader(shaderNames[i], shaderNames[i+1], shaderNames[i+2]);
	}
}

void ShaderManager::loadShader(char* shaderName, char* vertShaderFile, char* pixShaderFile)
{
	allShaders.push_back(new Shader(shaderName, vertShaderFile, pixShaderFile));
}

GLuint ShaderManager::fetchShader(char* shaderToSearch)
{
	for(int i=0; i<int(allShaders.size()); i++)
	{
		Shader* s = allShaders.at(i);
		if (!strcmp(s->shaderName, shaderToSearch)) 
		{
			return s->glShaderName;
		}	  
	}
	printf("Failed to get the shader %s; shader not loaded!\n", shaderToSearch);
	return -1;
}


void ShaderManager::deleteAllShaders(void)
{
	for(int i=0; i<int(allShaders.size()); i++)
	{
		delete allShaders.at(i);
	}
	allShaders.clear();
}


string ShaderManager::LoadShader(const char* filename)
{
	std::ifstream stream(filename, std::ios::in);

    if(!stream.is_open())
        return "";

    std::string Line = "";
	std::string Text = "";

	while(std::getline(stream, Line))
    	Text += "\n" + Line;

	stream.close();

	return Text;
}

bool ShaderManager::CheckShader(int Object, int Type, const char* Message, char* name)
{
	//printf("Shader: %s:\n", name);
    int Success;
    glGetShaderiv(Object, Type, &Success);

	int InfoLogSize;
    glGetShaderiv(Object, GL_INFO_LOG_LENGTH, &InfoLogSize);
	//if(Success == GL_FALSE)
	if (InfoLogSize>1)
	{
        fprintf(stderr, Message);
        int InfoLogSize;
        glGetShaderiv(Object, GL_INFO_LOG_LENGTH, &InfoLogSize);
        char* Buffer = new char[InfoLogSize];
		glGetShaderInfoLog(Object, InfoLogSize, NULL, Buffer);
		fprintf(stderr, "%s\n", Buffer);
        delete[] Buffer;

		if(Success == GL_FALSE)
			return false;
	}
    return true;
}

bool ShaderManager::CheckProgram(int Object, int Type, const char* Message)
{
	//printf("LINKING LINKING LINKING\n");
    int Success;
    glGetProgramiv(Object, Type, &Success);
	if(Success == GL_FALSE)
	{
        fprintf(stderr, Message);
        int InfoLogSize;
        glGetProgramiv(Object, GL_INFO_LOG_LENGTH, &InfoLogSize);
        char* Buffer = new char[InfoLogSize];
		glGetProgramInfoLog(Object, InfoLogSize, NULL, Buffer);
		fprintf(stderr, "%s\n", Buffer);
        delete[] Buffer;
		return false;
	}
    return true;
}
