#include "Texture.h"

using namespace std;


// the list of all textures loaded in already
std::vector<Texture*> TextureManager::allTextures;

GLuint TextureManager::lc; 
GLuint TextureManager::lp;
GLuint TextureManager::ln;

int TextureManager::lightTextureH;
int TextureManager::lightTextureW;
int TextureManager::anzLights;
float* TextureManager::pixelsLC;
float* TextureManager::pixelsLP;
float* TextureManager::pixelsLN;
deque<Cluster*> TextureManager::lightDeque;
CpuLightTree* TextureManager::lightTree;

float TextureManager::sumOfCutsizes;

float TextureManager::sumOfWeightedIntensities;

vector<VPL*> TextureManager::vplVector; 

double getMaterialTermBound(Cluster* nextNode, double posx, double posy, double posz, double nx, double ny, double nz);
double getGeometryTermBound(Cluster* nextNode, double posx, double posy, double posz);
double getGeometryTerm(Cluster* nextNode, double posx, double posy, double posz);
double getVisibility(Cluster* nextNode, double posx, double posy, double posz);

//**********************************************************************************
// The places to look for textures...
//    this should be modified to look where you might find your textures.
//    ultimately, there should be a better way than just sticking this in
//    the code.
//**********************************************************************************
char* TextureManager::texturePaths[] = 
{
	"Resources/Textures",
	0 
};

//For freeing all reserved space (including shaders).
void TextureManager::deleteAllTextures(void)
{
	printf("Deleting Textures...");
	for(int i=0; i<int(allTextures.size()); i++)
	{
		delete allTextures.at(i);
	}
	allTextures.clear();

	printf(" Done.\n");

	printf("Deleting temporary buffers...");
	if (pixelsLP!=0)
	{
		free(pixelsLP);
		pixelsLP = 0;
	}

	if (pixelsLC!=0)
	{
		free(pixelsLC);
		pixelsLC = 0;
	}

	if (pixelsLN!=0)
	{
		free(pixelsLN);
		pixelsLN = 0;
	}
	printf(" Done.\n");

	printf("Deleting VPLs and LightTree...");
	vector<VPL*>::iterator it;
	for (it = vplVector.begin(); it < vplVector.end(); it++)
	{
		delete (*it);
		(*it) = 0;
	}

	delete lightTree;
	lightTree = 0;
	printf(" Done.\n");
}

Texture::~Texture()
{
	glDeleteTextures(1, &glName);
	delete[] texName;
	delete[] bits;
}


#define MAX_IMAGEFORMAT_NUM 7
char* S_ImgFormatArray[MAX_IMAGEFORMAT_NUM]= 
{
	".TGA", 
	".BMP", 
	".GIF", 
	".PNG", 
	".JPEG", 
	".JPG",
	NULL
};



//*************************************************************************************
//
// * Return the file format of the file
//=====================================================================================
int ImageType(char *imgName)
//=====================================================================================
{
	if(!imgName) 
	{
		return -1;
	}

	// Get the input directory and the file extension
	char *imgNameWODirName = (char *) fl_filename_name(imgName); 
	char* ext=(char*) fl_filename_ext(imgName);
	char* capital_ext = new char[strlen(ext)+1];

	for(int i = 0; i < (int)strlen(ext); i++)
	{
		capital_ext[i] = toupper(ext[i]);
	}
	capital_ext[strlen(ext)] = 0;


	//*****************************************************************
	// Scan the image format array to see which image format it is
	//*****************************************************************
	int imgFormatIdx = -1;
	//int a = strcmp(capital_ext, S_ImgFormatArray[5]);
	//printf("strcmp = %i\n", a);
	for(int i = 0;i < MAX_IMAGEFORMAT_NUM && S_ImgFormatArray[i] != NULL;i++)
	{
		if(strcmp(capital_ext, S_ImgFormatArray[i]) == 0)	
		{
			imgFormatIdx = i;
			break;
		} // end of if
	} // end of for
	if(imgFormatIdx == -1)
	{ // Not a valid image format
		printf("Not supported image format...\n");	
		delete[] capital_ext;
		return -1;		
	}

	delete[] capital_ext;

	return imgFormatIdx;
}

//*****************************************************************************
//
// * Copy this into a new image, reversing the rows as it goes. A pointer
//   to the new image is returned.
//=============================================================================
unsigned byte* Reverse_Rows(unsigned byte * data, int *width, int *height)
//=============================================================================
{
	if (! data)
		return NULL;


	unsigned byte* dest = new unsigned byte[*width * *height * 4];


	for (int i = 0 ; i < *height ; i++) 
	{
		int in_offset = (*height - i - 1) * *width * 4;
		int out_offset = i * *width * 4;

		for (int j = 0 ; j < *width ; j++) 
		{
			dest[out_offset + j * 4] = data[in_offset + j * 4];
			dest[out_offset + j * 4 + 1] = data[in_offset + j * 4 + 1];
			dest[out_offset + j * 4 + 2] = data[in_offset + j * 4 + 2];
			dest[out_offset + j * 4 + 3] = data[in_offset + j * 4 + 3];
		}
	}
	return dest;
}// Reverse_Rows


unsigned byte * read_JPG(char *filename, int *width, int *height)
//=============================================================================
{
	if (!filename) 
	{
		printf("No filename given.\n");
		return 0;
	}
	
	Fl_JPEG_Image* jpg_image = new Fl_JPEG_Image(filename);
	*width  = jpg_image->w();
	*height = jpg_image->h();
	int depth  = jpg_image->d();

	if(*width <= 0 || *height <= 0)
	{
		delete jpg_image;
		return NULL;
	}
	if(depth != 3) 
	{
		printf("The image is not in RGB\n");
		return NULL;
	}

	int total = *width * *height * (depth + 1);
	unsigned byte *temp_data = new unsigned byte[total];

	for(int i = 0; i < *width; i++)
	{
		int raw_index = i * *height;
		for(int j = 0; j < *height; j++)
		{
			int org_raw_height = (raw_index + j) * 3;
			int dst_raw_height = (raw_index + j) * 4;
			for(int k = 0; k < depth; k++)
			{
				int org_index = org_raw_height + k;
				int dst_index = dst_raw_height + k;
				temp_data[dst_index] =  jpg_image->data()[0][org_index];
			}
			temp_data[i * *height * 4 + j * 4 + 3] = 255;
		}
	}
	unsigned byte *result = Reverse_Rows(temp_data, width, height);
	delete [] temp_data; temp_data = NULL;
	delete [] jpg_image; jpg_image = NULL;
	return result;
}// read_JPG

//*****************************************************************************
//
// * Read a BMP image from a file.  Return a character array
//   Return 0 on failure.
//=============================================================================
void* read_BMP(char *filename, int *width, int *height)
//=============================================================================
{
	if (!filename) 
	{
		printf("No filename given.\n");
		return 0;
	}

	Fl_BMP_Image *bmp_image = new Fl_BMP_Image(filename);
	*width  = bmp_image->w();
	*height = bmp_image->h();
	int depth  = bmp_image->d();

	if(*width <= 0 || *height <= 0)
	{
		printf("Fail to load in image\n");
		return NULL;
	}
	if(depth != 3) 
	{
		printf("The image is not in RGB\n");
		return NULL;
	}

	int total = *width * *height * (depth + 1);
	unsigned byte *temp_data = new unsigned byte[total];

	for(int i = 0; i < *width; i++)
	{
		int raw_index = i * *height;
		for(int j = 0; j < *height; j++)
		{
			int org_raw_height = (raw_index + j) * 3;
			int dst_raw_height = (raw_index + j) * 4;
			for(int k = 0; k < depth; k++)
			{
				int org_index = org_raw_height + k;
				int dst_index = dst_raw_height + k;
				temp_data[dst_index] =  bmp_image->data()[0][org_index];
			}
			temp_data[i * *height * 4 + j * 4 + 3] = 255;
		}
	}
	unsigned byte *result = Reverse_Rows(temp_data, width, height);
	delete [] temp_data; temp_data = NULL;
	//	delete [] bmp_image; bmp_image = NULL;
	return result;
}// read_BMP

//*****************************************************************************
//
// * Read a BMP image from a file.  Return a character array
//   Return 0 on failure.
//=============================================================================
void* read_GIF(char *filename, int *width, int *height)
//=============================================================================
{
	return 0;
	if (!filename) {
		printf("No filename given.\n");
		return 0;
	}// if

	Fl_GIF_Image *gif_image = new Fl_GIF_Image(filename);
	*width  = gif_image->w();
	*height = gif_image->h();
	int depth  = gif_image->d();

	if(*width <= 0 || *height <= 0){
		printf("Fail to load in image\n");
		return NULL;
	}
	if(depth != 3) {
		printf("The image is not in RGB\n");
		return NULL;
	}

	int total = *width * *height * (depth + 1);
	unsigned byte *temp_data = new unsigned byte[total];

	for(int i = 0; i < *width; i++){
		int raw_index = i * *height;
		for(int j = 0; j < *height; j++){
			int org_raw_height = (raw_index + j) * 3;
			int dst_raw_height = (raw_index + j) * 4;
			for(int k = 0; k < depth; k++){
				int org_index = org_raw_height + k;
				int dst_index = dst_raw_height + k;
				temp_data[dst_index] =  gif_image->data()[0][org_index];
			}
			temp_data[i * *height * 4 + j * 4 + 3] = 255;
		}
	}
	unsigned byte *result = Reverse_Rows(temp_data, width, height);
	delete [] temp_data;
	return result;
}// read_GIP

//*****************************************************************************
//
// * Read a PNG image from a file.  Return a character array
//   Return 0 on failure.
//=============================================================================
void* read_PNG(char *filename, int *width, int *height)
//=============================================================================
{
	return 0;
	if (!filename) {
		printf("No filename given.\n");
		return 0;
	}// if

	Fl_PNG_Image *png_image = new Fl_PNG_Image(filename);
	*width  = png_image->w();
	*height = png_image->h();
	int depth  = png_image->d();

	if(*width <= 0 || *height <= 0){
		printf("Fail to load in image\n");
		return NULL;
	}
	if(depth != 3) {
		printf("The image is not in RGB\n");
		return NULL;
	}

	int total = *width * *height * (depth + 1);
	unsigned byte *temp_data = new unsigned byte[total];

	for(int i = 0; i < *width; i++){
		int raw_index = i * *height;
		for(int j = 0; j < *height; j++){
			int org_raw_height = (raw_index + j) * 3;
			int dst_raw_height = (raw_index + j) * 4;
			for(int k = 0; k < depth; k++){
				int org_index = org_raw_height + k;
				int dst_index = dst_raw_height + k;
				temp_data[dst_index] =  png_image->data()[0][org_index];
			}
			temp_data[i * *height * 4 + j * 4 + 3] = 255;
		}
	}
	unsigned byte *result = Reverse_Rows(temp_data, width, height);
	delete [] temp_data;
	return result;
}// read_BMP

//*************************************************************************************
//
// * Returns pointer to the image, or 0
//=====================================================================================
unsigned byte* read_image(char* filename, int type, int *width, int *height)
//=====================================================================================
{
	switch(type)
	{
	case 0: // TGA
		return 0;//(unsigned byte*)tga_load(filename, *width, *height, TGA_TRUECOLOR_32);
	case 1: // BMP
		return 0;//(unsigned byte*)read_BMP(filename, *width, *height);
	case 2: // GIF
		return 0;//(unsigned byte*)read_GIF(filename, *width, *height);
	case 3: // PNG
		return 0;//(unsigned byte*)read_PNG(filename, *width, *height);
	case 4:
	case 5:
		return read_JPG(filename, width, height);
	default:
		printf("Unknown image file format\n");
		return 0;
	}
	return 0;
}

//*************************************************************************************
//
// * The core guts of the thing...
//=====================================================================================
GLuint TextureManager::fetchTexture(char* name, bool wrapS, bool wrapT)
//=====================================================================================
{
	if(!name)
		return -1;

	//cut off last \n if imported
	char* shortname = name;
	if(name[strlen(name)-1]=='\n')
	{
		int length = strlen(name);

		for(int x=0; x<length-1; x++)
		{
			shortname[x]=name[x];
		}
		shortname[length-1]='\0';
	}

	char buf[200]; // String buffer for all possible filenames

	int   w, h;    // *width and *height of the loading in image
	unsigned byte* b = 0;     // The buffer data for the image

	int i;

	//************************************************************
	// Check to see if we've loaded the texture already
	//************************************************************
	for(int i=0; i<int(allTextures.size()); i++)
	{
		Texture* t = allTextures.at(i);
		if (!strcmp(t->texName, shortname)) 
		{
			glBindTexture(GL_TEXTURE_2D, t->glName);
			return t->glName;
		}	  
	}

	//************************************************************
	// Find the type of the textures
	//************************************************************
	int type = ImageType(shortname);
	if(type < 0 || type > 6)
	{
		printf("!! Unrecognize type of image '%s'~\n",shortname);
		return -1;
	}

	//************************************************************
	// We need to load the file
	// Search everywhere on the path until we find it
	//************************************************************

	//first try to load the file directly (if path was specified).. edit by John
	_snprintf_s(buf, 200,"%s", name);
	b = read_image(buf, type, &w, &h);

	for(i = 0; texturePaths[i] && !b; i++) 
	{
		_snprintf_s(buf, 200,"%s/%s", texturePaths[i], shortname);
		//	printf("Trying: '%s'\n", buf);
		b = read_image(buf, type, &w, &h);
	}

	if (!b)
	{	
		printf("!! Couldn't load texture '%s'~\n", shortname);
		return -1;
	}

	// If we found an image, load it into Texture data structure and put it into the
	// link list

	// Create the structure
	Texture* t  = new Texture();

	// Set up the field in the data structure
	t->texName = new char[strlen(shortname) + 1];
	strcpy_s(t->texName, strlen(shortname) + 1, shortname);
	t->bits     = b;
	t->width    = w;
	t->height   = h;


	// Generate a unique ID inside opengl
	glGenTextures(1,&t->glName);

	// Print out the creating message
	//printf("Created texture %u for %s\n",t->glName,shortname);

	allTextures.push_back(t);

	glBindTexture(GL_TEXTURE_2D, t->glName);

	// Set up the texture clamping information in s direction
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, 
	//									wrapS ? GL_REPEAT : GL_CLAMP_TO_EDGE);
	//// Set up the texture clamping information in t direction
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, 
	//									wrapT ? GL_REPEAT : GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
	// Set up the texture clamping information in t direction
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	// Set up the alignment inside the texture
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	// Generate the mip maps
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE,b);
	//	int e = gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, w, h, GL_RGBA, GL_UNSIGNED_BYTE,b);

	//	printf("Build MipMap(%d %d) returns %d = %s (%d)\n", w, h, e, gluErrorString(e), e);
	return t->glName;
}

Texture* TextureManager::getTextureObject(char* textureName)
{
	for(int i=0; i<int(allTextures.size()); i++)
	{
		Texture* t = allTextures.at(i);
		if (!strcmp(t->texName, textureName)) 
		{
			return t;
		}	  
	}

	return NULL;

}


void TextureManager::generateLightTextures()
{	
	int anzAllLights=0;
	anzClusters = 0;
	list<AreaLight*>::iterator it;
	for(it=Scene::areaLightSources.begin(); it!=Scene::areaLightSources.end(); it++)
	{
		int numberOfPrimaryLights=(*it)->anzPrimaryLights;
		int virtualToPrimaryRatio=(*it)->virtualToPrimaryRatio;
		anzAllLights += (numberOfPrimaryLights + (numberOfPrimaryLights*virtualToPrimaryRatio));
	}

	if (pixelsLP!=0)
	{
		free(pixelsLP);
		pixelsLP = 0;
	}

	if (pixelsLC!=0)
	{
		free(pixelsLC);
		pixelsLC = 0;
	}

	if (pixelsLN!=0)
	{
		free(pixelsLN);
		pixelsLN = 0;
	}

	//for(unsigned int i=0; i<lightDeque.size(); i++)
		//delete lightDeque.at(i);
	lightDeque.clear();

	for(unsigned int i=0; i<vplVector.size(); i++)
	{
		delete vplVector.at(i);
	}
	vplVector.clear();

	cout << "VPL vector size = " << vplVector.size() << endl;

	void *dataLC = malloc(4*sizeof(float)*anzAllLights);
	pixelsLC = (float*) dataLC;

	//Licht Positions:
	void *dataLP = malloc(4*sizeof(float)*anzAllLights);
	pixelsLP = (float*) dataLP;

	//Licht Normals
	void *dataLN = malloc(4*sizeof(float)*anzAllLights);
	pixelsLN = (float*) dataLN;

	CUDA_LightCuts::deleteCudaArrays();

	int sumOfAllLights = 0;
	for(it=Scene::areaLightSources.begin(); it!=Scene::areaLightSources.end(); it++)
	{
		int numberOfPrimaryLights=(*it)->anzPrimaryLights;
		int virtualToPrimaryRatio=(*it)->virtualToPrimaryRatio;

		sumOfAllLights += numberOfPrimaryLights + (numberOfPrimaryLights*virtualToPrimaryRatio);
	}

//		CUDA_LightCuts::clustersInCUDA = new ClusterInCUDA[(sumOfAllLights*2)-1];
//		CUDA_LightCuts::representativeLights = new VPL_InCUDA[(sumOfAllLights*2)-1];

		CUDA_LightCuts::positionOfVPL = new float4[(sumOfAllLights*2)-1];
		CUDA_LightCuts::normalOfVPL_IntensityRatio = new float4[(sumOfAllLights*2)-1]; //intensity ratio wird in normalOfVPL_IntensityRatio.w gespeichert!!!
		CUDA_LightCuts::colorOfVPL = new float4[(sumOfAllLights*2)-1];
		CUDA_LightCuts::shadowMapAndLightSpaceOfVPL = new int4[(sumOfAllLights*2)-1];

		CUDA_LightCuts::indices = new int4[(sumOfAllLights*2)-1]; //indices.r = leftChild, indices.g = rightChild, indices.b = index of representative Light (in datastructure above!)
		CUDA_LightCuts::bounds_A = new float4[(sumOfAllLights*2)-1];
		CUDA_LightCuts::bounds_B = new float4[(sumOfAllLights*2)-1];
		CUDA_LightCuts::color_lch = new float4[(sumOfAllLights*2)-1]; //geometry term, visibility term, material term, intensity
		CUDA_LightCuts::estRad_RadUpperBound_errorBound = new float4[(sumOfAllLights*2)-1];


	printf("------------------------------------------------\n");
	printf("\n");
	int lightsCreated = 0;
	sumOfWeightedIntensities = 0.0;
	Timer timTimer;

	for(it=Scene::areaLightSources.begin(); it!=Scene::areaLightSources.end(); it++)
	{
		int numberOfPrimaryLights=(*it)->anzPrimaryLights;
		int virtualToPrimaryRatio=(*it)->virtualToPrimaryRatio;
		float intensity = (*it)->intensity;

		lightTextureW = int(ceil(sqrt(double(numberOfPrimaryLights + (numberOfPrimaryLights*virtualToPrimaryRatio)))));
		lightTextureH = lightTextureW;
		anzLights = numberOfPrimaryLights + (numberOfPrimaryLights*virtualToPrimaryRatio);

		for(int i=0; i < anzLights*4; i+= (4 *(virtualToPrimaryRatio+1)) )
		{
			//primary light:
			pixelsLP[lightsCreated*4] =	float((*it)->bBoxMin.x + (float(rand()%100)/100.0)*((*it)->bBoxMax.x-(*it)->bBoxMin.x));
			pixelsLP[lightsCreated*4+1] = float((*it)->bBoxMin.y + (float(rand()%100)/100.0)*((*it)->bBoxMax.y-(*it)->bBoxMin.y));
			pixelsLP[lightsCreated*4+2] = float((*it)->bBoxMin.z + (float(rand()%100)/100.0)*((*it)->bBoxMax.z-(*it)->bBoxMin.z));
			pixelsLP[lightsCreated*4+3]=	(1.0f); //grundintensity-verhaeltnis wird im alpha der position gespeichert

			float3 rgbWhite = make_float3(1.0f, 1.0f, 1.0f);
			//float3 lchWhite = Color::rgb2lch(rgbWhite);

			pixelsLC[lightsCreated*4] = rgbWhite.x;
			pixelsLC[lightsCreated*4+1] = rgbWhite.y;
			pixelsLC[lightsCreated*4+2] = rgbWhite.z;
			pixelsLC[lightsCreated*4+3] = intensity;

			pixelsLN[lightsCreated*4] = (*it)->primaryDirection.x;
			pixelsLN[lightsCreated*4+1] = (*it)->primaryDirection.y;
			pixelsLN[lightsCreated*4+2] = (*it)->primaryDirection.z;
			pixelsLN[lightsCreated*4+3]= 0.0f;  //0=primary


			VPL* newVPL = new VPL();
			newVPL->position = Point(pixelsLP[lightsCreated*4], pixelsLP[lightsCreated*4+1], pixelsLP[lightsCreated*4+2]); 
			newVPL->normal = Vector(pixelsLN[lightsCreated*4], pixelsLN[lightsCreated*4+1], pixelsLN[lightsCreated*4+2]);
			newVPL->color[0] = pixelsLC[lightsCreated*4];
			newVPL->color[1] = pixelsLC[lightsCreated*4+1];
			newVPL->color[2] = pixelsLC[lightsCreated*4+2];
			newVPL->color[3] = pixelsLC[lightsCreated*4+3];
			newVPL->intensityRatio = 1.0f;

			sumOfWeightedIntensities+=1.0f;

			lightsCreated++;
			printf("Generating %i lights... (%05i remaining)\r", anzLights, anzLights - lightsCreated);

			newVPL->truncateData();
			vplVector.push_back(newVPL);
			Cluster* someCluster = new Cluster(newVPL);
			someCluster->sameOrientation=true;

	//		ClusterInCUDA* cudaCluster = new ClusterInCUDA;
	//		VPL_InCUDA* cudaVPL = new VPL_InCUDA;

			setCUDA_ClusterProperties(someCluster);

			

	//		CUDA_LightCuts::clustersInCUDA[anzClusters] = *cudaCluster;
	//		CUDA_LightCuts::representativeLights[anzClusters] = *cudaVPL;

			anzClusters++;

			lightDeque.push_back(someCluster);	

			//now generate VPLs for the current primary light:
			for (int j = 1; j <= virtualToPrimaryRatio; j++)
			{
				float col[] = {pixelsLC[i], pixelsLC[i+1], pixelsLC[i+2], pixelsLC[i+3]};
				VPL* vpl = new VPL();
				ModelManager::generateVPL(vpl, Point(pixelsLP[i], pixelsLP[i+1], pixelsLP[i+2]),
					Vector(pixelsLN[i], pixelsLN[i+1], pixelsLN[i+2]),
					col);

				pixelsLP[lightsCreated*4] =	vpl->position.x;
				pixelsLP[lightsCreated*4+1] = vpl->position.y;
				pixelsLP[lightsCreated*4+2] = vpl->position.z;
				pixelsLP[lightsCreated*4+3]=	vpl->intensityRatio;

				pixelsLC[lightsCreated*4] = vpl->color[0];
				pixelsLC[lightsCreated*4+1] = vpl->color[1];
				pixelsLC[lightsCreated*4+2] = vpl->color[2];
				pixelsLC[lightsCreated*4+3] = vpl->color[3];

				pixelsLN[lightsCreated*4] = vpl->normal.x;
				pixelsLN[lightsCreated*4+1] = vpl->normal.y;
				pixelsLN[lightsCreated*4+2] = vpl->normal.z;
				pixelsLN[lightsCreated*4+3]= 1.0f; //1=firstbounce

				vpl->truncateData();

				Cluster* someCluster2 = new Cluster(vpl);
				someCluster2->sameOrientation=true;
				vplVector.push_back(vpl);

	//			ClusterInCUDA* cudaCluster2 = new ClusterInCUDA;
	//			VPL_InCUDA* cudaVPL2 = new VPL_InCUDA;
				setCUDA_ClusterProperties(someCluster2);
				


		//		CUDA_LightCuts::clustersInCUDA[anzClusters] = *cudaCluster2;
		//		CUDA_LightCuts::representativeLights[anzClusters] = *cudaVPL2;

				anzClusters++;

				lightDeque.push_back(someCluster2);	

				sumOfWeightedIntensities+=vpl->intensityRatio;

				lightsCreated++;
				printf("Generating %i lights... (%05i remaining)\r", anzLights, anzLights - lightsCreated);
			}

		}
	}

	const double dTimePerFrame = double(timTimer);

	timTimer.start();
	printf("\nDone (%f sek).", dTimePerFrame);

	printf("\n\n");
	//printf("Done.\n");
/*
	int lightTextureWH = int(ceil(sqrt(double(anzAllLights))));

	glGenTextures(1,&lc); // Allocate space for texture
	glBindTexture(GL_TEXTURE_2D,lc); 

	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, lightTextureWH , lightTextureWH , 0, GL_RGBA,GL_FLOAT,dataLC); 

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);

	glGenTextures(1,&lp); // Allocate space for texture
	glBindTexture(GL_TEXTURE_2D,lp); // Set our Tex handle as current 

	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, lightTextureWH , lightTextureWH , 0, GL_RGBA,GL_FLOAT, dataLP); 

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);

	glGenTextures(1,&ln); // Allocate space for texture
	glBindTexture(GL_TEXTURE_2D,ln); // Set our Tex handle as current 

	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, lightTextureWH , lightTextureWH , 0, GL_RGBA, GL_FLOAT, dataLN); 

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
*/
}

void TextureManager::createkDTreeAndClusterLightTree(void)
{
	int a=0;

	if(lightTree!=0)
	{
		delete lightTree;
		lightTree=0;
	}

	lightTree = new CpuLightTree();

	if(clusterLightTree)
	{
		deque<Cluster*>::iterator it;

		for (it = lightDeque.begin(); it < lightDeque.end(); it++)
		{
			lightTree->addCluster(*it);
		}

		lightTree->buildClusters2();

		//	lightTree->debugOutput(lightTree->root);
		printf("rootIntensity = %f\n", lightTree->root->intensity);

	}
}
/*
void removeErrorBoundFromVector(vector<double>* errorVector, double error)
{
	errorVector;
	int a=0;
	vector<double>::iterator it;
	for(it=errorVector->begin(); it!=errorVector->end(); it++)
	{
		if((*it) == error)
		{
			errorVector->erase(it);
			return;
		}
	}
}

float findLargestErrorInVector(vector<double>* errorVector)
{
	double largestError = 0;
	for(unsigned int i=0; i<errorVector->size(); i++)
	{
		largestError=max(largestError, errorVector->at(i));
	}
	return largestError;
}*/

void TextureManager::shadeLightcuts(double posx, double posy, double posz, double nx, double ny, double nz, double cr, double cg, double cb, float* retVal)
{
	FILE* file;
	if (printLightcutDebugInfoToFile)
	{
		file = fopen("test2.test", "a");
	}
//	fprintf(file, "\n######################\nNEUER PIXEL\n");

	//no normal vector defined->skybox or whatever (no geometry)
	if(Vector(nx,ny,nz).getLength()<=0.0f)
	{
		retVal[0] = 1.0f;
		for(int i=1; i<6; i++) retVal[i] = 0.0f;
		fclose(file);
		return;
	}		

	double criterion = 0.02;
	LightCut lc;
	lc.sumIlluminationEstimates=0.0;
	lc.sumRadianceUpperBounds=0.0;

	//mit nextNode ist hier jener cluster gemeint, der als naechstes in den cut kommen wird
	Cluster* nextNode = lightTree->root;

	//calculate estimated radiance of nextNode:
	nextNode->visibilityTerm = getVisibility(nextNode, posx, posy, posz);
	nextNode->geometryTerm = getGeometryTerm(nextNode, posx, posy, posz);
	Vector lightPos = Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
	Vector L = lightPos-Vector(posx, posy, posz);
	L.normalize();
	//check ob stelle hinter licht:
	Vector lightNorm = nextNode->representativeLight->normal;
	nextNode->materialTerm = max(0.0f, (Vector(nx, ny, nz)*L)  * (lightNorm*(L*(-1.0f)))); //shade pixel with this light/cluster
	nextNode->estimatedRadiance = nextNode->visibilityTerm * nextNode->geometryTerm * nextNode->materialTerm * nextNode->intensity;

	//errechne radiance upper bound von nextNode 
	double visibilityTermBound = 1.0;
	double geometryTermBound = getGeometryTermBound(nextNode, posx, posy, posz);
	if(geometryTermBound < nextNode->geometryTerm)
		cout << geometryTermBound << " " << nextNode->geometryTerm << endl;
	double materialTermBound = getMaterialTermBound(nextNode, posx, posy, posz, nx, ny, nz);
	nextNode->radianceUpperBound = visibilityTermBound*geometryTermBound*materialTermBound*nextNode->intensity;

	//errechne fehler bei Nicht-Unterteilung:
	nextNode->errorBound = abs(nextNode->radianceUpperBound-nextNode->estimatedRadiance);

	lc.cut.push(nextNode);
	lc.sumIlluminationEstimates += nextNode->estimatedRadiance;
	lc.sumRadianceUpperBounds +=nextNode->radianceUpperBound;
	
	lc.accColor.x = nextNode->estimatedRadiance * nextNode->representativeLight->color[0];
	lc.accColor.y = nextNode->estimatedRadiance * nextNode->representativeLight->color[1];
	lc.accColor.z = nextNode->estimatedRadiance * nextNode->representativeLight->color[2];

	Cluster* largest;

	bool leftChildIsLeaf = false;
	bool rightChildIsLeaf = false;
	 

//REFINEMENT:
	//while(int(lc.cut.size())<=1 || (int(lc.cut.size()) <= TextureManager::anzLights) && abs(lc.sumIlluminationEstimates-lc.sumRadianceUpperBounds) > criterion*lc.sumIlluminationEstimates))
	while(int(lc.cut.size())<=1 || (int(lc.cut.size()) <= TextureManager::anzLights))
	{
		//cluster with largest error bound will be replaced by children
		largest = lc.cut.top();

		if (printLightcutDebugInfoToFile)
		{
			fprintf(file, "\nBeginn einer neuen Iteration\n");
			fprintf(file, "Cluster - position mit dem groessten Fehler: %.2f, %.2f, %.2f\n", largest->representativeLight->position.x, largest->representativeLight->position.y, largest->representativeLight->position.z);
			fprintf(file, "largest.estimatedRadiance,radUpperBound = %.2f, %.2f\n", largest->estimatedRadiance, largest->radianceUpperBound);
			fprintf(file, "criterion = %.2f\n", lc.sumIlluminationEstimates * criterion);
		}
		if(largest->errorBound < criterion*lc.sumIlluminationEstimates)
			break;

		//if the cluster yielding the largest error has children, refine! else break
		if(largest->leftChild!=0 && largest->rightChild!=0)
		{
			lc.cut.pop();
			lc.sumIlluminationEstimates -= largest->estimatedRadiance;
			lc.sumRadianceUpperBounds -=largest->radianceUpperBound;

			lc.accColor.x -= largest->estimatedRadiance * largest->representativeLight->color[0];
			lc.accColor.y -= largest->estimatedRadiance * largest->representativeLight->color[1];
			lc.accColor.z -= largest->estimatedRadiance * largest->representativeLight->color[2];
		

		//INSERT LEFT CHILD
			nextNode = largest->leftChild;
			leftChildIsLeaf = nextNode->leftChild == 0;

			//calculate estimated radiance of nextNode (or reuse if same as with parent):
			if(largest->representativeLight == largest->leftChild->representativeLight)
			{
				nextNode->geometryTerm=largest->geometryTerm;
				nextNode->visibilityTerm=largest->visibilityTerm;
				nextNode->materialTerm=largest->materialTerm;			
			}
			else
			{
				nextNode->visibilityTerm = getVisibility(nextNode, posx, posy, posz);
				nextNode->geometryTerm = getGeometryTerm(nextNode, posx, posy, posz);
				lightPos = Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
				L = lightPos-Vector(posx, posy, posz);
				L.normalize();
				lightNorm = nextNode->representativeLight->normal;
				nextNode->materialTerm = max(0.0f, (Vector(nx, ny, nz)*L)  * (lightNorm*(L*(-1.0f))));
			}
			nextNode->estimatedRadiance = nextNode->visibilityTerm * nextNode->geometryTerm * nextNode->materialTerm * nextNode->intensity;

			if (!leftChildIsLeaf)
			{
				//errechne radiance upper bound von nextNode 
				visibilityTermBound = 1.0;
				geometryTermBound = getGeometryTermBound(nextNode, posx, posy, posz);
				materialTermBound = getMaterialTermBound(nextNode, posx, posy, posz, nx, ny, nz);
				nextNode->radianceUpperBound = visibilityTermBound*geometryTermBound*materialTermBound*nextNode->intensity;

				//errechne fehler bei Nicht-Unterteilung:
				nextNode->errorBound = abs(nextNode->radianceUpperBound-nextNode->estimatedRadiance);

				if (printLightcutDebugInfoToFile)
					fprintf(file, "Left Child, estimatedRadiance, upperBound = %.2f, %.2f\n", nextNode->estimatedRadiance, nextNode->radianceUpperBound);
			}
			else
			{
				if (printLightcutDebugInfoToFile)
					fprintf(file, "Left Child, estimatedRadiance, upperBound = %.2f, %.2f\n", nextNode->estimatedRadiance, nextNode->estimatedRadiance);

				nextNode->radianceUpperBound = nextNode->estimatedRadiance;
				nextNode->errorBound = 0.0f;
			}

			lc.cut.push(nextNode);
			lc.sumIlluminationEstimates += nextNode->estimatedRadiance;
			lc.sumRadianceUpperBounds +=nextNode->radianceUpperBound;

			lc.accColor.x += nextNode->estimatedRadiance * nextNode->representativeLight->color[0];
			lc.accColor.y += nextNode->estimatedRadiance * nextNode->representativeLight->color[1];
			lc.accColor.z += nextNode->estimatedRadiance * nextNode->representativeLight->color[2];

		//INSERT RIGHT CHILD boookmark
			nextNode = largest->rightChild;
			rightChildIsLeaf = nextNode->leftChild == 0;

			//calculate estimated radiance of nextNode (or reuse):
			if(largest->representativeLight == largest->rightChild->representativeLight)
			{
				nextNode->geometryTerm=largest->geometryTerm;
				nextNode->visibilityTerm=largest->visibilityTerm;
				nextNode->materialTerm=largest->materialTerm;	
			}
			else
			{
				nextNode->visibilityTerm = getVisibility(nextNode, posx, posy, posz);
				nextNode->geometryTerm = getGeometryTerm(nextNode, posx, posy, posz);
				lightPos = Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
				L = lightPos-Vector(posx, posy, posz);
				L.normalize();
				//check ob stelle hinter licht:
				lightNorm = nextNode->representativeLight->normal;
				nextNode->materialTerm = max(0.0f, (Vector(nx, ny, nz)*L)  * (lightNorm*(L*(-1.0f))));
			}
			nextNode->estimatedRadiance = nextNode->visibilityTerm * nextNode->geometryTerm * nextNode->materialTerm * nextNode->intensity;

			if (!rightChildIsLeaf)
			{
				//errechne radiance upper bound von nextNode 
				visibilityTermBound = 1.0;
				geometryTermBound = getGeometryTermBound(nextNode, posx, posy, posz);
				if(geometryTermBound < nextNode->geometryTerm)
					cout << geometryTermBound << " " << nextNode->geometryTerm << endl;
				materialTermBound = getMaterialTermBound(nextNode, posx, posy, posz, nx, ny, nz);
				nextNode->radianceUpperBound = visibilityTermBound*geometryTermBound*materialTermBound*nextNode->intensity;

				//errechne fehler bei Nicht-Unterteilung:
				nextNode->errorBound = abs(nextNode->radianceUpperBound-nextNode->estimatedRadiance);

				if (printLightcutDebugInfoToFile)
					fprintf(file, "Right Child, estimatedRadiance, upperBound = %.2f, %.2f\n", nextNode->estimatedRadiance, nextNode->radianceUpperBound);
			}
			else
			{
				nextNode->radianceUpperBound = nextNode->estimatedRadiance;
				nextNode->errorBound = 0.0f;
				if (printLightcutDebugInfoToFile)
					fprintf(file, "Right Child, estimatedRadiance, upperBound = %.2f, %.2f\n", nextNode->estimatedRadiance, nextNode->estimatedRadiance);

			}
			lc.cut.push(nextNode);
			lc.sumIlluminationEstimates += nextNode->estimatedRadiance;
			lc.sumRadianceUpperBounds +=nextNode->radianceUpperBound;

			lc.accColor.x += nextNode->estimatedRadiance * nextNode->representativeLight->color[0];
			lc.accColor.y += nextNode->estimatedRadiance * nextNode->representativeLight->color[1];
			lc.accColor.z += nextNode->estimatedRadiance * nextNode->representativeLight->color[2];

		}
		else
			break;		

	}

//	printf("SoftwareError = %f\n\n", abs(lc.sumIlluminationEstimates-lc.sumRadianceUpperBounds)/lc.sumIlluminationEstimates);
	
	//CUTSIZE VISUALIZATION:
	double csize = double(lc.cut.size())/double(TextureManager::anzLights);
	TextureManager::sumOfCutsizes += csize;

	//LIGHTCUTS SHADING
	retVal[0] = lc.accColor.x * cr / TextureManager::sumOfWeightedIntensities;
	retVal[1] = lc.accColor.y * cg / TextureManager::sumOfWeightedIntensities;
	retVal[2] = lc.accColor.z * cb/ TextureManager::sumOfWeightedIntensities;

	for(int i=0; i<3; i++)
	{
		retVal[3+i] = csize;
	}

	if (printLightcutDebugInfoToFile)
	{
		fprintf(file, "CUTSIZE = %i\n", lc.cut.size());
		fclose(file);
	}
	return;

	
}


void TextureManager::shadeIR(double posx, double posy, double posz, double nx, double ny, double nz, double cr, double cg, double cb, float* retVal)
{
	//no normal vector defined->skybox or whatever (no geometry)
	if(Vector(nx,ny,nz).getLength()<=0.0f)
	{
		retVal[0] = 1.0f;
		for(int i=1; i<6; i++) retVal[i] = 0.0f;
		return;
	}		

	deque<Cluster*>::iterator light;

	float visibilityTerm, geometryTerm, materialTerm;

	Vector illumination = Vector(0.0f, 0.0f, 0.0f);

	for (light = lightDeque.begin(); light < lightDeque.end(); light++)
	{
		visibilityTerm = getVisibility((*light), posx, posy, posz);
		geometryTerm = getGeometryTermBound((*light), posx, posy, posz);
		
		Vector lightPos = Vector((*light)->representativeLight->position.x, (*light)->representativeLight->position.y, (*light)->representativeLight->position.z);
		Vector L = lightPos-Vector(posx, posy, posz);
		L.normalize();
		//check ob stelle hinter licht:

		//check ob stelle hinter licht:
		Vector lightNorm = (*light)->representativeLight->normal;
		materialTerm = max(0.0f, (Vector(nx, ny, nz)*L)  * (lightNorm*(L*(-1.0f)))); //shade pixel with this light
		Vector vplColor = Vector((*light)->representativeLight->color[0], (*light)->representativeLight->color[1], (*light)->representativeLight->color[2]);
		illumination = illumination + vplColor * (visibilityTerm * geometryTerm * materialTerm * (*light)->intensity);
	}

	//LIGHTCUTS SHADING
	illumination = illumination * (1.0f/ TextureManager::sumOfWeightedIntensities);
	retVal[0] = illumination.x * cr;
	retVal[1] = illumination.y * cg;
	retVal[2] = illumination.z * cb;
	
	for(int i=0; i<3; i++)
	{
		retVal[3+i] = 1.0;
	}
	return;

	
}


double getMaterialTermBound(Cluster* nextNode, double posx, double posy, double posz, double nx, double ny, double nz)
{
	//Cluster BBOX
	//TODO evtl: das bei jedem cluster speichern statt fuer jedes pixel ausrechne
	//d = differenzvektor containing (width, height, depth) of bbox
	Vector d = nextNode->bounds.B - nextNode->bounds.A;
	
	//falls d laenge 0 hat, sind wir bereits bei einem blattknoten, so return die exakte brdf
	if(d.getLength()<=0.0)
	{
		Vector lightPos = Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
		Vector L = lightPos-Vector(posx, posy, posz);
		Vector lightNorm = nextNode->representativeLight->normal;
		L.normalize();
		return max(0.0f, (Vector(nx, ny, nz)*L)  * (lightNorm*(L*(-1.0f))));
	}

	//Beschleunigungs-Hack fuer stellen hinter dem licht: falls wir noch nicht bei einem blatt sind, die lichter im cluster aber 
	//alle die gleiche orientierung haben und von der stelle weg leuchten so returniere 0
	/*
	if(nextNode->sameOrientation)
	{
		Vector lightPos = Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
		Vector L = lightPos-Vector(posx, posy, posz);
		L.normalize();
		if(nextNode->representativeLight->normal*(L*(-1))<0)
			return 0.0;
	}
	*/

	Point bbox[8] = {nextNode->bounds.A, nextNode->bounds.A + Vector(d.x, 0.0, 0.0), 
					 nextNode->bounds.A + Vector(0.0, d.y, 0.0), nextNode->bounds.A + Vector(0.0, 0.0, d.z), 
					 nextNode->bounds.A + Vector(d.x, d.y, 0.0), nextNode->bounds.A + Vector(d.x, 0.0, d.z), 
					 nextNode->bounds.A + Vector(0.0, d.y, d.z), nextNode->bounds.B};

	//we want to transform bbox + shaded position in a way, that shaded position = origin and normal vector = zAxis
	
	//TRANSLATION VECTOR for bounding box
	Vector translateLightToOrigin = Vector(-posx, -posy, -posz);
	
	//rotation matrix, explained later
	double mat[9];

	//if the normal vector isn't already (0,0,1):
	if( ! (nx==0.0 && ny==0.0))
	{
		//SETUP ROTATION MATRIX so that normal vector of shaded position = zAxis
		double alpha = acos(nz / sqrt(nx*nx+ny*ny+nz*nz)); //drehwinkel alpha um folgende drehachse im raum:
			
		//v=Drehachse im Raum:
		double vX = ny;
		double vY = (-1)*nx;
		double vZ = 0.0;
		double lenAxis = sqrt(vX*vX+vY*vY);
		vX/=lenAxis;
		vY/=lenAxis;

		double cosA = cos(alpha);
		double sinA = sin(alpha);

		//ROT MAT:
		mat[0] = cosA + vX*vX*(1.0-cosA);
		mat[1] = vX*vY*(1.0-cosA)-vZ*sinA;
		mat[2] = vX*vZ*(1.0-cosA)+vY*sinA;
		mat[3] = vY*vX*(1.0-cosA)+vZ*sinA;
		mat[4] = cosA+vY*vY*(1.0-cosA);
		mat[5] = vY*vZ*(1.0-cosA)-vX*sinA;
		mat[6] = vZ*vX*(1.0-cosA)-vY*sinA;
		mat[7] = vZ*vY*(1.0-cosA)+vX*sinA;
		mat[8] = cosA+vZ*vZ*(1.0-cosA);
	}

	//transform all 8 vertices of bbox
	for(int i=0; i<8; i++)
	{
		//TRANSLATE BBOX (and light) so that shaded position = origin	
		bbox[i] = bbox[i] + translateLightToOrigin;

		//rotate only if normal vector isnt 0,0,1 yet	
		if( ! (nx==0.0 && ny==0.0))
		{
			//ROTATE BBOX
			double P[3] = {bbox[i].x, bbox[i].y, bbox[i].z};	
			bbox[i].x = P[0]*mat[0] + P[1]*mat[1] + P[2]*mat[2];
			bbox[i].y = P[0]*mat[3] + P[1]*mat[4] + P[2]*mat[5];
			bbox[i].z = P[0]*mat[6] + P[1]*mat[7] + P[2]*mat[8];	
		}
	}

	//GET NEW axis aligned bounds:
	double maxX, maxY, maxZ;
	maxX = maxY = maxZ = -999999.9f;
	double minX, minY, minZ;
	minX = minY = minZ = 999999.9f;
	for(int i=0; i<8; i++)
	{
		if(bbox[i].x > maxX) maxX = bbox[i].x;
		if(bbox[i].x < minX) minX = bbox[i].x;
		if(bbox[i].y > maxY) maxY = bbox[i].y;
		if(bbox[i].y < minY) minY = bbox[i].y;
		if(bbox[i].z > maxZ) maxZ = bbox[i].z;
		if(bbox[i].z < minZ) minZ = bbox[i].z;
	}

	//FIND MAXIMUM COSINE:
	double maxX2, maxY2, maxZ2, minX2, minY2;
	if(maxX*minX<0.0) //0 liegt dazwischen
		minX2 = 0.0;
	else
		minX2 = min(maxX*maxX, minX*minX);
	if(maxY*minY<0.0) //0 liegt dazwischen
		minY2 = 0.0;
	else
		minY2 = min(maxY*maxY, minY*minY);
	maxX2 = max(maxX*maxX, minX*minX);
	maxY2 = max(maxY*maxY, minY*minY);
	maxZ2 = max(maxZ*maxZ, minZ*minZ);

	//now calculate max cosine

	if(maxZ>=0.0)
	{
		if(sqrt(minX2+minY2+maxZ*maxZ) == 0.0)
			return 1.0;
		else
			return maxZ / sqrt(minX2+minY2+ maxZ*maxZ);
	}	
	else
		return 0.0; //WANN PASSIERT DAS? FALSCHE DREHUNG??
	/*
	{
		if(sqrt(maxX2+maxY2+maxZ*maxZ)==0.0)
			return 1.0;
		else
			return maxZ / sqrt(maxX2+maxY2+ maxZ*maxZ);
	}*/


}

double getGeometryTerm(Cluster* nextNode, double posx, double posy, double posz)
{
	Point light = nextNode->representativeLight->position;
	Vector lpos = Vector(light.x, light.y, light.z);
	Vector shadedPosition = Vector(posx, posy, posz);
	
	double distance = Scene::sceneUnitsToMeters*((lpos-shadedPosition).getLength());

	//TODO : tighter bound
	return min(1.0, (30.0)/((distance+3.0)*(distance+3.0)));
}

double getGeometryTermBound(Cluster* nextNode, double posx, double posy, double posz)
{	
	//d = differenzvektor containing (width, height, depth) of bbox
	Vector d = nextNode->bounds.B - nextNode->bounds.A;

	Vector v1 = Vector(d.x, 0.0f, 0.0f);
	Vector v2 = Vector(0.0f, d.y, 0.0f);
	Vector v3 = Vector(0.0f, 0.0f, d.z);

	Vector v = Vector(posx, posy, posz) - Vector(nextNode->bounds.A.x, nextNode->bounds.A.y, nextNode->bounds.A.z);

	//point is inside bounding box
	if (0.0f <= v*v1 && v*v1 <= v1*v1 && 
		0.0f <= v*v2 && v*v2 <= v2*v2 && 
		0.0f <= v*v3 && v*v3 <= v3*v3)
	{
		return 1.0f;
	}
	
	float minDist = 99999999.9f;

	minDist = min( abs(nextNode->bounds.A.x - posx), abs(nextNode->bounds.B.x - posx));
	minDist = min(min( abs(nextNode->bounds.A.y - posy), abs(nextNode->bounds.B.y - posy)), minDist);
	minDist = min(min( abs(nextNode->bounds.A.z - posz), abs(nextNode->bounds.B.z - posz)), minDist);
	
	minDist *= Scene::sceneUnitsToMeters;

	return min(1.0f, (30.0f)/((minDist+3.0f)*(minDist+3.0f)));

/*
	return 1.0f;
	
	Point light = nextNode->representativeLight->position;
	Vector lpos = Vector(light.x, light.y, light.z);
	Vector shadedPosition = Vector(posx, posy, posz);
	
	double distance = Scene::sceneUnitsToMeters*((lpos-shadedPosition).getLength());

	double maxCos = 1.0;

	//TODO : tighter bound
	return maxCos * min(1.0, (30.0)/((distance+3.0)*(distance+3.0)));
	*/
}
double getVisibility(Cluster* nextNode, double posx, double posy, double posz)
{
	int lspmX = nextNode->representativeLight->lightSpaceMatrixAtPixelX;
	int lspmY = nextNode->representativeLight->lightSpaceMatrixAtPixelY;
	int ismX = nextNode->representativeLight->shadowMapAtPixelX;
	int ismY = nextNode->representativeLight->shadowMapAtPixelY;

	float lsMatrix[16];

	for(int s=0; s<4; s++)
	{
		lsMatrix[s] =		lightSpaceMatrices[lspmY*lightSpaceMatrices_size*4 + lspmX*4  + s];
		lsMatrix[4+s] =		lightSpaceMatrices[lspmY*lightSpaceMatrices_size*4 + (lspmX+1)*4  + s];
		lsMatrix[8+s] =		lightSpaceMatrices[(lspmY+1)*lightSpaceMatrices_size*4 + lspmX*4  + s];
		lsMatrix[12+s] =	lightSpaceMatrices[(lspmY+1)*lightSpaceMatrices_size*4 + (lspmX+1)*4  + s];
	}
	
	//the shaded geometry position in world space
	float v[4] = {posx, posy, posz, 1.0};

	//vertex in light space
	float vLSpace[4];

	//rotate position to light space:
	for(int s=0; s<4; s++)
		vLSpace[s] = v[0]*lsMatrix[s] + v[1]*lsMatrix[4+s] + v[2]*lsMatrix[8+s] + v[3]*lsMatrix[12+s];

	//perspective division:
	for(int s=0; s<4; s++)
		vLSpace[s] /= vLSpace[3]; //vLSpace[3] is homogenous coordinate

//paraboloid mapping:
		
	// compute the halfway vector between viewing direction and direction to vertex			  
	float dist = sqrt(vLSpace[0]*vLSpace[0] + vLSpace[1]*vLSpace[1] + vLSpace[2]*vLSpace[2]);
	vLSpace[0] = (vLSpace[0] / dist);
	vLSpace[1] = (vLSpace[1] / dist);
	vLSpace[2] = (vLSpace[2] / dist) - 1.0;

  //adjust length so that z=-1
	vLSpace[0] /= -vLSpace[2];
    vLSpace[1] /= -vLSpace[2];		
		
//paraboloid code end
			
	//the vertex depth is the distance to the camera
	vLSpace[2] = (dist - GameStateVBO::nearZ) / (GameStateVBO::farZ-GameStateVBO::nearZ);
	
	vLSpace[0] = (vLSpace[0]+1.0) / 2.0;
	vLSpace[1] = (vLSpace[1]+1.0) / 2.0;
			
	//shadow test

	float distToM = sqrt(pow((vLSpace[0]-0.5), 2.0) + pow((vLSpace[1]-0.5), 2.0));
	

	Vector lvec = Vector(posx, posy, posz) - Vector(nextNode->representativeLight->position.x, nextNode->representativeLight->position.y, nextNode->representativeLight->position.z);
	lvec.normalize();
	bool behindlight = (lvec * nextNode->representativeLight->normal) < 0.0;
	
	//if position behindlight -> not visible
	if(behindlight)
		return 0.0f;

	//if at the very border of visible ism-> visible
	if(distToM > 0.48)
		return 1.0;
		
	int ismOffsetX = int(vLSpace[0] * ism_size);
	int ismOffsetY = int(vLSpace[1] * ism_size);

	float distXLeft = vLSpace[0]*ism_size - (float(ismOffsetX)); 
	float distYUp = vLSpace[1]*ism_size - (float(ismOffsetY)); 
	
	float* inShadow = new float[5];
	int valid =0;

	inShadow[valid++] = float(vLSpace[2] <= shadowMap[((ismY + ismOffsetY)*ism_atlas_size + ismX + ismOffsetX)*4]);
	float weightedSum = 1.0f;

	//simple shadow lookup:
	//return min(1.0f, inShadow[valid-1]);

	//averaged shadow lookup:
	if(((ismY + ismOffsetY - 1)*ism_atlas_size + ismX + ismOffsetX) >= 0)
	{
		inShadow[valid++] = float(vLSpace[2] <= shadowMap[((ismY + ismOffsetY - 1)*ism_atlas_size + ismX + ismOffsetX)*4]) * (1.0f- distYUp);
		weightedSum += (1.0f- distYUp);
	}

	if(((ismY + ismOffsetY + 1)*ism_atlas_size + ismX + ismOffsetX) < ism_atlas_size*ism_atlas_size)
	{
		inShadow[valid++] = float(vLSpace[2] <= shadowMap[((ismY + ismOffsetY + 1)*ism_atlas_size + ismX + ismOffsetX)*4]) * distYUp;
		weightedSum += distYUp;
	}

	if(((ismY + ismOffsetY)*ism_atlas_size + ismX + ismOffsetX - 1) >= 0)
	{
		inShadow[valid++] = float(vLSpace[2] <= shadowMap[((ismY + ismOffsetY)*ism_atlas_size + ismX + ismOffsetX - 1)*4]) * (1.0f- distXLeft);
		weightedSum += (1.0f- distXLeft);
	}

	if(((ismY + ismOffsetY)*ism_atlas_size + ismX + ismOffsetX + 1) < ism_atlas_size*ism_atlas_size)
	{
		inShadow[valid++] = float(vLSpace[2] <= shadowMap[((ismY + ismOffsetY)*ism_atlas_size + ismX + ismOffsetX + 1)*4]) * distXLeft;
		weightedSum += distXLeft;
	}

	float visibility=0.0f;
	for(int i=0; i<valid; i++)
	{
		visibility+=float(inShadow[i]);
	}

	delete[] inShadow;

	//shadow test
	return min(1.0f, (visibility / weightedSum));// + vLSpace[2]);
	
}
