#ifndef TEXTURE_H
#define TEXTURE_H

#include <vector>

//#include "../../GameStates/common.h"
//#include <Fl/Fl.h>
//#include <Fl/Gl.h>
//#include <FL/Glu.h>

#include "../ModelManager/ModelManager.h"

#include "../../Math/AreaLight.h"
#include <list>
#include "../../Math/CudaStuff.h"
#include "../Scene.h"
#include "../../CUDA/CUDA_PushPull.h"
#include "../../CUDA/CUDA_Lightcuts.h"
#include "../../GameStates/GameStateVBO.h"
#include <queue>

#include <sstream>
#include <stdio.h>
#include "libtarga.h"


#pragma warning (push)              // Disable the warning between the block of push and pop
// The list of numbers of warning we will like to disable 
#pragma warning(disable : 4204)
#pragma warning(disable : 4244)
#pragma warning(disable : 4312)
#pragma warning(disable : 4311)
#pragma warning(disable : 4083)
#pragma warning(disable : 4193)
#pragma warning(disable : 4996)
// List of all fltk include file for example #include <Fl/Fl_Window.h>
#include <Fl/Fl.h>
#include <Fl/Fl_Widget.h>
#include <FL/Fl_Image.h>
#include <FL/Fl_JPEG_Image.h>
#include <FL/Fl_PNG_Image.h>
#include <FL/Fl_BMP_Image.h>
#include <FL/Fl_GIF_Image.h>
#include <Fl/filename.h>
#pragma warning (pop) // Enable the warning again

#include <math.h>
#include "../../Math/RaytracingMath.h"
#include "../../Math/CudaStuff.h"

class Cluster;
class CpuLightTree;

class Texture 
{
	public:
		
		~Texture(void);
		// What we give to LibTarga
//		char* name;                   //
//		char* fname;					// the file name

		// What libtarga gives back...
		unsigned byte* bits;
		int   width;
		int   height;
		GLuint glName;
		char* texName;

};

class TextureManager
{
	public:
		static GLuint lc, lp, ln;
		static int lightTextureH, lightTextureW;
		static float* pixelsLC;
		static float* pixelsLP;
		static float* pixelsLN;
		static int anzLights;
		static deque <Cluster*> lightDeque;
		static CpuLightTree* lightTree;

		static float sumOfWeightedIntensities;

		static std::vector<Texture*> allTextures; 

		static std::vector<VPL*> vplVector; 

		static char* texturePaths[];

		static float sumOfCutsizes;

		// Get a model- will hunt it down and load it if its not already loaded
		static GLuint fetchTexture(char* textureName, bool wrapS, bool wrapT);

		static void deleteAllTextures(void);

		static Texture* getTextureObject(char* textureName);

		static void generateLightTextures();
		static void createkDTreeAndClusterLightTree(void);

		static void shadeLightcuts(double posx, double posy, double posz, double nx, double ny, double nz, double cr, double cg, double cb, float* retVal);
		static void shadeIR(double posx, double posy, double posz, double nx, double ny, double nz, double cr, double cg, double cb, float* retVal);



};

#endif
